//***************************************************************************
//
// ARCBALL.CPP
// by Harlen C. Batagelo
//
//***************************************************************************

#include <stdio.h>
#include "arcball.h"

/////////////////////////////////////////////////////////////////////////////
// Constructor
/////////////////////////////////////////////////////////////////////////////
CArcBall::CArcBall()
{
    QuaternionIdentity( &m_qDown );
    QuaternionIdentity( &m_qNow );
    MatrixIdentity( &m_matRotation );
    MatrixIdentity( &m_matRotationDelta );
    MatrixIdentity( &m_matTranslation );
    MatrixIdentity( &m_matTranslationDelta );
    m_bDrag = false;
} // CArcBall::CArcBall

/////////////////////////////////////////////////////////////////////////////
// Initialize arcball with window settings
/////////////////////////////////////////////////////////////////////////////
void CArcBall::SetWindow( int iWidth, int iHeight, float fRadius )
{
    m_iWidth  = iWidth;
    m_iHeight = iHeight;
    m_fRadius = fRadius;
} // CArcBall::SetWindow

/////////////////////////////////////////////////////////////////////////////
// Convert screen coordinates to vector coordinates
/////////////////////////////////////////////////////////////////////////////
VECTOR3 CArcBall::ScreenToVector( int sx, int sy )
{
    // scale to screen
    float x   =  (sx - m_iWidth/2)  / (m_fRadius*m_iWidth/2);
    float y   = -(sy - m_iHeight/2) / (m_fRadius*m_iHeight/2);
    float z   = 0.0f;
    float mag = x*x + y*y;

    if( mag > 1.0f )
    {
        float scale = 1.0f/sqrtf(mag);
        x *= scale;
        y *= scale;
    }
    else
        z = sqrtf( 1.0f - mag );

    return VECTOR3( x, y, z );
} // CArcBall::ScreenToVector

/////////////////////////////////////////////////////////////////////////////
// Set arcball radius
/////////////////////////////////////////////////////////////////////////////
void CArcBall::SetRadius( float fRadius )
{
    m_fRadius2 = fRadius;
} // CArcBall::SetRadius

/////////////////////////////////////////////////////////////////////////////
// Set arcball when left button is down
/////////////////////////////////////////////////////////////////////////////
void CArcBall::LeftButtonDown (int iMouseX, int iMouseY)
{
	m_bDrag = true;
	m_vDown = ScreenToVector (iMouseX, iMouseY);
} // CArcBall::LeftButtonDown

/////////////////////////////////////////////////////////////////////////////
// Set arcball when left button is up
/////////////////////////////////////////////////////////////////////////////
void CArcBall::LeftButtonUp ()
{
	m_bDrag = false;
	m_qDown = m_qNow;
} // CArcBall::LeftButtonUp

/////////////////////////////////////////////////////////////////////////////
// Set arcball when left button is down and moving
/////////////////////////////////////////////////////////////////////////////
void CArcBall::LeftButtonDownMove (int iMouseX, int iMouseY)
{
	if (m_bDrag)
	{
		VECTOR3 vPart;
		VECTOR3 vCur = ScreenToVector (iMouseX, iMouseY);
		Vec3Cross (&vPart, &m_vDown, &vCur);

		m_qNow = m_qDown * QUATERNION (vPart.x, vPart.y, vPart.z,
									Vec3Dot (&m_vDown, &vCur));
	}

	QUATERNION qConj;
	QuaternionConjugate( &qConj, &m_qNow );

	MatrixRotationQuaternion( &m_matRotationDelta, &qConj );
	MatrixTranspose( &m_matRotationDelta, &m_matRotationDelta );
	MatrixMultiply( &m_matRotation, &m_matRotation, &m_matRotationDelta );
	
	QuaternionIdentity( &m_qDown );
	QuaternionIdentity( &m_qNow );
	m_vDown = ScreenToVector( iMouseX, iMouseY );
	m_bDrag = true;
} // CArcBall::LeftButtonDownMove