//***************************************************************************
//
// DISK.CPP
// by Harlen C. Batagelo
//
//***************************************************************************

#if defined(WIN32)
#include <windows.h>
#pragma warning (disable:4244)	// disable bogus conversion warnings
#endif

#include <math.h>
#include <GL/gl.h>

#include "disk.h"

#define PI2 6.28318530717958647692528676655901 // PI * 2

/////////////////////////////////////////////////////////////////////////////
// Move disk to next position according to the assigned velocity
/////////////////////////////////////////////////////////////////////////////
void CDisk::Move ()
{
	pos_x += inc_x;
	pos_y += inc_y;
	pos_z += inc_z;

	if (pos_x > size_x - externalRadius - 3) inc_x = -inc_x;
	if (pos_y > size_y - externalRadius - 3) inc_y = -inc_y;
	if (pos_z > size_z - externalRadius - 3) inc_z = -inc_z;
	if (pos_x < 3 + externalRadius) inc_x = -inc_x;
	if (pos_y < 3 + externalRadius) inc_y = -inc_y;
	if (pos_z < 3 + externalRadius) inc_z = -inc_z;
} // CDisk::Move

/////////////////////////////////////////////////////////////////////////////
// Draw a filled disk in light red with an outline in dark red
/////////////////////////////////////////////////////////////////////////////
void CDisk::DrawDisk (Vis3D *v3dpt)
{
	if (v3dpt->bCalcExactVis) // use item buffer?
		glColor3ub (ucItemR, ucItemG, ucItemB);
	else
		glMaterialfv (GL_FRONT, GL_DIFFUSE, mat_diffuse);

	glPushMatrix ();
		glTranslatef (pos_x, pos_y, pos_z);

		rot_x = rot_x + inc_rot_x; // increment rotation
		rot_y = rot_y + inc_rot_y;
		rot_z = rot_z + inc_rot_z;
		glRotatef (rot_x, 1.0, 0.0, 0.0);
		glRotatef (rot_y, 0.0, 1.0, 0.0);
		glRotatef (rot_z, 0.0, 0.0, 1.0);

		switch (disk_type) // teapot or sphere?
		{
		case 0:
			glutSolidSphere (internalRadius, slices, stacks);
			break;
		case 1:
			glutSolidSphere (internalRadius, slices, stacks);
			break;
		}
	glPopMatrix ();
} // CDisk::DrawDisk