//***************************************************************************
//
// DISK.H
// by Harlen C. Batagelo
//
//***************************************************************************

#ifndef _DISK_H
#define _DISK_H

#include "common.h"

/////////////////////////////////////////////////////////////////////////////
// Define a disk
/////////////////////////////////////////////////////////////////////////////
class CDisk
{
private:
	GLfloat mat_diffuse[4];			// disk diffuse color
	unsigned char ucItemR, ucItemG, ucItemB; // color for item buffer
	float pos_x, pos_y, pos_z;		// disk position
	float inc_x, inc_y, inc_z;		// velocity
	float rot_x, rot_y, rot_z;		// current rotation (internal)
	float inc_rot_x, inc_rot_y, inc_rot_z; // rotation increment
	float internalRadius;			// internal radius of the disk
	float externalRadius;			// external radius of the disk
	float slices, stacks;			// sphere resolution
	float size_x, size_y, size_z;	// width, height and depth of the moving plane

	// TBV date
	bool has_tbv;					// TRUE if this object has a TBV
	float tbv_x, tbv_y, tbv_z;		// TBV position
	long start_date;				// TBV start date (absolute date)
	long expiration_date;			// TBV expiration date (absolute date)
	float last_x, last_y, last_z;	// disk position at the last time this disk was drawn
	float TBV_radius;				// radius of the TBV

	int disk_type; // 0 = teapot, 1 = sphere

public:
	// used to track cells spanned by the object
	VECTOR3 vI[16384]; // ID cells occupied
	int iICells; // number of elements of vI

	CDisk ()
	{
		internalRadius = 0;			// default settings
		externalRadius = 0;
		slices = 25;//12; // 6 // 264 polygons each one
		stacks = 25;//12; // 6
		// 1200

		pos_x = pos_y = pos_z = 0.0;
		inc_x = inc_y = inc_z = 1.0; 
		rot_x = rot_y = rot_z = 0.0;
		inc_rot_x = inc_rot_y = inc_rot_z = 1.0;
		size_x = size_y = size_z = 0;
		disk_type = 1;
		ucItemR = ucItemG = ucItemB = 0;

		iICells = 0;
	}

	inline void SetType (int type) { disk_type = type; }
	inline void SetPos (float x, float y, float z) { pos_x = x; pos_y = y; pos_z = z; }
	inline void SetInternalRadius (float r) { internalRadius = r; }
	inline void SetExternalRadius (float r) { externalRadius = r; }
	inline void SetSlices (float s) { slices = s; }
	inline void SetStacks (float s) { stacks = s; }
	inline void SetColor (float fR, float fG, float fB) 
	{ 
		mat_diffuse[0] = fR; 
		mat_diffuse[1] = fG;
		mat_diffuse[2] = fB; 
		mat_diffuse[3] = 1.0;
	}
	inline void SetItemColor (unsigned char ucR, unsigned char ucG, unsigned char ucB) 
	{ 
		ucItemR = ucR;
		ucItemG = ucG;
		ucItemB = ucB;
	}
	inline void SetRot (float x, float y, float z) { rot_x = x; rot_y = y; rot_z = z; }
	inline void SetRotInc (float x, float y, float z) { inc_rot_x = x; inc_rot_y = y; inc_rot_z = z; }
	inline void SetVelocityX (float v) { inc_x = v; }
	inline void SetVelocityY (float v) { inc_y = v; }
	inline void SetVelocityZ (float v) { inc_z = v; }
	inline void SetBounds (float x, float y, float z) { size_x = x; size_y = y; size_z = z; }
	inline void SetStartDate (long date) { start_date = date; }
	inline void SetExpDate (long date) { expiration_date = date; }
	inline void SetTBVX (float x) { tbv_x = x; }
	inline void SetTBVY (float y) { tbv_y = y; }
	inline void SetTBVZ (float z) { tbv_z = z; }
	inline void SetLastX (float x) { last_x = x; }
	inline void SetLastY (float y) { last_y = y; }
	inline void SetLastZ (float z) { last_z = z; }
	inline void SetTBVState (bool s) { has_tbv = s; }
	inline void SetTBVRadius (int r) { TBV_radius = (float)r; }

	inline float GetX () { return pos_x; }
	inline float GetY () { return pos_y; }
	inline float GetZ () { return pos_z; }
	inline float GetInternalRadius () { return internalRadius; }
	inline float GetExternalRadius () { return externalRadius; }
	inline float GetSlices () { return slices; }
	inline float GetStacks () { return stacks; }
	inline float GetVelocityX () { return inc_x; }
	inline float GetVelocityY () { return inc_y; }
	inline float GetVelocityZ () { return inc_z; }
	inline float GetBoundX () { return size_x; }
	inline float GetBoundY () { return size_y; }
	inline float GetBoundZ () { return size_z; }
	inline long GetStartDate () { return start_date; }
	inline long GetExpDate () { return expiration_date; }
	inline long GetTBVX () { return (long)tbv_x; }
	inline long GetTBVY () { return (long)tbv_y; }
	inline long GetTBVZ () { return (long)tbv_z; }
	inline float GetLastX () { return last_x; }
	inline float GetLastY () { return last_y; }
	inline float GetLastZ () { return last_z; }
	inline bool HasTBV () { return has_tbv; }
	inline float GetTBVRadius () { return TBV_radius; }

	void Move ();		// move disk to next position
	void DrawDisk (Vis3D *v3dpt); // draw a filled disk
};

#endif