//***************************************************************************
//
// GEOPRI.CPP
// by Harlen C. Batagelo
//
//***************************************************************************

#include "common.h"

int iX, iY, iZ;

#define EPSILON (.375)

/////////////////////////////////////////////////////////////////////////////
// Convert a raster image to a simple vectorized representation
// iRas: raster plane
// iSize: raster plane dimensions (iSize X iSize)
/////////////////////////////////////////////////////////////////////////////
void Vis3D::ras2Vec (int *iRas, int iSize)
{
	// look for hidden cells inside the range of iRas
	if (fusionEnabled)
	{
		switch (iRasterPlane)
		{
		case raFront:
			for (iY = (int)fCPMinY; iY <= (int)fCPMaxY; iY++)
				for (iX = (int)fCPMinX; iX <= (int)fCPMaxX; iX++)
					if(H.get(MK_OFF_GRID((-iX+iDist)+iOrX,(-iY+iDist)+iOrY,iDist+iOrZ)))
						iRas[MK_OFF(iY,iX,iDistSize)] = 1;
			break;
		case raBack:
			for (iY = (int)fCPMinY; iY <= (int)fCPMaxY; iY++)
				for (iX = (int)fCPMinX; iX <= (int)fCPMaxX; iX++)
					if(H.get(MK_OFF_GRID((iX-iDist)+iOrX,(-iY+iDist)+iOrY,-iDist+iOrZ)))
						iRas[MK_OFF(iY,iX,iDistSize)] = 1;
			break;
		case raLeft:
			for (iY = (int)fCPMinY; iY <= (int)fCPMaxY; iY++)
				for (iX = (int)fCPMinX; iX <= (int)fCPMaxX; iX++)
					if(H.get(MK_OFF_GRID(iDist+iOrX,(-iY+iDist)+iOrY,(iX-iDist)+iOrZ)))
						iRas[MK_OFF(iY,iX,iDistSize)] = 1;
			break;
		case raRight:
			for (iY = (int)fCPMinY; iY <= (int)fCPMaxY; iY++)
				for (iX = (int)fCPMinX; iX <= (int)fCPMaxX; iX++)
					if(H.get(MK_OFF_GRID(-iDist+iOrX,(-iY+iDist)+iOrY,(-iX+iDist)+iOrZ)))
						iRas[MK_OFF(iY,iX,iDistSize)] = 1;
			break;
		case raTop:
			for (iY = (int)fCPMinY; iY <= (int)fCPMaxY; iY++)
				for (iX = (int)fCPMinX; iX <= (int)fCPMaxX; iX++)
					if(H.get(MK_OFF_GRID((-iX+iDist)+iOrX,iDist+iOrY,(iY-iDist)+iOrZ)))
						iRas[MK_OFF(iY,iX,iDistSize)] = 1;
			break;
		case raDown:
			for (iY = (int)fCPMinY; iY <= (int)fCPMaxY; iY++)
				for (iX = (int)fCPMinX; iX <= (int)fCPMaxX; iX++)
					if(H.get(MK_OFF_GRID((-iX+iDist)+iOrX, -iDist+iOrY, (-iY+iDist)+iOrZ)))
						iRas[MK_OFF(iY,iX,iDistSize)] = 1;
			break;
		}
	}

	unsigned char iValCur, iValPrev = 0;
	char cHFlag = 0, cVFlag = 0;
	int iHXStart, iVYStart, iISave, iJSave;

	bRasterFull = true;

	// reset
	iVertices = 0;
	iHLines = 0;
	cpHLine[0].Reset ();
	cpHLineEnd[0].Reset ();

	bool bIncreaseH, bIncreaseV, bDontAdd;

	int i, j;

	// horizontal spans (vertical sweep)
	for (i = (int)fCPMinY; i <= (int)fCPMaxY; i++)
	{
		bIncreaseH = false;

		for (j = (int)fCPMinX; j <= (int)fCPMaxX; j++)
		{
			iValCur = iRas[MK_OFF(i,j,iSize)];

			// look for internal horizontal interval vertices

			if (!iValCur) bRasterFull = false;
			else
			{
				if (!iValPrev)	{ cHFlag = 1; iHXStart = j; }

				// new horizontal interval is coming...
				// look extension below

				if (i == (int)fCPMinY || ((i != (int)fCPMinY) && (!iRas[MK_OFF(i-1,j,iSize)])))
				{
					iISave = i;
					while ((iISave < (int)fCPMaxY) && iRas[MK_OFF(iISave+1,j,iSize)]) iISave++;
					// add vertex
					if (iRas[MK_OFF(iISave,j,iSize)] == 1)
					{
						iRas[MK_OFF(iISave,j,iSize)] = iVertices+2;
						vtVertices[iVertices].Set (j, iISave);
						iVertices++;
					}

					// set ending vertex
					bDontAdd = false;
					int iYTemp = vtVertices[iRas[MK_OFF(iISave,j,iSize)]-2].fY;
					for (int iCt = 0; iCt < cpHLineEnd[iHLines].iElements; iCt++)
					{
						int iYTemp2 = vtVertices[cpHLineEnd[iHLines].iIdxEnd[iCt]].fY;
						if (iYTemp == iYTemp2)
							bDontAdd = true;
					}
					if (!bDontAdd)
					{
						cpHLineEnd[iHLines].Add (iRas[MK_OFF(iISave,j,iSize)]-2);
						bDontAdd = true;
					}
				}

			}

			if (!iValCur && iValPrev && cHFlag)
			{
				// set first vertex
				if (iRas[MK_OFF(i,iHXStart,iSize)] == 1)
				{
					iRas[MK_OFF(i,iHXStart,iSize)] = iVertices+2;
					vtVertices[iVertices].Set (iHXStart, i);
					iVertices++;
				}
				cpHLine[iHLines].Set1stVertex (iRas[MK_OFF(i,iHXStart,iSize)]-2);

				// set second vertex
				if (iRas[MK_OFF(i,j-1,iSize)] == 1)
				{
					iRas[MK_OFF(i,j-1,iSize)] = iVertices+2;
					vtVertices[iVertices].Set (j-1, i);
					iVertices++;
				}
				cpHLine[iHLines].Set2ndVertex (iRas[MK_OFF(i,j-1,iSize)]-2);

				cpHLine[iHLines].NewEdge ();
				cHFlag = 0;

				bIncreaseH = true;
			}
			iValPrev = iValCur;
		}

		if (iValCur && cHFlag)
		{
			// set first vertex
			if (iRas[MK_OFF(i,iHXStart,iSize)] == 1)
			{
				iRas[MK_OFF(i,iHXStart,iSize)] = iVertices+2;
				vtVertices[iVertices].Set (iHXStart, i);
				iVertices++;
			}
			cpHLine[iHLines].Set1stVertex (iRas[MK_OFF(i,iHXStart,iSize)]-2);

			// set second vertex
			if (iRas[MK_OFF(i,j-1,iSize)] == 1)
			{
				iRas[MK_OFF(i,j-1,iSize)] = iVertices+2;
				vtVertices[iVertices].Set (j-1, i);
				iVertices++;
			}
			cpHLine[iHLines].Set2ndVertex (iRas[MK_OFF(i,j-1,iSize)]-2);

			cpHLine[iHLines].NewEdge ();

			bIncreaseH = true;
		}

		if (bIncreaseH)
		{
			iHLines++;
			cpHLine[iHLines].Reset ();
			cpHLineEnd[iHLines].Reset ();
		}

		cHFlag    = 0;
		iValPrev = 0;
	}

	// one more condition to test bRasterFull
	if ((int)fCPMinY || (int)fCPMinX || (int)fCPMaxY != iSize-1 || (int)fCPMaxX != iSize-1)
		bRasterFull = false;

	iValPrev = 0;
	cpVLine[0].Reset ();
	cpVLineEnd[0].Reset ();

	iVLines = 0;

	// vertical spans (horizontal sweep)
	for (j = (int)fCPMinX; j <= (int)fCPMaxX; j++)
	{
		bIncreaseV = false;

		for (i = (int)fCPMinY; i <= (int)fCPMaxY; i++)
		{
			iValCur = iRas[MK_OFF(i,j,iSize)];

			// look for internal vertical interval vertices

			if (iValCur)
			{
				if (!iValPrev) { cVFlag = 1; iVYStart = i; }

				// new vertical interval is coming...
				// look extension at right

				if (j == (int)fCPMinX || ((j != (int)fCPMinX) && (!iRas[MK_OFF(i,j-1,iSize)])))
				{
					iJSave = j;
					while ((iJSave < (int)fCPMaxX) && iRas[MK_OFF(i, iJSave+1,iSize)]) iJSave++;
					// add vertex
					if (iRas[MK_OFF(i,iJSave,iSize)] == 1)
					{
						iRas[MK_OFF(i,iJSave,iSize)] = iVertices+2;
						vtVertices[iVertices].Set (i, iJSave);
						iVertices++;
					}
					
					// set ending vertex
					bDontAdd = false;
					int iXTemp = vtVertices[iRas[MK_OFF(i,iJSave,iSize)]-2].fX;
					for (int iCt = 0; iCt < cpVLineEnd[iVLines].iElements; iCt++)
					{
						int iXTemp2 = vtVertices[cpVLineEnd[iVLines].iIdxEnd[iCt]].fX;
						if (iXTemp == iXTemp2)
							bDontAdd = true;
					}
					if (!bDontAdd)
					{
						cpVLineEnd[iVLines].Add (iRas[MK_OFF(i,iJSave,iSize)]-2);
						bDontAdd = true;
					}
				}
			}

			if (!iValCur && iValPrev && cVFlag)
			{
				// set first vertex
				if (iRas[MK_OFF(iVYStart,j,iSize)] == 1)
				{
					iRas[MK_OFF(iVYStart,j,iSize)] = iVertices+2;
					vtVertices[iVertices].Set (j, iVYStart);
					iVertices++;
				}
				cpVLine[iVLines].Set1stVertex (iRas[MK_OFF(iVYStart,j,iSize)]-2);

				// set second vertex
				if (iRas[MK_OFF(i-1,j,iSize)] == 1)
				{
					iRas[MK_OFF(i-1,j,iSize)] = iVertices+2;
					vtVertices[iVertices].Set (j, i-1);
					iVertices++;
				}
				cpVLine[iVLines].Set2ndVertex (iRas[MK_OFF(i-1,j,iSize)]-2);

				cpVLine[iVLines].NewEdge ();
				cVFlag = 0;

				bIncreaseV = true;
			}
			iValPrev = iValCur;
		}

		if (iValCur && cVFlag)
		{
			// set first vertex
			if (iRas[MK_OFF(iVYStart,j,iSize)] == 1)
			{
				iRas[MK_OFF(iVYStart,j,iSize)] = iVertices+2;
				vtVertices[iVertices].Set (j, iVYStart);
				iVertices++;
			}
			cpVLine[iVLines].Set1stVertex (iRas[MK_OFF(iVYStart,j,iSize)]-2);

			// set second vertex
			if (iRas[MK_OFF(i-1,j,iSize)] == 1)
			{
				iRas[MK_OFF(i-1,j,iSize)] = iVertices+2;
				vtVertices[iVertices].Set (j, i-1);
				iVertices++;
			}
			cpVLine[iVLines].Set2ndVertex (iRas[MK_OFF(i-1,j,iSize)]-2);

			cpVLine[iVLines].NewEdge ();

			bIncreaseV = true;
		}

		if (bIncreaseV) 
		{
			iVLines++;
			cpVLine[iVLines].Reset ();
			cpVLineEnd[iVLines].Reset ();
		}

		cVFlag    = 0;
		iValPrev = 0;
	}
} // Vis3D::ras2Vec

static int iSpanEnd[512];

/////////////////////////////////////////////////////////////////////////////
// Rasterize the vectorized representation given by ras2Vec
// 'Front' version
/////////////////////////////////////////////////////////////////////////////
void Vis3D::drawConcPolF (void)
{
	int iCt;

	//
	// compute the increment of each vertex and store it for further use
	//
	for (iCt = 0; iCt < iVertices; iCt++)
		vtVertices[iCt].SetInc ((float)(vtVertices[iCt].fX-iDist)/iDist, 
								(float)(vtVertices[iCt].fY-iDist)/iDist);

	iZ = iDist;

	float fCPMinXInc = (float)(fCPMinX-iDist)/iDist,
		  fCPMaxXInc = (float)(fCPMaxX-iDist)/iDist,
		  fCPMinYInc = (float)(fCPMinY-iDist)/iDist,
		  fCPMaxYInc = (float)(fCPMaxY-iDist)/iDist;

	for (int iDepth = 0; iDepth < iMaxDist-iDist; iDepth++)
	{

		iZ++;

		// return if the current concave polygon slice is out of bounds
		if ((iZ+iOrZ) > SIZE_ZM1) return;

		//
		// increment all vertices of the raster plane
		//
		for (iCt = 0; iCt < iVertices; iCt++)
			vtVertices[iCt].Increment ();

		// min/max values must also be transformed
		fCPMinX += fCPMinXInc; fCPMaxX += fCPMaxXInc;
		fCPMinY += fCPMinYInc; fCPMaxY += fCPMaxYInc;

		//
		// rasterize
		//

		int iNSpanEnd = 0;
		int icpCounterH = -1;
		char cChanged = 1;
		int iEg;

		// vertical sweep
		for (iY = ROUND(fCPMinY+EPSILON); iY <= ROUND(fCPMaxY-EPSILON); iY++)
		{
			// increment icpCounterH?
			if (icpCounterH < iHLines-1)
				if (ROUND(vtVertices[cpHLine[icpCounterH+1].iIdx1[0]].fY+EPSILON) <= iY)
				{
					cChanged = 0;
					icpCounterH++;

					// add to the span set the vertical lines starting here
					for (iEg = 0; iEg < cpHLineEnd[icpCounterH].iElements; iEg++)
						iSpanEnd[iNSpanEnd++] = ROUND(vtVertices[cpHLineEnd[icpCounterH].iIdxEnd[iEg]].fY+EPSILON);

					// look for ending vertical spans
					for (iEg = 0; iEg < iNSpanEnd; iEg++)
						if (iSpanEnd[iEg] == iY)
							cChanged = 1;
				}

			// if there aren't horizontal spans and vertical spans didn't change,
			// draw current horizontal lines (if any)
			if (!cChanged)
			{
				int iYTrans = (-iY+iDist)+iOrY;
				if (!(iYTrans < 0 || iYTrans > SIZE_YM1))
				{
					// horizontal rasterization
					for (iEg = 0; iEg < cpHLine[icpCounterH].iEdges; iEg++)
					{	
						// end/start inverted
						int iXEnd   = (-ROUND(vtVertices[cpHLine[icpCounterH].iIdx1[iEg]].fX+EPSILON)+iDist)+iOrX;
						int iXStart = (-ROUND(vtVertices[cpHLine[icpCounterH].iIdx2[iEg]].fX-EPSILON)+iDist)+iOrX;

						if (iXStart < 0) iXStart = 0;
						if (iXEnd > SIZE_XM1) iXEnd = SIZE_XM1;

						if (iXEnd >= iXStart)
							H.set (MK_OFF_GRID(iXStart,iYTrans,iZ+iOrZ), (iXEnd-iXStart)+1);
					} 
				}

				// look for ending vertical spans
				for (iEg = 0; iEg < iNSpanEnd; iEg++)
					if (iSpanEnd[iEg] == iY)
						cChanged = 1;
			}
		}

		iNSpanEnd = 0;
		int icpCounterV = -1;
		cChanged = 1;

		// horizontal sweep
		for (iX = ROUND(fCPMinX+EPSILON); iX <= ROUND(fCPMaxX-EPSILON); iX++)
		{
			// increment icpCounterV?
			if (icpCounterV < iVLines-1)
				if (ROUND(vtVertices[cpVLine[icpCounterV+1].iIdx1[0]].fX+EPSILON) <= iX)
				{
					cChanged = 0;
					icpCounterV++;

					// add to the span set the horizontal lines starting here
					for (iEg = 0; iEg < cpVLineEnd[icpCounterV].iElements; iEg++)
						iSpanEnd[iNSpanEnd++] = ROUND(vtVertices[cpVLineEnd[icpCounterV].iIdxEnd[iEg]].fX+EPSILON);

					// look for ending horizontal spans
					for (iEg = 0; iEg < iNSpanEnd; iEg++)
						if (iSpanEnd[iEg] == iX)
							cChanged = 1;
				}

			// if there aren't vertical spans and horizontal spans didn't change...
			// draw current vertical lines (if any)
			if (!cChanged)
			{
				int iXTrans = (-iX+iDist)+iOrX;
				if (!(iXTrans < 0 || iXTrans > SIZE_XM1))
				{
					// vertical rasterization
					for (iEg = 0; iEg < cpVLine[icpCounterV].iEdges; iEg++)
					{
						// end/start inverted
						int iYEnd   = (-ROUND(vtVertices[cpVLine[icpCounterV].iIdx1[iEg]].fY+EPSILON)+iDist)+iOrY;
						int iYStart = (-ROUND(vtVertices[cpVLine[icpCounterV].iIdx2[iEg]].fY-EPSILON)+iDist)+iOrY;

						if (iYStart < 0) iYStart = 0;
						if (iYEnd > SIZE_YM1) iYEnd = SIZE_YM1;

						// front plane (x' = -x) (y' = -y) (z' = z++)
						for (iY = iYStart; iY <= iYEnd; iY++)
							H.set(MK_OFF_GRID(iXTrans,iY,iZ+iOrZ));
					}
				}

				// look for ending horizontal spans
				for (iEg = 0; iEg < iNSpanEnd; iEg++)
					if (iSpanEnd[iEg] == iX)
						cChanged = 1;
			} 
		}
	}
} // Vis3D::drawConcPolF

/////////////////////////////////////////////////////////////////////////////
// Rasterize the vectorized representation given by ras2Vec
// 'Back' version
/////////////////////////////////////////////////////////////////////////////
void Vis3D::drawConcPolB (void)
{
	int iCt;

	//
	// compute the increment of each vertex and store it for further use
	//
	for (iCt = 0; iCt < iVertices; iCt++)
		vtVertices[iCt].SetInc ((float)(vtVertices[iCt].fX-iDist)/iDist, 
								(float)(vtVertices[iCt].fY-iDist)/iDist);

	iZ = iDist;

	float fCPMinXInc = (float)(fCPMinX-iDist)/iDist,
		  fCPMaxXInc = (float)(fCPMaxX-iDist)/iDist,
		  fCPMinYInc = (float)(fCPMinY-iDist)/iDist,
		  fCPMaxYInc = (float)(fCPMaxY-iDist)/iDist;

	for (int iDepth = 0; iDepth < iMaxDist-iDist; iDepth++)
	{
		iZ++;

		// return if the current concave polygon slice is out of bounds
		if ((-iZ+iOrZ) < 0) return;

		//
		// increment all vertices of the raster plane
		//
		for (iCt = 0; iCt < iVertices; iCt++)
			vtVertices[iCt].Increment ();

		// min/max values must also be transformed
		fCPMinX += fCPMinXInc; fCPMaxX += fCPMaxXInc; 
		fCPMinY += fCPMinYInc; fCPMaxY += fCPMaxYInc;

		//
		// rasterize
		//

		int iNSpanEnd = 0;
		int icpCounterH = -1;
		char cChanged = 1;
		int iEg;

		// vertical sweep
		for (iY = ROUND(fCPMinY+EPSILON); iY <= ROUND(fCPMaxY-EPSILON); iY++)
		{
			// increment icpCounterH?
			if (icpCounterH < iHLines-1)
				if (ROUND(vtVertices[cpHLine[icpCounterH+1].iIdx1[0]].fY+EPSILON) <= iY)
				{
					cChanged = 0;

					// add to the span set the vertical lines starting here
					for (iEg = 0; iEg < cpHLineEnd[icpCounterH].iElements; iEg++)
						iSpanEnd[iNSpanEnd++] = ROUND(vtVertices[cpHLineEnd[icpCounterH].iIdxEnd[iEg]].fY+EPSILON);

					// look for ending vertical spans
					for (iEg = 0; iEg < iNSpanEnd; iEg++)
						if (iSpanEnd[iEg] == iY)
							cChanged = 1;

					icpCounterH++;
				}

			// if there aren't horizontal spans and vertical spans didn't change,
			// draw current horizontal lines (if any)
			if (!cChanged)
			{

				int iYTrans = (-iY+iDist)+iOrY;
				if (!(iYTrans < 0 || iYTrans > SIZE_YM1))
				{
					// horizontal rasterization
					for (iEg = 0; iEg < cpHLine[icpCounterH].iEdges; iEg++)
					{
						for (iX = ROUND(vtVertices[cpHLine[icpCounterH].iIdx1[iEg]].fX+EPSILON); 
							iX <= ROUND(vtVertices[cpHLine[icpCounterH].iIdx2[iEg]].fX-EPSILON); 
							iX++)
						{
							// back plane (x' = x) (y' = -y) (z' = z--)
							int iXTrans = ( iX-iDist)+iOrX;
							if (!(iXTrans < 0 || iXTrans > SIZE_XM1))
								H.set(MK_OFF_GRID(iXTrans,iYTrans,-iZ+iOrZ));
						}
					}
				}

				// look for ending vertical spans
				for (iEg = 0; iEg < iNSpanEnd; iEg++)
					if (iSpanEnd[iEg] == iY)
						cChanged = 1;
			}
		}

		cChanged = 0;
		// look for ending vertical spans
		for (iEg = 0; iEg < iNSpanEnd; iEg++)
			if (iSpanEnd[iEg] == iY)
				cChanged = 1;

		iY--;

		// if there aren't horizontal spans and vertical spans didn't change,
		// draw current horizontal lines (if any)
		if (!cChanged)
		{
			int iYTrans = (-iY+iDist)+iOrY;
			if (!(iYTrans < 0 || iYTrans > SIZE_YM1))
			{
				// horizontal rasterization
				for (iEg = 0; iEg < cpHLine[icpCounterH].iEdges; iEg++)
				{
					for (iX = ROUND(vtVertices[cpHLine[icpCounterH].iIdx1[iEg]].fX+EPSILON); 
						iX <= ROUND(vtVertices[cpHLine[icpCounterH].iIdx2[iEg]].fX-EPSILON); 
						iX++)
					{
						// back plane (x' = x) (y' = -y) (z' = z--)
						int iXTrans = ( iX-iDist)+iOrX;
						if (!(iXTrans < 0 || iXTrans > SIZE_XM1))
							H.set(MK_OFF_GRID(iXTrans,iYTrans,-iZ+iOrZ));
					}
				}
			}
		}


		iNSpanEnd = 0;
		int icpCounterV = -1;
		cChanged = 1;

		// horizontal sweep
		for (iX = ROUND(fCPMinX+EPSILON); iX <= ROUND(fCPMaxX-EPSILON); iX++)
		{
			// increment icpCounterV?
			if (icpCounterV < iVLines-1)
				if (ROUND(vtVertices[cpVLine[icpCounterV+1].iIdx1[0]].fX+EPSILON) <= iX)
				{
					cChanged = 0;

					// add to the span set the horizontal lines starting here
					for (iEg = 0; iEg < cpVLineEnd[icpCounterV].iElements; iEg++)
						iSpanEnd[iNSpanEnd++] = ROUND(vtVertices[cpVLineEnd[icpCounterV].iIdxEnd[iEg]].fX+EPSILON);

					// look for ending horizontal spans
					for (iEg = 0; iEg < iNSpanEnd; iEg++)
						if (iSpanEnd[iEg] == iX)
							cChanged = 1;

					icpCounterV++;
				}

			// if there aren't vertical spans and horizontal spans didn't change...
			// draw current vertical lines (if any)
			if (!cChanged)
			{
				int iXTrans = ( iX-iDist)+iOrX;
				if (!(iXTrans < 0 || iXTrans > SIZE_XM1))
				{
					// vertical rasterization
					for (iEg = 0; iEg < cpVLine[icpCounterV].iEdges; iEg++)
					{
						for (iY = ROUND(vtVertices[cpVLine[icpCounterV].iIdx1[iEg]].fY+EPSILON); 
							iY <= ROUND(vtVertices[cpVLine[icpCounterV].iIdx2[iEg]].fY-EPSILON); 
							iY++)
						{
							// back plane (x' = x) (y' = -y) (z' = z--)
							int iYTrans = (-iY+iDist)+iOrY;
							if (!(iYTrans < 0 || iYTrans > SIZE_YM1))
								H.set(MK_OFF_GRID(iXTrans,iYTrans,-iZ+iOrZ));
						}
					} 
				}

				// look for ending horizontal spans
				for (iEg = 0; iEg < iNSpanEnd; iEg++)
					if (iSpanEnd[iEg] == iX)
						cChanged = 1;
			}
		}

		cChanged = 0;
		// look for ending horizontal spans
		for (iEg = 0; iEg < iNSpanEnd; iEg++)
			if (iSpanEnd[iEg] == iX)
				cChanged = 1;

		iX--;

		// if there aren't vertical spans and horizontal spans didn't change...
		// draw current vertical lines (if any)
		if (!cChanged)
		{
			int iXTrans = ( iX-iDist)+iOrX;
			if (!(iXTrans < 0 || iXTrans > SIZE_XM1))
			{
				// vertical rasterization
				for (iEg = 0; iEg < cpVLine[icpCounterV].iEdges; iEg++)
				{
					for (iY = ROUND(vtVertices[cpVLine[icpCounterV].iIdx1[iEg]].fY+EPSILON); 
						iY <= ROUND(vtVertices[cpVLine[icpCounterV].iIdx2[iEg]].fY-EPSILON); 
						iY++)
					{
						// back plane (x' = x) (y' = -y) (z' = z--)
						int iYTrans = (-iY+iDist)+iOrY;
						if (!(iYTrans < 0 || iYTrans > SIZE_YM1))
							H.set(MK_OFF_GRID(iXTrans,iYTrans,-iZ+iOrZ));
					}
				} 
			}
		}
	}
} // Vis3D::drawConcPolB

/////////////////////////////////////////////////////////////////////////////
// Rasterize the vectorized representation given by ras2Vec
// 'Left' version
/////////////////////////////////////////////////////////////////////////////
void Vis3D::drawConcPolL (void)
{
	int iCt;

	//
	// compute the increment of each vertex and store it for further use
	//
	for (iCt = 0; iCt < iVertices; iCt++)
		vtVertices[iCt].SetInc ((float)(vtVertices[iCt].fX-iDist)/iDist, 
								(float)(vtVertices[iCt].fY-iDist)/iDist);

	iZ = iDist;

	float fCPMinXInc = (float)(fCPMinX-iDist)/iDist,
		  fCPMaxXInc = (float)(fCPMaxX-iDist)/iDist,
		  fCPMinYInc = (float)(fCPMinY-iDist)/iDist,
		  fCPMaxYInc = (float)(fCPMaxY-iDist)/iDist;

	for (int iDepth = 0; iDepth < iMaxDist-iDist; iDepth++)	
	{
		iZ++;

		// return if the current concave polygon slice is out of bounds
		if ((iZ+iOrX) > SIZE_XM1) return;

		//
		// increment all vertices of the raster plane
		//
		for (iCt = 0; iCt < iVertices; iCt++)
			vtVertices[iCt].Increment ();

		// min/max values must also be transformed
		fCPMinX += fCPMinXInc; fCPMaxX += fCPMaxXInc; 
		fCPMinY += fCPMinYInc; fCPMaxY += fCPMaxYInc;

		//
		// rasterize
		//

		int iNSpanEnd = 0;
		int icpCounterH = -1;
		char cChanged = 1;
		int iEg;

		// vertical sweep
		for (iY = ROUND(fCPMinY+EPSILON); iY <= ROUND(fCPMaxY-EPSILON); iY++)
		{
			// increment icpCounterH?
			if (icpCounterH < iHLines-1)
				if (ROUND(vtVertices[cpHLine[icpCounterH+1].iIdx1[0]].fY+EPSILON) <= iY)
				{
					cChanged = 0;

					// add to the span set the vertical lines starting here
					for (iEg = 0; iEg < cpHLineEnd[icpCounterH].iElements; iEg++)
						iSpanEnd[iNSpanEnd++] = ROUND(vtVertices[cpHLineEnd[icpCounterH].iIdxEnd[iEg]].fY+EPSILON);

					// look for ending vertical spans
					for (iEg = 0; iEg < iNSpanEnd; iEg++)
						if (iSpanEnd[iEg] == iY)
							cChanged = 1;

					icpCounterH++;
				}

			// if there aren't horizontal spans and vertical spans didn't change,
			// draw current horizontal lines (if any)
			if (!cChanged)
			{
				int iYTrans = (-iY+iDist)+iOrY;
				if (!(iYTrans < 0 || iYTrans > SIZE_YM1))
				{
					// horizontal rasterization
					for (iEg = 0; iEg < cpHLine[icpCounterH].iEdges; iEg++)
					{
						for (iX = ROUND(vtVertices[cpHLine[icpCounterH].iIdx1[iEg]].fX+EPSILON); 
							 iX <= ROUND(vtVertices[cpHLine[icpCounterH].iIdx2[iEg]].fX-EPSILON); 
							iX++)
						{
							// left plane (x' = z) (y' = -y) (z' = x++)
							int iZTrans = ( iX-iDist)+iOrZ;
							if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
								H.set(MK_OFF_GRID(iZ+iOrX,iYTrans,iZTrans));
						}
					} 
				}

				// look for ending vertical spans
				for (iEg = 0; iEg < iNSpanEnd; iEg++)
					if (iSpanEnd[iEg] == iY)
						cChanged = 1;
			}
		}

		cChanged = 0;
		// look for ending vertical spans
		for (iEg = 0; iEg < iNSpanEnd; iEg++)
			if (iSpanEnd[iEg] == iY)
				cChanged = 1;

		iY--;

		// if there aren't horizontal spans and vertical spans didn't change,
		// draw current horizontal lines (if any)
		if (!cChanged)
		{
			int iYTrans = (-iY+iDist)+iOrY;
			if (!(iYTrans < 0 || iYTrans > SIZE_YM1))
			{
				// horizontal rasterization
				for (iEg = 0; iEg < cpHLine[icpCounterH].iEdges; iEg++)
				{
					for (iX = ROUND(vtVertices[cpHLine[icpCounterH].iIdx1[iEg]].fX+EPSILON); 
						 iX <= ROUND(vtVertices[cpHLine[icpCounterH].iIdx2[iEg]].fX-EPSILON); 
						iX++)
					{
						// left plane (x' = z) (y' = -y) (z' = x++)
						int iZTrans = ( iX-iDist)+iOrZ;
						if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
							H.set(MK_OFF_GRID(iZ+iOrX,iYTrans,iZTrans));
					}
				} 
			}
		}

		iNSpanEnd = 0;
		int icpCounterV = -1;
		cChanged = 1;

		// horizontal sweep
		for (iX = ROUND(fCPMinX+EPSILON); iX <= ROUND(fCPMaxX-EPSILON); iX++)
		{
			// increment icpCounterV?
			if (icpCounterV < iVLines-1)
				if (ROUND(vtVertices[cpVLine[icpCounterV+1].iIdx1[0]].fX+EPSILON) <= iX)
				{
					cChanged = 0;

					// add to the span set the horizontal lines starting here
					for (iEg = 0; iEg < cpVLineEnd[icpCounterV].iElements; iEg++)
						iSpanEnd[iNSpanEnd++] = ROUND(vtVertices[cpVLineEnd[icpCounterV].iIdxEnd[iEg]].fX+EPSILON);

					// look for ending horizontal spans
					for (iEg = 0; iEg < iNSpanEnd; iEg++)
						if (iSpanEnd[iEg] == iX)
							cChanged = 1;

					icpCounterV++;
				}

			// if there aren't vertical spans and horizontal spans didn't change...
			// draw current vertical lines (if any)
			if (!cChanged)
			{
				int iZTrans = ( iX-iDist)+iOrZ;
				if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
				{
					// vertical rasterization
					for (iEg = 0; iEg < cpVLine[icpCounterV].iEdges; iEg++)
					{
						for (iY = ROUND(vtVertices[cpVLine[icpCounterV].iIdx1[iEg]].fY+EPSILON); 
							iY <= ROUND(vtVertices[cpVLine[icpCounterV].iIdx2[iEg]].fY-EPSILON); 
							iY++)
						{
							// left plane (x' = z) (y' = -y) (z' = x++)
							int iYTrans = (-iY+iDist)+iOrY;
							if (!(iYTrans < 0 || iYTrans > SIZE_YM1))
								H.set(MK_OFF_GRID(iZ+iOrX,iYTrans,iZTrans));
						}
					} 
				}

				// look for ending horizontal spans
				for (iEg = 0; iEg < iNSpanEnd; iEg++)
					if (iSpanEnd[iEg] == iX)
						cChanged = 1;
			}
		}

		cChanged = 0;
		// look for ending horizontal spans
		for (iEg = 0; iEg < iNSpanEnd; iEg++)
			if (iSpanEnd[iEg] == iX)
				cChanged = 1;

		iX--;

		// if there aren't vertical spans and horizontal spans didn't change...
		// draw current vertical lines (if any)
		if (!cChanged)
		{
			int iZTrans = ( iX-iDist)+iOrZ;
			if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
			{
				// vertical rasterization
				for (iEg = 0; iEg < cpVLine[icpCounterV].iEdges; iEg++)
				{
					for (iY = ROUND(vtVertices[cpVLine[icpCounterV].iIdx1[iEg]].fY+EPSILON); 
						iY <= ROUND(vtVertices[cpVLine[icpCounterV].iIdx2[iEg]].fY-EPSILON); 
						iY++)
					{
						// left plane (x' = z) (y' = -y) (z' = x++)
						int iYTrans = (-iY+iDist)+iOrY;
						if (!(iYTrans < 0 || iYTrans > SIZE_YM1))
							H.set(MK_OFF_GRID(iZ+iOrX,iYTrans,iZTrans));
					}
				} 
			}
		}
	}
} // Vis3D::drawConcPolL

/////////////////////////////////////////////////////////////////////////////
// Rasterize the vectorized representation given by ras2Vec
/////////////////////////////////////////////////////////////////////////////
void Vis3D::drawConcPolR (void)
{
	int iCt;

	//
	// compute the increment of each vertex and store it for further use
	//
	for (iCt = 0; iCt < iVertices; iCt++)
		vtVertices[iCt].SetInc ((float)(vtVertices[iCt].fX-iDist)/iDist, 
								(float)(vtVertices[iCt].fY-iDist)/iDist);

	iZ = iDist;

	float fCPMinXInc = (float)(fCPMinX-iDist)/iDist,
		  fCPMaxXInc = (float)(fCPMaxX-iDist)/iDist,
		  fCPMinYInc = (float)(fCPMinY-iDist)/iDist,
		  fCPMaxYInc = (float)(fCPMaxY-iDist)/iDist;

	for (int iDepth = 0; iDepth < iMaxDist-iDist; iDepth++)
	{
		iZ++;

		// return if the current concave polygon slice is out of bounds
		if ((-iZ+iOrX) < 0) return;

		//
		// increment all vertices of the raster plane
		//
		for (iCt = 0; iCt < iVertices; iCt++)
			vtVertices[iCt].Increment ();

		// min/max values must also be transformed
		fCPMinX += fCPMinXInc; fCPMaxX += fCPMaxXInc; 
		fCPMinY += fCPMinYInc; fCPMaxY += fCPMaxYInc;

		//
		// rasterize
		//

		int iNSpanEnd = 0;
		int icpCounterH = -1;
		char cChanged = 1;
		int iEg;

		// vertical sweep
		for (iY = ROUND(fCPMinY+EPSILON); iY <= ROUND(fCPMaxY-EPSILON); iY++)
		{
			// increment icpCounterH?
			if (icpCounterH < iHLines-1)
				if (ROUND(vtVertices[cpHLine[icpCounterH+1].iIdx1[0]].fY+EPSILON) <= iY)
				{
					cChanged = 0;

					// add to the span set the vertical lines starting here
					for (iEg = 0; iEg < cpHLineEnd[icpCounterH].iElements; iEg++)
						iSpanEnd[iNSpanEnd++] = ROUND(vtVertices[cpHLineEnd[icpCounterH].iIdxEnd[iEg]].fY+EPSILON);

					// look for ending vertical spans
					for (iEg = 0; iEg < iNSpanEnd; iEg++)
						if (iSpanEnd[iEg] == iY)
							cChanged = 1;

					icpCounterH++;
				}

			// if there aren't horizontal spans and vertical spans didn't change,
			// draw current horizontal lines (if any)
			if (!cChanged)
			{
				int iYTrans = (-iY+iDist)+iOrY;
				if (!(iYTrans < 0 || iYTrans > SIZE_YM1))
				{
					// horizontal rasterization
					for (iEg = 0; iEg < cpHLine[icpCounterH].iEdges; iEg++)
					{
						for (iX = ROUND(vtVertices[cpHLine[icpCounterH].iIdx1[iEg]].fX+EPSILON); 
							 iX <= ROUND(vtVertices[cpHLine[icpCounterH].iIdx2[iEg]].fX-EPSILON); 
							iX++)
						{
							// right plane (x' = -z) (y' = -y) (z' = x--)
							int iZTrans = (-iX+iDist)+iOrZ;
							if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
								H.set(MK_OFF_GRID(-iZ+iOrX,iYTrans,iZTrans));
						}
					} 
				}

				// look for ending vertical spans
				for (iEg = 0; iEg < iNSpanEnd; iEg++)
					if (iSpanEnd[iEg] == iY)
						cChanged = 1;
			}
		}

		cChanged = 0;
		// look for ending vertical spans
		for (iEg = 0; iEg < iNSpanEnd; iEg++)
			if (iSpanEnd[iEg] == iY)
				cChanged = 1;

		iY--;

		// if there aren't horizontal spans and vertical spans didn't change,
		// draw current horizontal lines (if any)
		if (!cChanged)
		{
			int iYTrans = (-iY+iDist)+iOrY;
			if (!(iYTrans < 0 || iYTrans > SIZE_YM1))
			{
				// horizontal rasterization
				for (iEg = 0; iEg < cpHLine[icpCounterH].iEdges; iEg++)
				{
					for (iX = ROUND(vtVertices[cpHLine[icpCounterH].iIdx1[iEg]].fX+EPSILON); 
						 iX <= ROUND(vtVertices[cpHLine[icpCounterH].iIdx2[iEg]].fX-EPSILON); 
						iX++)
					{
						// right plane (x' = -z) (y' = -y) (z' = x--)
						int iZTrans = (-iX+iDist)+iOrZ;
						if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
							H.set(MK_OFF_GRID(-iZ+iOrX,iYTrans,iZTrans));
					}
				} 
			}
		}

		iNSpanEnd = 0;
		int icpCounterV = -1;
		cChanged = 1;

		// horizontal sweep
		for (iX = ROUND(fCPMinX+EPSILON); iX <= ROUND(fCPMaxX-EPSILON); iX++)
		{
			// increment icpCounterV?
			if (icpCounterV < iVLines-1)
				if (ROUND(vtVertices[cpVLine[icpCounterV+1].iIdx1[0]].fX+EPSILON) <= iX)
				{
					cChanged = 0;

					// add to the span set the horizontal lines starting here
					for (iEg = 0; iEg < cpVLineEnd[icpCounterV].iElements; iEg++)
						iSpanEnd[iNSpanEnd++] = ROUND(vtVertices[cpVLineEnd[icpCounterV].iIdxEnd[iEg]].fX+EPSILON);

					// look for ending horizontal spans
					for (iEg = 0; iEg < iNSpanEnd; iEg++)
						if (iSpanEnd[iEg] == iX)
							cChanged = 1;

					icpCounterV++;
				}

			// if there aren't vertical spans and horizontal spans didn't change...
			// draw current vertical lines (if any)
			if (!cChanged)
			{
				int iZTrans = (-iX+iDist)+iOrZ;
				if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
				{
					// vertical rasterization
					for (iEg = 0; iEg < cpVLine[icpCounterV].iEdges; iEg++)
					{
						for (iY = ROUND(vtVertices[cpVLine[icpCounterV].iIdx1[iEg]].fY+EPSILON); 
							iY <= ROUND(vtVertices[cpVLine[icpCounterV].iIdx2[iEg]].fY-EPSILON); 
							iY++)
						{
							// right plane (x' = -z) (y' = -y) (z' = x--)
							int iYTrans = (-iY+iDist)+iOrY;
							if (!(iYTrans < 0 || iYTrans > SIZE_YM1))
								H.set(MK_OFF_GRID(-iZ+iOrX,iYTrans,iZTrans));
						}
					} 
				}

				// look for ending horizontal spans
				for (iEg = 0; iEg < iNSpanEnd; iEg++)
					if (iSpanEnd[iEg] == iX)
						cChanged = 1;
			}
		}

		cChanged = 0;
		// look for ending horizontal spans
		for (iEg = 0; iEg < iNSpanEnd; iEg++)
			if (iSpanEnd[iEg] == iX)
				cChanged = 1;

		iX--;

		// if there aren't vertical spans and horizontal spans didn't change...
		// draw current vertical lines (if any)
		if (!cChanged)
		{
			int iZTrans = (-iX+iDist)+iOrZ;
			if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
			{
				// vertical rasterization
				for (iEg = 0; iEg < cpVLine[icpCounterV].iEdges; iEg++)
				{
					for (iY = ROUND(vtVertices[cpVLine[icpCounterV].iIdx1[iEg]].fY+EPSILON); 
						iY <= ROUND(vtVertices[cpVLine[icpCounterV].iIdx2[iEg]].fY-EPSILON); 
						iY++)
					{
						// right plane (x' = -z) (y' = -y) (z' = x--)
						int iYTrans = (-iY+iDist)+iOrY;
						if (!(iYTrans < 0 || iYTrans > SIZE_YM1))
							H.set(MK_OFF_GRID(-iZ+iOrX,iYTrans,iZTrans));
					}
				} 
			}
		}
	}
} // Vis3D::drawConcPolR

/////////////////////////////////////////////////////////////////////////////
// Rasterize the vectorized representation given by ras2Vec
// 'Top' version
/////////////////////////////////////////////////////////////////////////////
void Vis3D::drawConcPolT (void)
{
	int iCt;

	//
	// compute the increment of each vertex and store it for further use
	//
	for (iCt = 0; iCt < iVertices; iCt++)
		vtVertices[iCt].SetInc ((float)(vtVertices[iCt].fX-iDist)/iDist, 
								(float)(vtVertices[iCt].fY-iDist)/iDist);

	iZ = iDist;

	float fCPMinXInc = (float)(fCPMinX-iDist)/iDist,
		  fCPMaxXInc = (float)(fCPMaxX-iDist)/iDist,
		  fCPMinYInc = (float)(fCPMinY-iDist)/iDist,
		  fCPMaxYInc = (float)(fCPMaxY-iDist)/iDist;

	for (int iDepth = 0; iDepth < iMaxDist-iDist; iDepth++)
	{
		iZ++;

		// return if the current concave polygon slice is out of bounds
		if ((iZ+iOrY) > SIZE_YM1) return;

		//
		// increment all vertices of the raster plane
		//
		for (iCt = 0; iCt < iVertices; iCt++)
			vtVertices[iCt].Increment ();

		// min/max values must also be transformed
		fCPMinX += fCPMinXInc; fCPMaxX += fCPMaxXInc; 
		fCPMinY += fCPMinYInc; fCPMaxY += fCPMaxYInc;

		//
		// rasterize
		//

		int iNSpanEnd = 0;
		int icpCounterH = -1;
		char cChanged = 1;
		int iEg;

		// vertical sweep
		for (iY = ROUND(fCPMinY+EPSILON); iY <= ROUND(fCPMaxY-EPSILON); iY++)
		{
			// increment icpCounterH?
			if (icpCounterH < iHLines-1)
				if (ROUND(vtVertices[cpHLine[icpCounterH+1].iIdx1[0]].fY+EPSILON) <= iY)
				{
					cChanged = 0;

					// add to the span set the vertical lines starting here
					for (iEg = 0; iEg < cpHLineEnd[icpCounterH].iElements; iEg++)
						iSpanEnd[iNSpanEnd++] = ROUND(vtVertices[cpHLineEnd[icpCounterH].iIdxEnd[iEg]].fY+EPSILON);

					// look for ending vertical spans
					for (iEg = 0; iEg < iNSpanEnd; iEg++)
						if (iSpanEnd[iEg] == iY)
							cChanged = 1;

					icpCounterH++;
				}

			// if there aren't horizontal spans and vertical spans didn't change,
			// draw current horizontal lines (if any)
			if (!cChanged)
			{
				int iZTrans = ( iY-iDist)+iOrZ;
				if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
				{
					// horizontal rasterization
					for (iEg = 0; iEg < cpHLine[icpCounterH].iEdges; iEg++)
					{
						for (iX = ROUND(vtVertices[cpHLine[icpCounterH].iIdx1[iEg]].fX+EPSILON); 
							 iX <= ROUND(vtVertices[cpHLine[icpCounterH].iIdx2[iEg]].fX-EPSILON); 
							iX++)
						{
							// top plane (x' = -x) (y' = z) (z' = y++)
							int iXTrans = (-iX+iDist)+iOrX;
							if (!(iXTrans < 0 || iXTrans > SIZE_XM1))
								H.set(MK_OFF_GRID(iXTrans,iZ+iOrY,iZTrans));
						}
					} 
				}

				// look for ending vertical spans
				for (iEg = 0; iEg < iNSpanEnd; iEg++)
					if (iSpanEnd[iEg] == iY)
						cChanged = 1;
			}
		}

		cChanged = 0;
		// look for ending vertical spans
		for (iEg = 0; iEg < iNSpanEnd; iEg++)
			if (iSpanEnd[iEg] == iY)
				cChanged = 1;

		iY--;

		// if there aren't horizontal spans and vertical spans didn't change,
		// draw current horizontal lines (if any)
		if (!cChanged)
		{
			int iZTrans = ( iY-iDist)+iOrZ;
			if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
			{
				// horizontal rasterization
				for (iEg = 0; iEg < cpHLine[icpCounterH].iEdges; iEg++)
				{
					for (iX = ROUND(vtVertices[cpHLine[icpCounterH].iIdx1[iEg]].fX+EPSILON); 
						 iX <= ROUND(vtVertices[cpHLine[icpCounterH].iIdx2[iEg]].fX-EPSILON); 
						iX++)
					{
						// top plane (x' = -x) (y' = z) (z' = y++)
						int iXTrans = (-iX+iDist)+iOrX;
						if (!(iXTrans < 0 || iXTrans > SIZE_XM1))
							H.set(MK_OFF_GRID(iXTrans,iZ+iOrY,iZTrans));
					}
				} 
			}
		}

		iNSpanEnd = 0;
		int icpCounterV = -1;
		cChanged = 1;

		// horizontal sweep
		for (iX = ROUND(fCPMinX+EPSILON); iX <= ROUND(fCPMaxX-EPSILON); iX++)
		{
			// increment icpCounterV?
			if (icpCounterV < iVLines-1)
				if (ROUND(vtVertices[cpVLine[icpCounterV+1].iIdx1[0]].fX+EPSILON) <= iX)
				{
					cChanged = 0;

					// add to the span set the horizontal lines starting here
					for (iEg = 0; iEg < cpVLineEnd[icpCounterV].iElements; iEg++)
						iSpanEnd[iNSpanEnd++] = ROUND(vtVertices[cpVLineEnd[icpCounterV].iIdxEnd[iEg]].fX+EPSILON);

					// look for ending horizontal spans
					for (iEg = 0; iEg < iNSpanEnd; iEg++)
						if (iSpanEnd[iEg] == iX)
							cChanged = 1;

					icpCounterV++;
				}

			// if there aren't vertical spans and horizontal spans didn't change...
			// draw current vertical lines (if any)
			if (!cChanged)
			{
				int iXTrans = (-iX+iDist)+iOrX;
				if (!(iXTrans < 0 || iXTrans > SIZE_XM1))
				{
					// vertical rasterization
					for (iEg = 0; iEg < cpVLine[icpCounterV].iEdges; iEg++)
					{
						for (iY = ROUND(vtVertices[cpVLine[icpCounterV].iIdx1[iEg]].fY+EPSILON); 
							iY <= ROUND(vtVertices[cpVLine[icpCounterV].iIdx2[iEg]].fY-EPSILON); 
							iY++)
						{
							// top plane (x' = -x) (y' = z) (z' = y++)
							int iZTrans = ( iY-iDist)+iOrZ;
							if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
								H.set(MK_OFF_GRID(iXTrans,iZ+iOrY,iZTrans));
						}
					} 
				}

				// look for ending horizontal spans
				for (iEg = 0; iEg < iNSpanEnd; iEg++)
					if (iSpanEnd[iEg] == iX)
						cChanged = 1;
			}
		}

		cChanged = 0;
		// look for ending horizontal spans
		for (iEg = 0; iEg < iNSpanEnd; iEg++)
			if (iSpanEnd[iEg] == iX)
				cChanged = 1;

		iX--;

		// if there aren't vertical spans and horizontal spans didn't change...
		// draw current vertical lines (if any)
		if (!cChanged)
		{
			int iXTrans = (-iX+iDist)+iOrX;
			if (!(iXTrans < 0 || iXTrans > SIZE_XM1))
			{
				// vertical rasterization
				for (iEg = 0; iEg < cpVLine[icpCounterV].iEdges; iEg++)
				{
					for (iY = ROUND(vtVertices[cpVLine[icpCounterV].iIdx1[iEg]].fY+EPSILON); 
						iY <= ROUND(vtVertices[cpVLine[icpCounterV].iIdx2[iEg]].fY-EPSILON); 
						iY++)
					{
						// top plane (x' = -x) (y' = z) (z' = y++)
						int iZTrans = ( iY-iDist)+iOrZ;
						if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
							H.set(MK_OFF_GRID(iXTrans,iZ+iOrY,iZTrans));
					}
				} 
			}
		}
	}
} // Vis3D::drawConcPolT

/////////////////////////////////////////////////////////////////////////////
// Rasterize the vectorized representation given by ras2Vec
// 'Down' version
/////////////////////////////////////////////////////////////////////////////
void Vis3D::drawConcPolD (void)
{
	int iCt;

	//
	// compute the increment of each vertex and store it for further use
	//
	for (iCt = 0; iCt < iVertices; iCt++)
		vtVertices[iCt].SetInc ((float)(vtVertices[iCt].fX-iDist)/iDist, 
								(float)(vtVertices[iCt].fY-iDist)/iDist);

	iZ = iDist;

	float fCPMinXInc = (float)(fCPMinX-iDist)/iDist,
		  fCPMaxXInc = (float)(fCPMaxX-iDist)/iDist,
		  fCPMinYInc = (float)(fCPMinY-iDist)/iDist,
		  fCPMaxYInc = (float)(fCPMaxY-iDist)/iDist;

	for (int iDepth = 0; iDepth < iMaxDist-iDist; iDepth++)
	{
		iZ++;

		// return if the current concave polygon slice is out of bounds
		if ((-iZ+iOrY) < 0) return;

		//
		// increment all vertices of the raster plane
		//
		for (iCt = 0; iCt < iVertices; iCt++)
			vtVertices[iCt].Increment ();

		// min/max values must also be transformed
		fCPMinX += fCPMinXInc; fCPMaxX += fCPMaxXInc; 
		fCPMinY += fCPMinYInc; fCPMaxY += fCPMaxYInc;

		//
		// rasterize
		//

		int iNSpanEnd = 0;
		int icpCounterH = -1;
		char cChanged = 1;
		int iEg;

		// vertical sweep
		for (iY = ROUND(fCPMinY+EPSILON); iY <= ROUND(fCPMaxY-EPSILON); iY++)
		{
			// increment icpCounterH?
			if (icpCounterH < iHLines-1)
				if (ROUND(vtVertices[cpHLine[icpCounterH+1].iIdx1[0]].fY+EPSILON) <= iY)
				{
					cChanged = 0;

					// add to the span set the vertical lines starting here
					for (iEg = 0; iEg < cpHLineEnd[icpCounterH].iElements; iEg++)
						iSpanEnd[iNSpanEnd++] = ROUND(vtVertices[cpHLineEnd[icpCounterH].iIdxEnd[iEg]].fY+EPSILON);

					// look for ending vertical spans
					for (iEg = 0; iEg < iNSpanEnd; iEg++)
						if (iSpanEnd[iEg] == iY)
							cChanged = 1;

					icpCounterH++;
				}

			// if there aren't horizontal spans and vertical spans didn't change,
			// draw current horizontal lines (if any)
			if (!cChanged)
			{
				int iZTrans = (-iY+iDist)+iOrZ;
				if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
				{
					// horizontal rasterization
					for (iEg = 0; iEg < cpHLine[icpCounterH].iEdges; iEg++)
					{
						for (iX = ROUND(vtVertices[cpHLine[icpCounterH].iIdx1[iEg]].fX+EPSILON); 
							 iX <= ROUND(vtVertices[cpHLine[icpCounterH].iIdx2[iEg]].fX-EPSILON); 
							iX++)
						{
							// down plane (x' = -x) (y' = -z) (z' = y--)
							int iXTrans = (-iX+iDist)+iOrX;
							if (!(iXTrans < 0 || iXTrans > SIZE_XM1))
								H.set(MK_OFF_GRID(iXTrans,-iZ+iOrY,iZTrans));
						}
					} 
				}

				// look for ending vertical spans
				for (iEg = 0; iEg < iNSpanEnd; iEg++)
					if (iSpanEnd[iEg] == iY)
						cChanged = 1;
			}
		}

		cChanged = 0;
		// look for ending vertical spans
		for (iEg = 0; iEg < iNSpanEnd; iEg++)
			if (iSpanEnd[iEg] == iY)
				cChanged = 1;

		iY--;

		// if there aren't horizontal spans and vertical spans didn't change,
		// draw current horizontal lines (if any)
		if (!cChanged)
		{
			int iZTrans = (-iY+iDist)+iOrZ;
			if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
			{
				// horizontal rasterization
				for (iEg = 0; iEg < cpHLine[icpCounterH].iEdges; iEg++)
				{
					for (iX = ROUND(vtVertices[cpHLine[icpCounterH].iIdx1[iEg]].fX+EPSILON); 
						 iX <= ROUND(vtVertices[cpHLine[icpCounterH].iIdx2[iEg]].fX-EPSILON); 
						iX++)
					{
						// down plane (x' = -x) (y' = -z) (z' = y--)
						int iXTrans = (-iX+iDist)+iOrX;
						if (!(iXTrans < 0 || iXTrans > SIZE_XM1))
							H.set(MK_OFF_GRID(iXTrans,-iZ+iOrY,iZTrans));
					}
				} 
			}
		}

		iNSpanEnd = 0;
		int icpCounterV = -1;
		cChanged = 1;

		// horizontal sweep
		for (iX = ROUND(fCPMinX+EPSILON); iX <= ROUND(fCPMaxX-EPSILON); iX++)
		{
			// increment icpCounterV?
			if (icpCounterV < iVLines-1)
				if (ROUND(vtVertices[cpVLine[icpCounterV+1].iIdx1[0]].fX+EPSILON) <= iX)
				{
					cChanged = 0;

					// add to the span set the horizontal lines starting here
					for (iEg = 0; iEg < cpVLineEnd[icpCounterV].iElements; iEg++)
						iSpanEnd[iNSpanEnd++] = ROUND(vtVertices[cpVLineEnd[icpCounterV].iIdxEnd[iEg]].fX+EPSILON);

					// look for ending horizontal spans
					for (iEg = 0; iEg < iNSpanEnd; iEg++)
						if (iSpanEnd[iEg] == iX)
							cChanged = 1;

					icpCounterV++;
				}

			// if there aren't vertical spans and horizontal spans didn't change...
			// draw current vertical lines (if any)
			if (!cChanged)
			{
				int iXTrans = (-iX+iDist)+iOrX;
				if (!(iXTrans < 0 || iXTrans > SIZE_XM1))
				{
					// vertical rasterization
					for (iEg = 0; iEg < cpVLine[icpCounterV].iEdges; iEg++)
					{
						for (iY = ROUND(vtVertices[cpVLine[icpCounterV].iIdx1[iEg]].fY+EPSILON); 
							iY <= ROUND(vtVertices[cpVLine[icpCounterV].iIdx2[iEg]].fY-EPSILON); 
							iY++)
						{
							// down plane (x' = -x) (y' = -z) (z' = y--)
							int iZTrans = (-iY+iDist)+iOrZ;
							if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
								H.set(MK_OFF_GRID(iXTrans,-iZ+iOrY,iZTrans));
						}
					} 
				}

				// look for ending horizontal spans
				for (iEg = 0; iEg < iNSpanEnd; iEg++)
					if (iSpanEnd[iEg] == iX)
						cChanged = 1;
			}
		}

		cChanged = 0;
		// look for ending horizontal spans
		for (iEg = 0; iEg < iNSpanEnd; iEg++)
			if (iSpanEnd[iEg] == iX)
				cChanged = 1;

		iX--;

		// if there aren't vertical spans and horizontal spans didn't change...
		// draw current vertical lines (if any)
		if (!cChanged)
		{
			int iXTrans = (-iX+iDist)+iOrX;
			if (!(iXTrans < 0 || iXTrans > SIZE_XM1))
			{
				// vertical rasterization
				for (iEg = 0; iEg < cpVLine[icpCounterV].iEdges; iEg++)
				{
					for (iY = ROUND(vtVertices[cpVLine[icpCounterV].iIdx1[iEg]].fY+EPSILON); 
						iY <= ROUND(vtVertices[cpVLine[icpCounterV].iIdx2[iEg]].fY-EPSILON); 
						iY++)
					{
						// down plane (x' = -x) (y' = -z) (z' = y--)
						int iZTrans = (-iY+iDist)+iOrZ;
						if (!(iZTrans < 0 || iZTrans > SIZE_ZM1))
							H.set(MK_OFF_GRID(iXTrans,-iZ+iOrY,iZTrans));
					}
				} 
			}
		}
	}
} // Vis3D::drawConcPolD

/////////////////////////////////////////////////////////////////////////////
// Test whether cell (x, y, z) is inside the view frustum and structs bounds
// iOFB is true if any component is out of bounds
// iXOFB, iYOFB, iZOFB are flags for individual components
/////////////////////////////////////////////////////////////////////////////
int Vis3D::isInsideVFBnd (float x, float y, float z, VECTOR3 &vBnd, int &iOFB, int &iXOFB, int &iYOFB, int &iZOFB)
{
	iOFB = iXOFB = iYOFB = iZOFB = 0; // out-of-bounds flag

	// test against data structures bounds
	if (x < 0)        { vBnd.x = 0;        iOFB = iXOFB = 1; }
	if (x > SIZE_XM1) { vBnd.x = SIZE_XM1; iOFB = iXOFB = 1; }
	if (y < 0)        { vBnd.y = 0;        iOFB = iYOFB = 1; }
	if (y > SIZE_YM1) { vBnd.y = SIZE_YM1; iOFB = iYOFB = 1; }
	if (z < 0)        { vBnd.z = 0;        iOFB = iZOFB = 1; }
	if (z > SIZE_ZM1) { vBnd.z = SIZE_ZM1; iOFB = iZOFB = 1; }

	vBnd.x -= ROUND(vPos.x);
	vBnd.y -= ROUND(vPos.y);
	vBnd.z -= ROUND(vPos.z);

	// test against frustum
	if ((a_right  * x + b_right  * y + c_right  * z + d_right  < 0) &&
		(a_left   * x + b_left   * y + c_left   * z + d_left   < 0) &&
		(a_bottom * x + b_bottom * y + c_bottom * z + d_bottom < 0) &&
		(a_top    * x + b_top    * y + c_top    * z + d_top    < 0) &&
		(a_near   * x + b_near   * y + c_near   * z + d_near   < 0))
		return 1;

	return 0;
} // Vis3D::isInsideVFBnd

/////////////////////////////////////////////////////////////////////////////
// Calculates the intersection between a ray and a frustum centered at origin
// vOrg is the origin of the ray
// vDir is the direction of the ray
/////////////////////////////////////////////////////////////////////////////
VECTOR3 Vis3D::rayFrustumIntersect (VECTOR3 &vOrg, VECTOR3 &vDir)
{
	VECTOR3 vOrg2 = vOrg + vPos;

	// left clipping plane
	float tl = -((a_left*vOrg2.x)+(b_left*vOrg2.y)+(c_left*vOrg2.z)+d_left)/((a_left*vDir.x)+(b_left*vDir.y)+(c_left*vDir.z));
	if (tl < 0.0) tl = FLT_MAX;
	
	// right clipping plane
	float tr = -((a_right*vOrg2.x)+(b_right*vOrg2.y)+(c_right*vOrg2.z)+d_right)/((a_right*vDir.x)+(b_right*vDir.y)+(c_right*vDir.z));
	if (tr < 0.0) tr = FLT_MAX;
	
	// top clipping plane
	float tt = -((a_top*vOrg2.x)+(b_top*vOrg2.y)+(c_top*vOrg2.z)+d_top)/((a_top*vDir.x)+(b_top*vDir.y)+(c_top*vDir.z));
	if (tt < 0.0) tt = FLT_MAX;
	
	// bottom clipping plane
	float tb = -((a_bottom*vOrg2.x)+(b_bottom*vOrg2.y)+(c_bottom*vOrg2.z)+d_bottom)/((a_bottom*vDir.x)+(b_bottom*vDir.y)+(c_bottom*vDir.z));
	if (tb < 0.0) tb = FLT_MAX;
	
	// near clipping plane
	float tn = -((a_near*vOrg2.x)+(b_near*vOrg2.y)+(c_near*vOrg2.z)+d_near)/((a_near*vDir.x)+(b_near*vDir.y)+(c_near*vDir.z));
	if (tn < 0.0) tn = FLT_MAX;

	float t = MIN(MIN(MIN(MIN(tl, tr), tt), tb), tn);

	// check bounds
	VECTOR3 vRes = (vDir*t)+vOrg;
	if (!isInsideVF (vRes.x+vPos.x, vRes.y+vPos.y, vRes.z+vPos.z))
	{
		vRes = (vDir*t)+vOrg - vDir;
		if (!isInsideVF (vRes.x+vPos.x, vRes.y+vPos.y, vRes.z+vPos.z))
			return vOrg;
	}
	else return vRes;

	return (vDir*t)+vOrg;
} // Vis3D::rayFrustumIntersect

/////////////////////////////////////////////////////////////////////////////
// Calculates the intersection between a ray and a frustum centered at origin
// Also test bounds
// vOrg is the origin of the ray
// vDir is the direction of the ray
/////////////////////////////////////////////////////////////////////////////
VECTOR3 Vis3D::rayFrustumIntersectBnd (VECTOR3 &vOrg, VECTOR3 &vDir, VECTOR3 &vBnd, int &iOFB, int &iXOFB, int &iYOFB, int &iZOFB)
{
	iOFB = iXOFB = iYOFB = iZOFB = 0; // out-of-bounds flag

	VECTOR3 vOrg2 = vOrg + vPos;

	// left clipping plane
	float tl = -((a_left*vOrg2.x)+(b_left*vOrg2.y)+(c_left*vOrg2.z)+d_left)/((a_left*vDir.x)+(b_left*vDir.y)+(c_left*vDir.z));
	if (tl < 0.0) tl = FLT_MAX;
	
	// right clipping plane
	float tr = -((a_right*vOrg2.x)+(b_right*vOrg2.y)+(c_right*vOrg2.z)+d_right)/((a_right*vDir.x)+(b_right*vDir.y)+(c_right*vDir.z));
	if (tr < 0.0) tr = FLT_MAX;
	
	// top clipping plane
	float tt = -((a_top*vOrg2.x)+(b_top*vOrg2.y)+(c_top*vOrg2.z)+d_top)/((a_top*vDir.x)+(b_top*vDir.y)+(c_top*vDir.z));
	if (tt < 0.0) tt = FLT_MAX;
	
	// bottom clipping plane
	float tb = -((a_bottom*vOrg2.x)+(b_bottom*vOrg2.y)+(c_bottom*vOrg2.z)+d_bottom)/((a_bottom*vDir.x)+(b_bottom*vDir.y)+(c_bottom*vDir.z));
	if (tb < 0.0) tb = FLT_MAX;
	
	// near clipping plane
	float tn = -((a_near*vOrg2.x)+(b_near*vOrg2.y)+(c_near*vOrg2.z)+d_near)/((a_near*vDir.x)+(b_near*vDir.y)+(c_near*vDir.z));
	if (tn < 0.0) tn = FLT_MAX;

	float t = MIN(MIN(MIN(MIN(tl, tr), tt), tb), tn);

	// check bounds
	VECTOR3 vRes = (vDir*t)+vOrg;

	if (!isInsideVF (vRes.x+vPos.x, vRes.y+vPos.y, vRes.z+vPos.z))
	{
		VECTOR3 vRes2 = (vDir*t)+vOrg - vDir;

		if (!isInsideVF (vRes2.x+vPos.x, vRes2.y+vPos.y, vRes2.z+vPos.z))
		{
			// test bounds
			if (vOrg.x+vPos.x < 0)        { vBnd.x = 0;        iOFB = iXOFB = 1; }
			if (vOrg.x+vPos.x > SIZE_XM1) { vBnd.x = SIZE_XM1; iOFB = iXOFB = 1; }
			if (vOrg.y+vPos.y < 0)        { vBnd.y = 0;        iOFB = iYOFB = 1; }
			if (vOrg.y+vPos.y > SIZE_YM1) { vBnd.y = SIZE_YM1; iOFB = iYOFB = 1; }
			if (vOrg.z+vPos.z < 0)        { vBnd.z = 0;        iOFB = iZOFB = 1; }
			if (vOrg.z+vPos.z > SIZE_ZM1) { vBnd.z = SIZE_ZM1; iOFB = iZOFB = 1; }

			vBnd -= vPos;

			return vOrg;
		}
	}

	// test bounds
	if (vRes.x+vPos.x < 0)        { vBnd.x = 0;        iOFB = iXOFB = 1; }
	if (vRes.x+vPos.x > SIZE_XM1) { vBnd.x = SIZE_XM1; iOFB = iXOFB = 1; }
	if (vRes.y+vPos.y < 0)        { vBnd.y = 0;        iOFB = iYOFB = 1; }
	if (vRes.y+vPos.y > SIZE_YM1) { vBnd.y = SIZE_YM1; iOFB = iYOFB = 1; }
	if (vRes.z+vPos.z < 0)        { vBnd.z = 0;        iOFB = iZOFB = 1; }
	if (vRes.z+vPos.z > SIZE_ZM1) { vBnd.z = SIZE_ZM1; iOFB = iZOFB = 1; }

	vBnd -= vPos;

	return vRes;
} // Vis3D::rayFrustumIntersectBnd

/////////////////////////////////////////////////////////////////////////////
// Compute the normal vector determined by three points in a plane
/////////////////////////////////////////////////////////////////////////////
VECTOR3 Vis3D::getNormal (VECTOR3 &p1, VECTOR3 &p2, VECTOR3 &p3)
{
	return CrossProduct(p2-p1, p3-p1);
} // Vis3D::getNormal

/////////////////////////////////////////////////////////////////////////////
// Transform view frustum coordinates
/////////////////////////////////////////////////////////////////////////////
void Vis3D::transformVFC (void)
{
	float fFOVRad = (float)RAD(fFOV); // convert degrees to radians

	VECTOR3 corner1, corner2, corner3, corner4;

	corner1.x = (float)-tan(fFOVRad/2.0)*zFar;
	corner1.y = (float)-tan(fFOVRad/2.0)*zFar;
	corner1.z = zFar;

	corner2.x = (float)tan(fFOVRad/2.0)*zFar;
	corner2.y = (float)tan(fFOVRad/2.0)*zFar;
	corner2.z = zFar;

	corner3.x = corner2[0];
	corner3.y = corner1[1];
	corner3.z = zFar;

	corner4.x = corner1[0];
	corner4.y = corner2[1];
	corner4.z = zFar;

	// transformed corner points
	
	//         top
	//      2--------4
	//      |        |
	// left |        | right
	//      |        |
	//      3--------1
	//        bottom

	tCorner1 = TransformVector(&mView, &corner1);
	tCorner2 = TransformVector(&mView, &corner2);
	tCorner3 = TransformVector(&mView, &corner3);
	tCorner4 = TransformVector(&mView, &corner4);

	tCorner1 += vPos; tCorner2 += vPos;	tCorner3 += vPos; tCorner4 += vPos;

	// clipping plane coefficients
	VECTOR3 n_left, n_right, n_top, n_bottom, n_near;

	n_right  = getNormal (tCorner1, tCorner4, vPos);
	n_left   = getNormal (tCorner2, tCorner3, vPos);
	n_bottom = getNormal (tCorner3, tCorner1, vPos);
	n_top    = getNormal (tCorner4, tCorner2, vPos);

	n_right  = Normalize (n_right);
	n_left   = Normalize (n_left);
	n_bottom = Normalize (n_bottom);
	n_top    = Normalize (n_top);

	// translate view frustum planes to guarantee conservative queries
	float scalar;
	float halfdiag = .5;

	scalar = (float)(fabs(n_right.x)+fabs(n_right.y)+fabs(n_right.z))*halfdiag;
	VECTOR3 vTransRight = vPos + (n_right*scalar);
	a_right = n_right.x; b_right = n_right.y; c_right = n_right.z;
	d_right = -(n_right.x*vTransRight.x + n_right.y*vTransRight.y + n_right.z*vTransRight.z);

	scalar = (float)(fabs(n_left.x)+fabs(n_left.y)+fabs(n_left.z))*halfdiag;
	VECTOR3 vTransLeft = vPos + (n_left*scalar);
	a_left = n_left.x; b_left = n_left.y; c_left = n_left.z;
	d_left = -(n_left.x*vTransLeft.x + n_left.y*vTransLeft.y + n_left.z*vTransLeft.z);

	scalar = (float)(fabs(n_top.x)+fabs(n_top.y)+fabs(n_top.z))*halfdiag;
	VECTOR3 vTransTop = vPos + (n_top*scalar);
	a_top = n_top.x; b_top = n_top.y; c_top = n_top.z;
	d_top = -(n_top.x*vTransTop.x + n_top.y*vTransTop.y + n_top.z*vTransTop.z);

	scalar = (float)(fabs(n_bottom.x)+fabs(n_bottom.y)+fabs(n_bottom.z))*halfdiag;
	VECTOR3 vTransBottom = vPos + (n_bottom*scalar);
	a_bottom = n_bottom.x; b_bottom = n_bottom.y; c_bottom = n_bottom.z;
	d_bottom = -(n_bottom.x*vTransBottom.x + n_bottom.y*vTransBottom.y + n_bottom.z*vTransBottom.z);

	// near clipping plane
	n_near = vPos - (tCorner1+tCorner2)/2.0;
	n_near = Normalize (n_near);

	scalar = (float)(fabs(n_near.x)+fabs(n_near.y)+fabs(n_near.z))*halfdiag;
	if (fFOV <= 90.0) scalar = (float)-0.001; else scalar = 0.0;

	VECTOR3 vTransNear = vPos + (n_near*scalar);
	a_near = n_near.x; b_near = n_near.y; c_near = n_near.z;
	d_near = -(n_near.x*vTransNear.x + n_near.y*vTransNear.y + n_near.z*vTransNear.z);
} // Vis3D::transformVFC