//***************************************************************************
//
// MAIN.CPP
// by Harlen C. Batagelo
//
//***************************************************************************

#pragma warning(disable:4786)
#include <set>
#include <list>

#include <math.h>
#include <stdlib.h>

#include "disk.h"
#include "common.h"

// arcball variables
static BOOL bLeftButtonPressed = FALSE,
			bRightButtonPressed = FALSE;
static float fWorldZTranslate;
static int fStartTranslate, fEndTranslate;

void keyboard (unsigned char key, int x, int y);
void reshape (int w, int h);

Vis3D *MyVis; // main application class

typedef void (Vis3D::*MemFntPtr)();
MemFntPtr drawConcPol;

FILE *logFileHnd;

/////////////////////////////////////////////////////////////////////////////
// Constructor
/////////////////////////////////////////////////////////////////////////////
Vis3D::Vis3D() 
{
	vPos = VECTOR3((float)SIZE_X/2.0, (float)SIZE_Y/2.0, (float)SIZE_Z/2.0);

	fFOV = (float)55; // 55o = 35mm camera 
	zFar = (float)sqrt(SIZE_X*SIZE_X + SIZE_Y*SIZE_Y + SIZE_Z*SIZE_Z);

	rotX = 0.0; rotY = 0.0; rotZ = 0.0;
} // Vis3D::Vis3D

/////////////////////////////////////////////////////////////////////////////
// Destructor
/////////////////////////////////////////////////////////////////////////////
Vis3D::~Vis3D()
{
	gluDeleteQuadric (qobj);
	if (disks) delete []disks;
	if (iMatrixF) delete []iMatrixF;
	if (iMatrixB) delete []iMatrixB;
	if (iMatrixL) delete []iMatrixL;
	if (iMatrixR) delete []iMatrixR;
	if (iMatrixT) delete []iMatrixT;
	if (iMatrixD) delete []iMatrixD;
	if (I)   delete []I;
	if (PVS) delete []PVS;
	if (PVSVF) delete []PVSVF;
	if (PVSEXACT) delete []PVSEXACT;
	if (ucReadBufferR) delete []ucReadBufferR;
	if (ucReadBufferG) delete []ucReadBufferG;
	if (ucReadBufferB) delete []ucReadBufferB;

	if (TxyN) delete []TxyN;
	if (TyzN) delete []TyzN;
	if (TxzN) delete []TxzN;
} // Vis3D::~Vis3D

void Vis3D::putIntoGrid (int iCt, bool bStatic, bool bEraseOnly)
{
	// Set O and I
	float fZCt, fYCt, fXCt;

	float fX = disks[iCt].GetX (),
		  fY = disks[iCt].GetY (),
		  fZ = disks[iCt].GetZ ();

	float fMaxRadius = disks[iCt].GetExternalRadius (),
		  fMinRadius = disks[iCt].GetInternalRadius ();

	// remove previous object
	if(bStatic)
		for (int iCt2 = 0; iCt2 < disks[iCt].iICells; iCt2++)
			I[MK_OFF_GRID((int)disks[iCt].vI[iCt2].x, (int)disks[iCt].vI[iCt2].y, (int)disks[iCt].vI[iCt2].z)].erase (iCt);

	if (bEraseOnly) return;

	disks[iCt].iICells = 0;
	for (fZCt = ROUND(fZ-fMaxRadius); fZCt <= ROUND(fZ+fMaxRadius); fZCt++)
	{
		for (fYCt = ROUND(fY-fMaxRadius); fYCt <= ROUND(fY+fMaxRadius); fYCt++)
		{
			for (fXCt = ROUND(fX-fMaxRadius); fXCt <= ROUND(fX+fMaxRadius); fXCt++)
			{
				if (setI(fXCt, fYCt, fZCt, iCt))
					disks[iCt].vI[disks[iCt].iICells++] = VECTOR3(fXCt, fYCt, fZCt);
			}
		}
	}
	if (setI(ROUND(fX), ROUND(fY), ROUND(fZ), iCt))
		disks[iCt].vI[disks[iCt].iICells++] = VECTOR3(ROUND(fX), ROUND(fY), ROUND(fZ));

	if (setI(fX, fY, fZ, iCt))
		disks[iCt].vI[disks[iCt].iICells++] = VECTOR3(fX, fY, fZ);


	if (!bStatic)
	{
		for (fZCt = ROUND(fZ-fMinRadius); fZCt <= ROUND(fZ+fMinRadius); fZCt++)
		{
			for (fYCt = ROUND(fY-fMinRadius); fYCt <= ROUND(fY+fMinRadius); fYCt++)
			{
				for (fXCt = ROUND(fX-fMinRadius); fXCt <= ROUND(fX+fMinRadius); fXCt++)
				{
					float fXCtM5 = (fXCt-fX)-.5,
						fXCtP5 = (fXCt-fX)+.5,
						fYCtM5 = (fYCt-fY)-.5,
						fYCtP5 = (fYCt-fY)+.5,
						fZCtM5 = (fZCt-fZ)-.5,
						fZCtP5 = (fZCt-fZ)+.5;
					
					if (ISINSIDE(fXCtM5,fYCtM5,fZCtP5,fMinRadius) &&
						ISINSIDE(fXCtM5,fYCtP5,fZCtP5,fMinRadius) &&
						ISINSIDE(fXCtP5,fYCtP5,fZCtP5,fMinRadius) &&
						ISINSIDE(fXCtP5,fYCtM5,fZCtP5,fMinRadius) &&
						ISINSIDE(fXCtM5,fYCtM5,fZCtM5,fMinRadius) &&
						ISINSIDE(fXCtM5,fYCtP5,fZCtM5,fMinRadius) &&
						ISINSIDE(fXCtP5,fYCtP5,fZCtM5,fMinRadius) &&
						ISINSIDE(fXCtP5,fYCtM5,fZCtM5,fMinRadius))
						setO(fXCt, fYCt, fZCt);
				}
			}
		}
	}
	else
	{
		for (fZCt = ROUND(fZ-fMinRadius); fZCt <= ROUND(fZ+fMinRadius); fZCt++)
		{
			for (fYCt = ROUND(fY-fMinRadius); fYCt <= ROUND(fY+fMinRadius); fYCt++)
			{
				for (fXCt = ROUND(fX-fMinRadius); fXCt <= ROUND(fX+fMinRadius); fXCt++)
				{
					float fXCtM5 = (fXCt-fX)-.5,
						fXCtP5 = (fXCt-fX)+.5,
						fYCtM5 = (fYCt-fY)-.5,
						fYCtP5 = (fYCt-fY)+.5,
						fZCtM5 = (fZCt-fZ)-.5,
						fZCtP5 = (fZCt-fZ)+.5;
					
					if (ISINSIDE(fXCtM5,fYCtM5,fZCtP5,fMinRadius) &&
						ISINSIDE(fXCtM5,fYCtP5,fZCtP5,fMinRadius) &&
						ISINSIDE(fXCtP5,fYCtP5,fZCtP5,fMinRadius) &&
						ISINSIDE(fXCtP5,fYCtM5,fZCtP5,fMinRadius) &&
						ISINSIDE(fXCtM5,fYCtM5,fZCtM5,fMinRadius) &&
						ISINSIDE(fXCtM5,fYCtP5,fZCtM5,fMinRadius) &&
						ISINSIDE(fXCtP5,fYCtP5,fZCtM5,fMinRadius) &&
						ISINSIDE(fXCtP5,fYCtM5,fZCtM5,fMinRadius))
						setOs(fXCt, fYCt, fZCt);
				}
			}
		}
	}
} // Vis3D::putIntoGrid

/////////////////////////////////////////////////////////////////////////////
// Setup structures and state variables
/////////////////////////////////////////////////////////////////////////////
void Vis3D::init (void) 
{
	fusionEnabled  = true;   // true when occluder fusion is enabled
	hiddenGeometry = false;  // true to disable scene rendering
	bCalcExactVis  = false;  // true to enable exact visibility calculation
	bDynamicScene  = true;   // true to use moving objects

	current_date = fps_avg = fps_avg_ct = 0;

	int iCt;
	bShowData = FALSE;

	glClearColor (0.0, 0.0, 0.0, 0.0);
	glShadeModel (GL_FLAT);

	qobj = gluNewQuadric ();

	// allocate memory for PVS list
	if (!(PVS = new bool[N_OBJS_DYN+N_OBJS_STC]))
	{
		fprintf (stderr, "Not enough memory for PVS.\n");
		exit (-1);
	}
	printf (".");

	// allocate memory for PVSVF list
	if (!(PVSVF = new bool[N_OBJS_DYN+N_OBJS_STC]))
	{
		fprintf (stderr, "Not enough memory for PVSVF.\n");
		exit (-1);
	}
	printf (".");

	// allocate memory for PVSEXACT list
	if (!(PVSEXACT = new bool[N_OBJS_DYN+N_OBJS_STC]))
	{
		fprintf (stderr, "Not enough memory for PVSEXACT.\n");
		exit (-1);
	}
	printf (".");

	// allocate memory for ucReadBufferR
	if (!(ucReadBufferR = new unsigned char[(WIN_WIDTH/2)*WIN_HEIGHT]))
	{
		fprintf (stderr, "Not enough memory for ucReadBufferR.\n");
		exit (-1);
	}
	printf (".");

	// allocate memory for ucReadBufferG
	if (!(ucReadBufferG = new unsigned char[(WIN_WIDTH/2)*WIN_HEIGHT]))
	{
		fprintf (stderr, "Not enough memory for ucReadBufferG.\n");
		exit (-1);
	}
	printf (".");

	// allocate memory for ucReadBufferB
	if (!(ucReadBufferB = new unsigned char[(WIN_WIDTH/2)*WIN_HEIGHT]))
	{
		fprintf (stderr, "Not enough memory for ucReadBufferB.\n");
		exit (-1);
	}
	printf (".");

	// allocate and reset memory for temporary raster images
	iMatrixF = new int[512*512]; 
	iMatrixB = new int[512*512]; 
	iMatrixL = new int[512*512]; 
	iMatrixR = new int[512*512]; 
	iMatrixT = new int[512*512]; 
	iMatrixD = new int[512*512]; 
	memset (iMatrixF, 0, sizeof(int)*512*512);
	memset (iMatrixB, 0, sizeof(int)*512*512);
	memset (iMatrixL, 0, sizeof(int)*512*512);
	memset (iMatrixR, 0, sizeof(int)*512*512);
	memset (iMatrixT, 0, sizeof(int)*512*512);
	memset (iMatrixD, 0, sizeof(int)*512*512);
	printf (".");

	// allocate memory for H
	if(!(H.SetUp(SIZE_X*SIZE_Y*SIZE_Z)))
	{
		fprintf (stderr, "Not enough memory for H.\n");
		exit (-1);
	}
	printf (".");

	// allocate memory for O
	if(!(O.SetUp(SIZE_X*SIZE_Y*SIZE_Z)))
	{
		fprintf (stderr, "Not enough memory for O.\n");
		exit (-1);
	}
	printf (".");

	// allocate memory for Os
	if(!(Os.SetUp(SIZE_X*SIZE_Y*SIZE_Z)))
	{
		fprintf (stderr, "Not enough memory for Os.\n");
		exit (-1);
	}
	printf (".");

	// allocate memory for Txy, Tyz and Txz
	for (iCt = 0; iCt < N_OBJS_DYN; iCt++)
	{
		if(!(Txy[iCt].SetUp(SIZE_X*SIZE_Y)))
		{
			fprintf (stderr, "Not enough memory for Txy[%d].\n", iCt);
			exit (-1);
		}
		if(!(Tyz[iCt].SetUp(SIZE_Y*SIZE_Z)))
		{
			fprintf (stderr, "Not enough memory for Tyz[%d].\n", iCt);
			exit (-1);
		}
		if(!(Txz[iCt].SetUp(SIZE_X*SIZE_Z)))
		{
			fprintf (stderr, "Not enough memory for Txz[%d].\n", iCt);
			exit (-1);
		}
	}
	printf (".");

	// allocate memory for TxyN, TyzN and TxzN
	if (!(TxyN = new set_of_ids[SIZE_X*SIZE_Y]))
	{
		fprintf (stderr, "Not enough memory for TxyN.\n");
		exit (-1);
	}
	for (iCt = 0; iCt < SIZE_X*SIZE_Y; iCt++)
		TxyN[iCt].clear ();

	if (!(TyzN = new set_of_ids[SIZE_Y*SIZE_Z]))
	{
		fprintf (stderr, "Not enough memory for TyzN.\n");
		exit (-1);
	}
	for (iCt = 0; iCt < SIZE_Y*SIZE_Z; iCt++)
		TyzN[iCt].clear ();

	if (!(TxzN = new set_of_ids[SIZE_X*SIZE_Z]))
	{
		fprintf (stderr, "Not enough memory for TxzN.\n");
		exit (-1);
	}
	for (iCt = 0; iCt < SIZE_X*SIZE_Z; iCt++)
		TxzN[iCt].clear ();
	printf (".");

	// allocate memory for I
	if (!(I = new set_of_ids[SIZE_X*SIZE_Y*SIZE_Z]))
	{
		fprintf (stderr, "Not enough memory for I.\n");
		exit (-1);
	}
	// clear I
	for (iCt = 0; iCt < SIZE_X*SIZE_Y*SIZE_Z; iCt++)
		I[iCt].clear ();
	printf (".");

	// initialize disks
	if (!(disks = new CDisk[N_OBJS_DYN + N_OBJS_STC]))
	{
		fprintf (stderr, "Not enough memory to disks.\n");
		exit (-1);
	}
	printf (" done!\n\n");

	fWorldZTranslate = 0.0;

	XMATRIX mX, mY, mZ;

	MatrixIdentity(&mView);
	MatrixMultiply(&mView, MatrixRotationZ(&mZ, RAD(rotZ)), &mView);
	MatrixMultiply(&mView, MatrixRotationX(&mX, RAD(rotX)), &mView);
	MatrixMultiply(&mView, MatrixRotationY(&mY, RAD(rotY)), &mView);

	float fMinRadius;
	float fMaxRadius;

	for (iCt = 0; iCt < N_OBJS_DYN; iCt++)
	{
		fMinRadius = fMaxRadius = ((2.5 + (2.0 / (1 + (rand() % 5)))) * SIZE_X) / 50.0;

		disks[iCt].SetInternalRadius (fMinRadius);
		disks[iCt].SetExternalRadius (fMaxRadius);

		disks[iCt].SetVelocityX (-.05 + (2.0 / (1 + (rand() % 50))));
		disks[iCt].SetVelocityY (-.05 + (2.0 / (1 + (rand() % 50))));
		disks[iCt].SetVelocityZ (-.05 + (2.0 / (1 + (rand() % 50))));

		disks[iCt].SetBounds (SIZE_X, SIZE_Y, SIZE_Z);
		disks[iCt].iICells = 0;

		float fX = 1+(rand()%(SIZE_X-2)), // disk center position
			  fY = 1+(rand()%(SIZE_Y-2)),
			  fZ = 1+(rand()%(SIZE_Z-2));

		fX = (rand()%(SIZE_X*4)) / 4.0;
		fY = (rand()%(SIZE_Y*4)) / 4.0;
		fZ = (rand()%(SIZE_Z*4)) / 4.0;

		fX = SIZE_X / 2.0;
		fY = SIZE_Y / 2.0;
		fZ = SIZE_Z / 2.0;

		float t1 = (float)rand() / RAND_MAX,
			  t2 = (float)rand() / RAND_MAX,
			  t3 = (float)rand() / RAND_MAX;

		fX = t1 * SIZE_X;
		fY = t2 * SIZE_Y;
		fZ = (t3+.125) * SIZE_Z;

		disks[iCt].SetPos (fX, fY, fZ);
		
		disks[iCt].SetRot (rand()%360, rand()%360, rand()%360);
		disks[iCt].SetRotInc (5-rand()%10, 5-rand()%10, 5-rand()%10);

		float fColor = .3+(.7/(1+rand()%10));
		disks[iCt].SetColor (fColor, fColor, fColor);
		disks[iCt].SetItemColor ((iCt >> 16) & 0xff, (iCt >> 8) & 0xff, (iCt) & 0xff);

		// TBV settings
		disks[iCt].SetTBVX (fX); disks[iCt].SetTBVY (fY); disks[iCt].SetTBVZ (fZ);
		disks[iCt].SetStartDate (current_date);
		disks[iCt].SetExpDate (current_date + 5 + (rand() % 40));
		disks[iCt].SetLastX (fX); disks[iCt].SetLastY (fY); disks[iCt].SetLastZ (fZ);
		disks[iCt].SetTBVState (false); // TBV not assigned

#ifndef USETBV
		putIntoGrid (iCt, true, false);
#endif

		PVS[iCt] = true; // everything's initially visible
		PVSVF[iCt] = true;
	}

	for (iCt = N_OBJS_DYN; iCt < N_OBJS_STC+N_OBJS_DYN; iCt++)
	{
		fMaxRadius = fMinRadius = ((0.5 + (4.5 / (1 + (rand() % 50)))) * SIZE_X) / 50.0;

		disks[iCt].SetInternalRadius (fMinRadius);
		disks[iCt].SetExternalRadius (fMaxRadius);

		int type = rand () % 10;

		if (type > 5)
			disks[iCt].SetType (0);

		disks[iCt].SetVelocityX (0.0);
		disks[iCt].SetVelocityY (0.0);
		disks[iCt].SetVelocityZ (0.0);

		disks[iCt].SetBounds (SIZE_X, SIZE_Y, SIZE_Z);
		disks[iCt].iICells = 0;

		float fX = 1+(rand()%(SIZE_X-2)), // disk center position
			  fY = 1+(rand()%(SIZE_Y-2)),
			  fZ = 1+(rand()%(SIZE_Z-2));

		fX = (rand()%(SIZE_X*4)) / 4.0;
		fY = (rand()%(SIZE_Y*4)) / 4.0;
		fZ = (rand()%(SIZE_Z*4)) / 4.0;

		fX = SIZE_X / 2.0;
		fY = SIZE_Y / 2.0;
		fZ = SIZE_Z / 2.0;

		float t1 = (float)rand() / RAND_MAX,
			  t2 = (float)rand() / RAND_MAX,
			  t3 = (float)rand() / RAND_MAX;

		fX = t1 * SIZE_X;
		fY = t2 * SIZE_Y;
//		fZ = t3 * SIZE_Z;
		fZ = (t3+.125) * SIZE_Z;

		disks[iCt].SetPos (fX, fY, fZ);

		disks[iCt].SetRot (rand()%360, rand()%360, rand()%360);
		disks[iCt].SetRotInc (5-rand()%10, 5-rand()%10, 5-rand()%10);

		float fColor = .3+(.7/(1+rand()%10));
		disks[iCt].SetColor (fColor, fColor, fColor);
		disks[iCt].SetItemColor ((iCt >> 16) & 0xff, (iCt >> 8) & 0xff, (iCt) & 0xff);

		// TBV settings
		disks[iCt].SetTBVX (fX); disks[iCt].SetTBVY (fY); disks[iCt].SetTBVZ (fZ);
		disks[iCt].SetStartDate (current_date);
		disks[iCt].SetExpDate (current_date + 5 + (rand() % 40));
		disks[iCt].SetLastX (fX); disks[iCt].SetLastY (fY); disks[iCt].SetLastZ (fZ);
		disks[iCt].SetTBVState (false); // TBV not assigned

		// Set O and I
		putIntoGrid (iCt, true, false);

		PVS[iCt] = true; // everything's initially visible
		PVSVF[iCt] = true;
	}

	// grid lines' display list
	glNewList (STRUCTURE_DISPLAY_LIST, GL_COMPILE);
		glLineStipple (2, 0xF0F0); // grid line stipple
		glEnable (GL_LINE_STIPPLE);

		glBegin (GL_LINES);
			for (int cz = 0; cz <= SIZE_Z; cz++)
			{
				for (int cx = 0; cx <= SIZE_X; cx++)
				{
					glVertex3f (cx*STEP, 0, cz*STEP); 
					glVertex3f (cx*STEP, SIZE_Y*STEP, cz*STEP);
				}

				for (int cy = 0; cy <= SIZE_Y; cy++)
				{
					glVertex3f (0, cy*STEP, cz*STEP); 
					glVertex3f (SIZE_X*STEP, cy*STEP, cz*STEP);
				}
			}

			for (int cy = 0; cy <= SIZE_Y; cy++)
			{
				for (int cx = 0; cx <= SIZE_X; cx++)
				{
					glVertex3f (cx*STEP, cy*STEP, 0); 
					glVertex3f (cx*STEP, cy*STEP, SIZE_Z*STEP);
				}
			}
		glEnd ();
		
		glDisable (GL_LINE_STIPPLE);
	glEndList();

    arcBall.SetWindow (WIN_WIDTH, WIN_HEIGHT, 0.85f); // setup arcball parameters

	bMatrixF = bMatrixB = bMatrixL = bMatrixR = bMatrixT = bMatrixD = 0; // reset raster flags

	iMaxDist = MAX(SIZE_X,MAX(SIZE_Y,SIZE_Z)); // maximum depth of a shadow volume

	// clear grid
	H.reset ();
	O.reset ();
	O.getFromBitset (Os);

	iDist = 0;
	
	resetFPS ();
} // Vis3D::init

/////////////////////////////////////////////////////////////////////////////
// Start view frustum traversal from a given seed cell
/////////////////////////////////////////////////////////////////////////////
void Vis3D::startTraversal (int iX, int iY, int iZ)
{
	// min/max initial values
	iCPMinXF = iCPMinXB = iCPMinXL = iCPMinXR = iCPMinXT = iCPMinXD = 512;
	iCPMinYF = iCPMinYB = iCPMinYL = iCPMinYR = iCPMinYT = iCPMinYD = 512;
	iCPMaxXF = iCPMaxXB = iCPMaxXL = iCPMaxXR = iCPMaxXT = iCPMaxXD = 0;
	iCPMaxYF = iCPMaxYB = iCPMaxYL = iCPMaxYR = iCPMaxYT = iCPMaxYD = 0;

	iDist = MAX(MAX(abs(iX),abs(iY)),abs(iZ)); // current chess distance

	bPrematureEnd = true;

	if (!iDist) bPrematureEnd = false;

	iDistSize = (iDist<<1)+1+2;

	{
		int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
		int iSeedOFB; // flag that indicates whether a seed cell is fully outside data structures bounds
		VECTOR3 vOFB; // out-of-bounds vector
		
		BOOL bTurnNeg, bTurnPos; // make a 90deg turn during the traversal? 
		int iInitX, iInitY, iInitZ; // seed cell
		
		bTurnNeg = bTurnPos = FALSE;
		iInitX = iX; iInitY = iY; iInitZ = iZ;
		
		int tmpX = iX, tmpY = iY, tmpZ = iZ;
		
		int iOFBNeg, iOFBPos;
		
		//
		//  X ^- - - - - - - - - - -|
		//    | \                 / 
		//    |   \     here    /   |
		//    |     \         /     
		//    |       \     /       |
		//    |         \ /         
		//    |         / \         |
		//    |       /     \       
		//    |     /         \     |
		//    |   /     here    \   
		//    | /                 \ | 
		//    *---------------------> Z
		//  any Y
		// 
		
		if (abs(iInitX) > abs(iInitY) && abs(iInitX) > abs(iInitZ))
		{
			int iStopZNeg, iStopZPos; // stop points
			int iOFBStopZNeg, iOFBStopZPos; // ofb stop points
			
			// test bounds for seed cell
			iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
			
			// 1st directions +z -z
			// 2nd directions +y -y
			
			// determine Z extents
			// -z
			if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, -abs(iInitX)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
			{
				iStopZNeg = -abs(iInitX);
				bTurnNeg = TRUE;
			}
			else
			{
				// determine stop point
				VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, 0, -1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
				
				if (iInitZ+iOrZ > SIZE_ZM1 && vIntersect.z+iOrZ > SIZE_ZM1) return;
				
				iStopZNeg = vIntersect.z;
				bTurnNeg = FALSE;
			}
			iOFBNeg = iZOFB;
			iOFBStopZNeg = vOFB.z;
			
			// +z
			if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, abs(iInitX)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
			{
				iStopZPos = abs(iInitX);
				bTurnPos = TRUE;
			}
			else
			{
				// determine stop point
				VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, 0, +1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
				
				if (iInitZ+iOrZ < 0 && vIntersect.z+iOrZ < 0) return;
				
				iStopZPos = vIntersect.z;
				bTurnPos = FALSE;
			}
			iOFBPos = iZOFB;
			iOFBStopZPos = vOFB.z;
			
			// added chunk
			// put seed cell within Z bounds
			if (iSeedOFB)
			{
				if (iInitZ+iOrZ > SIZE_ZM1) iInitZ = SIZE_ZM1-iOrZ;
				if (iInitZ+iOrZ < 0)		iInitZ = -iOrZ;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
			}
			
			if (iInitX > 0)
			{
				if (!bTurnNeg)
				{
					if (!iOFBNeg)
					{
						for (iZ = iInitZ-1; iZ >= iStopZNeg; iZ--)
							Y1TraversalNeg (iInitX, iInitY, iZ, iXOFB);
					}
					else
					{
						for (iZ = iInitZ-1; iZ >= iOFBStopZNeg; iZ--)
							Y1TraversalNeg (iInitX, iInitY, iZ, iXOFB);
					}
				}
				else
				{
					if (!iOFBNeg)
					{
						for (iZ = iInitZ-1; iZ > iStopZNeg; iZ--)
							Y1TraversalNeg (iInitX, iInitY, iZ, iXOFB);
						
						if (iZ+iOrZ >= 0 && iZ+iOrZ < SIZE_Z)
							YXTraversalNeg (iInitX, iInitY, iZ, iXOFB, iYOFB);
					}
					else
					{
						for (iZ = iInitZ; iZ >= iOFBStopZNeg; iZ--)
							Y1TraversalNeg (iInitX, iInitY, iZ, iXOFB);
					}
				}
				
				if (!bTurnPos)
				{	
					if (!iOFBPos)
					{
						for (iZ = iInitZ; iZ <= iStopZPos; iZ++)
							Y1TraversalNeg (iInitX, iInitY, iZ, iXOFB);
					}
					else
					{
						for (iZ = iInitZ; iZ <= iOFBStopZPos; iZ++)
							Y1TraversalNeg (iInitX, iInitY, iZ, iXOFB);
					}
				}
				else
				{
					if (!iOFBPos)
					{
						for (iZ = iInitZ; iZ < iStopZPos; iZ++)
							Y1TraversalNeg (iInitX, iInitY, iZ, iXOFB);
						
						if (iZ+iOrZ >= 0 && iZ+iOrZ < SIZE_Z)
							YXTraversalNeg (iInitX, iInitY, iZ, iXOFB, iYOFB);
					}
					else
					{
						for (iZ = iInitZ; iZ <= iOFBStopZPos; iZ++)
							Y1TraversalNeg (iInitX, iInitY, iZ, iXOFB);
					}
				}
			}
			else
			{
				if (!bTurnNeg)
				{
					if (!iOFBNeg)
					{
						for (iZ = iInitZ-1; iZ >= iStopZNeg; iZ--)
							Y1TraversalPos (iInitX, iInitY, iZ, iXOFB);
					}
					else
					{
						for (iZ = iInitZ-1; iZ >= iOFBStopZNeg; iZ--)
							Y1TraversalPos (iInitX, iInitY, iZ, iXOFB);
					}
				}
				else
				{
					if (!iOFBNeg)
					{
						for (iZ = iInitZ-1; iZ > iStopZNeg; iZ--)
							Y1TraversalPos (iInitX, iInitY, iZ, iXOFB);
						
						if (iZ+iOrZ >= 0 && iZ+iOrZ < SIZE_Z)
							YXTraversalPos (iInitX, iInitY, iZ, iXOFB, iYOFB);
					}
					else
					{
						for (iZ = iInitZ; iZ >= iOFBStopZNeg; iZ--)
							Y1TraversalPos (iInitX, iInitY, iZ, iXOFB);
					}
				}
				
				if (!bTurnPos)
				{	
					if (!iOFBPos)
					{
						for (iZ = iInitZ; iZ <= iStopZPos; iZ++)
							Y1TraversalPos (iInitX, iInitY, iZ, iXOFB);
					}
					else
					{
						for (iZ = iInitZ; iZ <= iOFBStopZPos; iZ++)
							Y1TraversalPos (iInitX, iInitY, iZ, iXOFB);
					}
				}
				else
				{
					if (!iOFBPos)
					{
						for (iZ = iInitZ; iZ < iStopZPos; iZ++)
							Y1TraversalPos (iInitX, iInitY, iZ, iXOFB);
						
						if (iZ+iOrZ >= 0 && iZ+iOrZ < SIZE_Z)
							YXTraversalPos (iInitX, iInitY, iZ, iXOFB, iYOFB);
					}
					else
					{
						for (iZ = iInitZ; iZ <= iOFBStopZPos; iZ++)
							Y1TraversalPos (iInitX, iInitY, iZ, iXOFB);
					}
				}
			}
		}
		else
		//
		//  Y ^- - - - - - - - - - -|
		//    | \                 / 
		//    |   \     here    /   |
		//    |     \         /     
		//    |       \     /       |
		//    |         \ /         
		//    |         / \         |
		//    |       /     \       
		//    |     /         \     |
		//    |   /     here    \   
		//    | /                 \ | 
		//    *---------------------> Z
		//  any X
		// 
		if (abs(iInitY) > abs(iInitX) && abs(iInitY) > abs(iInitZ))
		{
			int iStopZNeg, iStopZPos; // stop points
			int iOFBStopZNeg, iOFBStopZPos; // ofb stop points
			
			// test bounds for seed cell
			iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
				
			// 1st directions +z -z
			// 2nd directions +x -x
				
			// determine Z extents
			// -z
			if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, -abs(iInitY)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
			{
				iStopZNeg = -abs(iInitY);
				bTurnNeg = TRUE;
			}
			else
			{
				// determine stop point
				VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, 0, -1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
					
				if (iInitZ+iOrZ > SIZE_ZM1 && vIntersect.z+iOrZ > SIZE_ZM1) return;
					
				iStopZNeg = vIntersect.z;
				bTurnNeg = FALSE;
			}
			iOFBNeg = iZOFB;
			iOFBStopZNeg = vOFB.z;
				
			// +z
			if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, abs(iInitY)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
			{
				iStopZPos = abs(iInitY);
				bTurnPos = TRUE;
			}
			else
			{
				// determine stop point
				VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, 0, +1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
				
				if (iInitZ+iOrZ < 0 && vIntersect.z+iOrZ < 0) return;
					
				iStopZPos = vIntersect.z;
				bTurnPos = FALSE;
			}
			iOFBPos = iZOFB;
			iOFBStopZPos = vOFB.z;
				
			// added chunk
			// put seed cell within Z bounds
			if (iSeedOFB)
			{
				if (iInitZ+iOrZ > SIZE_ZM1) iInitZ = SIZE_ZM1-iOrZ;
				if (iInitZ+iOrZ < 0) iInitZ = -iOrZ;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
			}
				
			if (iInitY > 0)
			{
				if (!bTurnNeg)
				{
					if (!iOFBNeg)
					{
						for (iZ = iInitZ-1; iZ >= iStopZNeg; iZ--)
							XTraversalNeg (iInitX, iInitY, iZ, iYOFB);
					}
					else
					{
						for (iZ = iInitZ-1; iZ >= iOFBStopZNeg; iZ--)
							XTraversalNeg (iInitX, iInitY, iZ, iYOFB);
					}
				}
				else
				{
					if (!iOFBNeg)
					{
						for (iZ = iInitZ-1; iZ > iStopZNeg; iZ--)
							XTraversalNeg (iInitX, iInitY, iZ, iYOFB);
							
						if (iZ+iOrZ >= 0 && iZ+iOrZ < SIZE_Z)
							XYTraversalNeg (iInitX, iInitY, iZ, iXOFB, iYOFB);
					}
					else
					{
						for (iZ = iInitZ; iZ >= iOFBStopZNeg; iZ--)
							XTraversalNeg (iInitX, iInitY, iZ, iYOFB);
					}
				}
				
				if (!bTurnPos)
				{	
					if (!iOFBPos)
					{
						for (iZ = iInitZ; iZ <= iStopZPos; iZ++)
							XTraversalNeg (iInitX, iInitY, iZ, iYOFB);
					}
					else
					{
						for (iZ = iInitZ; iZ <= iOFBStopZPos; iZ++)
							XTraversalNeg (iInitX, iInitY, iZ, iYOFB);
					}
				}
				else
				{
					if (!iOFBPos)
					{
						for (iZ = iInitZ; iZ < iStopZPos; iZ++)
							XTraversalNeg (iInitX, iInitY, iZ, iYOFB);
							
						if (iZ+iOrZ >=0 && iZ+iOrZ < SIZE_Z)
							XYTraversalNeg (iInitX, iInitY, iZ, iXOFB, iYOFB);
					}
					else
					{
						for (iZ = iInitZ; iZ <= iOFBStopZPos; iZ++)
							XTraversalNeg (iInitX, iInitY, iZ, iYOFB);
					}
				}
			}
			else
			{
				if (!bTurnNeg)
				{
					if (!iOFBNeg)
					{
						for (iZ = iInitZ-1; iZ >= iStopZNeg; iZ--)
							XTraversalPos (iInitX, iInitY, iZ, iYOFB);
					}
					else
					{
						for (iZ = iInitZ-1; iZ >= iOFBStopZNeg; iZ--)
							XTraversalPos (iInitX, iInitY, iZ, iYOFB);
					}
				}
				else
				{
					if (!iOFBNeg)
					{
						for (iZ = iInitZ-1; iZ > iStopZNeg; iZ--)
							XTraversalPos (iInitX, iInitY, iZ, iYOFB);
							
						if (iZ+iOrZ >= 0 && iZ+iOrZ < SIZE_Z)
							XYTraversalPos (iInitX, iInitY, iZ, iXOFB, iYOFB);
					}
					else
					{
						for (iZ = iInitZ; iZ >= iOFBStopZNeg; iZ--)
							XTraversalPos (iInitX, iInitY, iZ, iYOFB);
					}
				}
					
				if (!bTurnPos)
				{	
					if (!iOFBPos)
					{
						for (iZ = iInitZ; iZ <= iStopZPos; iZ++)
							XTraversalPos (iInitX, iInitY, iZ, iYOFB);
					}
					else
					{
						for (iZ = iInitZ; iZ <= iOFBStopZPos; iZ++)
							XTraversalPos (iInitX, iInitY, iZ, iYOFB);
					}
				}
				else
				{
					if (!iOFBPos)
					{
						for (iZ = iInitZ; iZ < iStopZPos; iZ++)
							XTraversalPos (iInitX, iInitY, iZ, iYOFB);
						
						if (iZ+iOrZ >= 0 && iZ+iOrZ < SIZE_Z)
							XYTraversalPos (iInitX, iInitY, iZ, iXOFB, iYOFB);
					}
					else
					{
						for (iZ = iInitZ; iZ <= iOFBStopZPos; iZ++)
							XTraversalPos (iInitX, iInitY, iZ, iYOFB);
					}
				}
			}
		}
		else
		//
		//  X ^- - - - - - - - - - -|
		//	  | \                 / 
		//    |   \             /   |
		//    |     \         /     
		//    |       \     /       |
		//    |  here   \ /   here 
		//    |         / \         |
		//    |       /     \       
		//    |     /         \     |
		//    |   /             \   
		//    | /                 \ | 
		//    *---------------------> Z
		//  any Y
		// 
		if (abs(iInitZ) > abs(iInitX) && abs(iInitZ) > abs(iInitY))
		{
			int iStopXNeg, iStopXPos; // stop points
			int iOFBStopXNeg, iOFBStopXPos; // ofb stop points
			
			// test bounds for seed cell
			iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
			
			// 1st directions +x -x
			// 2nd directions +y -y
			
			// determine X extents
			// -x
			if (isInsideVFBnd (-abs(iInitZ)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
			{
				iStopXNeg = -abs(iInitZ);
				bTurnNeg = TRUE;
			}
			else
			{
				// determine stop point
				VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
				
				if (iInitX+iOrX > SIZE_XM1 && vIntersect.x+iOrX > SIZE_XM1) return;
				
				iStopXNeg = vIntersect.x;
				bTurnNeg = FALSE;
			}
			iOFBNeg = iXOFB;
			iOFBStopXNeg = vOFB.x;
			
			// +x
			if (isInsideVFBnd (abs(iInitZ)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
			{
				iStopXPos = abs(iInitZ);
				bTurnPos = TRUE;
			}
			else
			{
				// determine stop point
				VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
				
				if (iInitX+iOrX < 0 && vIntersect.x+iOrX < 0) return;
				
				iStopXPos = vIntersect.x;
				bTurnPos = FALSE;
			}
			iOFBPos = iXOFB;
			iOFBStopXPos = vOFB.x;
			
			// added chunk
			// put seed cell within X bounds
			if (iSeedOFB)
			{
				if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
				if (iInitX+iOrX < 0)		iInitX = -iOrX;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
			}
			
			if (iInitZ > 0)
			{
				if (!bTurnNeg)
				{
					if (!iOFBNeg)
					{
						for (iX = iInitX-1; iX >= iStopXNeg; iX--)
							Y2TraversalNeg (iX, iInitY, iInitZ, iZOFB);
					}
					else
					{
						for (iX = iInitX-1; iX >= iOFBStopXNeg; iX--)
							Y2TraversalNeg (iX, iInitY, iInitZ, iZOFB);
					}
				}
				else
				{
					if (!iOFBNeg)
					{
						for (iX = iInitX-1; iX > iStopXNeg; iX--)
							Y2TraversalNeg (iX, iInitY, iInitZ, iZOFB);
						
						if (iX+iOrX >= 0 && iX+iOrX < SIZE_X)
							YZTraversalNeg (iX, iInitY, iInitZ, iYOFB, iZOFB);
					}
					else
					{
						for (iX = iInitX; iX >= iOFBStopXNeg; iX--)
							Y2TraversalNeg (iX, iInitY, iInitZ, iZOFB);
					}
				}
				
				if (!bTurnPos)
				{	
					if (!iOFBPos)
					{
						for (iX = iInitX; iX <= iStopXPos; iX++)
							Y2TraversalNeg (iX, iInitY, iInitZ, iZOFB);
					}
					else
					{
						for (iX = iInitX; iX <= iOFBStopXPos; iX++)
							Y2TraversalNeg (iX, iInitY, iInitZ, iZOFB);
					}
				}
				else
				{
					if (!iOFBPos)
					{
						for (iX = iInitX; iX < iStopXPos; iX++)
							Y2TraversalNeg (iX, iInitY, iInitZ, iZOFB);
						
						if (iX+iOrX >= 0 && iX+iOrX < SIZE_X)
							YZTraversalNeg (iX, iInitY, iInitZ, iYOFB, iZOFB);
					}
					else
					{
						for (iX = iInitX; iX <= iOFBStopXPos; iX++)
							Y2TraversalNeg (iX, iInitY, iInitZ, iZOFB);
					}
				}
			}
			else
			{
				if (!bTurnNeg)
				{
					if (!iOFBNeg)
					{
						for (iX = iInitX-1; iX >= iStopXNeg; iX--)
							Y2TraversalPos (iX, iInitY, iInitZ, iZOFB);
					}
					else
					{
						for (iX = iInitX-1; iX >= iOFBStopXNeg; iX--)
							Y2TraversalPos (iX, iInitY, iInitZ, iZOFB);
					}
				}
				else
				{
					if (!iOFBNeg)
					{
						for (iX = iInitX-1; iX > iStopXNeg; iX--)
							Y2TraversalPos (iX, iInitY, iInitZ, iZOFB);
						
						if (iX+iOrX >= 0 && iX+iOrX < SIZE_X)					
							YZTraversalPos (iX, iInitY, iInitZ, iYOFB, iZOFB);
					}
					else
					{
						for (iX = iInitX; iX >= iOFBStopXNeg; iX--)
							Y2TraversalPos (iX, iInitY, iInitZ, iZOFB);
					}
				}
				
				if (!bTurnPos)
				{	
					if (!iOFBPos)
					{
						for (iX = iInitX; iX <= iStopXPos; iX++)
							Y2TraversalPos (iX, iInitY, iInitZ, iZOFB);
					}
					else
					{
						for (iX = iInitX; iX <= iOFBStopXPos; iX++)
							Y2TraversalPos (iX, iInitY, iInitZ, iZOFB);
					}
				}
				else
				{
					if (!iOFBPos)
					{
						for (iX = iInitX; iX < iStopXPos; iX++)
							Y2TraversalPos (iX, iInitY, iInitZ, iZOFB);
						
						if (iX+iOrX >= 0 && iX+iOrX < SIZE_X)
							YZTraversalPos (iX, iInitY, iInitZ, iYOFB, iZOFB);
					}
					else
					{
						for (iX = iInitX; iX <= iOFBStopXPos; iX++)
							Y2TraversalPos (iX, iInitY, iInitZ, iZOFB);
					}
				}
			}
		}
		else
		//
		//	X ^- - - - - - - - - - -|
		//	  | \				  * 
		//	  |   \ 			*	|
		//	  | 	\		  * 	
		//	  | 	  \ 	*		|
		//	  | 		\ * 		
		//	  | 		* \ 		|
		//	  | 	  * 	\		
		//	  | 	*		  \ 	|
		//	  |   * 			\	
		//	  | *				  \ | 
		//	  *---------------------> Z
		//	any Y
		// 
		if (iInitX && iInitZ && (iInitX == iInitZ))
		{
			int iStopZ; // stop points
			int iFirstSweep = 0;
			
			BOOL bTurnZ, bTurnX; // make a 90deg turn during the traversal? 
			bTurnZ = bTurnX = FALSE;
			
			// test bounds for seed cell
			iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
			
			if (iInitX > 0)
			{
				// 1st directions -z -x
				// 2nd directions +y -y
				
				// determine Z extents
				// -z
				if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, -abs(iInitZ)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
				{
					iStopZ = -abs(iInitZ);
					bTurnZ = TRUE;
				}
				else
				{
					// determine stop point
					VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, 0, -1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
					
					if (iInitZ+iOrZ > SIZE_ZM1 && vIntersect.z+iOrZ > SIZE_ZM1) return;
					
					iStopZ = vIntersect.z;
					bTurnZ = FALSE;
				}
				if (iZOFB)
					iStopZ = vOFB.z;
				
				// put seed cell within Z bounds
				if (iInitZ+iOrZ > SIZE_ZM1) { iInitZ = SIZE_ZM1-iOrZ; iFirstSweep = 1; }
				if (iInitZ+iOrZ < 0)		{ iInitZ = -iOrZ; iFirstSweep = 1; }
				
				int iCpyInitY = iInitY; // get a copy of original Y
				if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
				if (iInitY+iOrY < 0)		iInitY = -iOrY;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
				
				if (!bTurnZ)
				{
					for (iZ = iInitZ-1; iZ >= iStopZ; iZ--)
						Y1TraversalNegDg (iInitX, iCpyInitY, iZ, iSeedOFB);
				}
				else
				{
					for (iZ = iInitZ-1; iZ > iStopZ; iZ--)
						Y1TraversalNegDg (iInitX, iCpyInitY, iZ, iSeedOFB);
					
					if (!iZOFB)
						YXTraversalNegDg (iInitX, iInitY, iZ, iSeedOFB);
				}
				
				if (!iFirstSweep)
					YXTraversalNegDg (iInitX, iInitY, iInitZ, iSeedOFB); 
				else
					Y1TraversalNegDg (iInitX, iCpyInitY, iInitZ, iSeedOFB);
			}
			else
			{
				// 1st directions +z +x
				// 2nd directions +y -y
				
				// determine Z extents
				// +z
				if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, abs(iInitZ)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
				{
					iStopZ = abs(iInitZ);
					bTurnZ = TRUE;
				}
				else
				{
					// determine stop point
					VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, 0, +1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
					
					if (iInitZ+iOrZ < 0 && vIntersect.z+iOrZ < 0) return;
					
					iStopZ = vIntersect.z;
					bTurnZ = FALSE;
				}
				
				if (iZOFB)
					iStopZ = vOFB.z;
				
				// put seed cell within Z bounds
				if (iInitZ+iOrZ > SIZE_ZM1) { iInitZ = SIZE_ZM1-iOrZ; iFirstSweep = 1; }
				if (iInitZ+iOrZ < 0)		{ iInitZ = -iOrZ; iFirstSweep = 1; }
				
				int iCpyInitY = iInitY; // get a copy of original Y
				if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
				if (iInitY+iOrY < 0)		iInitY = -iOrY;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
				
				if (!bTurnZ)
				{
					for (iZ = iInitZ+1; iZ <= iStopZ; iZ++)
						Y1TraversalPosDg (iInitX, iCpyInitY, iZ, iSeedOFB);
				}
				else
				{
					for (iZ = iInitZ+1; iZ < iStopZ; iZ++)
						Y1TraversalPosDg (iInitX, iCpyInitY, iZ, iSeedOFB);
					
					if (!iZOFB)
						YXTraversalPosDg (iInitX, iInitY, iZ, iSeedOFB);
				}
				
				if (!iFirstSweep)
					YXTraversalPosDg (iInitX, iInitY, iInitZ, iSeedOFB);
				else
					Y1TraversalPosDg (iInitX, iCpyInitY, iInitZ, iSeedOFB);
			}
		}
		else
		//
		//	X ^- - - - - - - - - - -|
		//	  | *				  / 
		//	  |   * 			/	|
		//	  | 	*		  / 	
		//	  | 	  * 	/		|
		//	  | 		* / 		
		//	  | 		/ * 		|
		//	  | 	  / 	*		
		//	  | 	/		  * 	|
		//	  |   / 			*	
		//	  | /				  * | 
		//	  *---------------------> Z
		//	any Y
		// 
		if (iInitX && iInitZ && (-iInitX == iInitZ))
		{
			int iStopZ; // stop points
			int iFirstSweep = 0;
			
			BOOL bTurnZ, bTurnX; // make a 90deg turn during the traversal? 
			bTurnZ = bTurnX = FALSE;
			
			// test bounds for seed cell
			iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
			
			if (iInitX > 0)
			{
				// 1st directions +z -x
				// 2nd directions +y -y
				
				// determine Z extents
				// +z
				if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, abs(iInitZ)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
				{
					iStopZ = abs(iInitZ);
					bTurnZ = TRUE;
				}
				else
				{
					// determine stop point
					VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, 0, +1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
					
					if (iInitZ+iOrZ < 0 && vIntersect.z+iOrZ < 0) return;
					
					iStopZ = vIntersect.z;
					bTurnZ = FALSE;
				}
				if (iZOFB)
					iStopZ = vOFB.z;
				
				// put seed cell within Z bounds
				if (iInitZ+iOrZ > SIZE_ZM1) { iInitZ = SIZE_ZM1-iOrZ; iFirstSweep = 1; }
				if (iInitZ+iOrZ < 0)		{ iInitZ = -iOrZ; iFirstSweep = 1; }
				
				int iCpyInitY = iInitY; // get a copy of original Y
				if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
				if (iInitY+iOrY < 0)		iInitY = -iOrY;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
				
				if (!bTurnZ)
				{
					for (iZ = iInitZ+1; iZ <= iStopZ; iZ++)
						Y1TraversalNegDg (iInitX, iCpyInitY, iZ, iSeedOFB);
				}
				else
				{
					for (iZ = iInitZ+1; iZ < iStopZ; iZ++)
						Y1TraversalNegDg (iInitX, iCpyInitY, iZ, iSeedOFB);
					
					if (!iZOFB)
						YXTraversalNegDg (iInitX, iInitY, iZ, iSeedOFB);
				}
				
				if (!iFirstSweep)
					YXTraversalNegDg (iInitX, iInitY, iInitZ, iSeedOFB); 
				else
					Y1TraversalNegDg (iInitX, iCpyInitY, iInitZ, iSeedOFB);
			}
			else
			{
				// 1st directions -z +x
				// 2nd directions +y -y
				
				// determine Z extents
				// -z
				if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, -abs(iInitZ)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
				{
					iStopZ = -abs(iInitZ);
					bTurnZ = TRUE;
				}
				else
				{
					// determine stop point
					VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, 0, -1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
					
					if (iInitZ+iOrZ > SIZE_ZM1 && vIntersect.z+iOrZ > SIZE_ZM1) return;
					
					iStopZ = vIntersect.z;
					bTurnZ = FALSE;
				}
				if (iZOFB)
					iStopZ = vOFB.z;
				
				// put seed cell within Z bounds
				if (iInitZ+iOrZ > SIZE_ZM1) { iInitZ = SIZE_ZM1-iOrZ; iFirstSweep = 1; }
				if (iInitZ+iOrZ < 0)		{ iInitZ = -iOrZ; iFirstSweep = 1; }
				
				int iCpyInitY = iInitY; // get a copy of original Y
				if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
				if (iInitY+iOrY < 0)		iInitY = -iOrY;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
				
				if (!bTurnZ)
				{
					for (iZ = iInitZ-1; iZ >= iStopZ; iZ--)
						Y1TraversalPosDg (iInitX, iCpyInitY, iZ, iSeedOFB);
				}
				else
				{
					for (iZ = iInitZ-1; iZ > iStopZ; iZ--)
						Y1TraversalPosDg (iInitX, iCpyInitY, iZ, iSeedOFB);
					
					if (!iZOFB)
						YXTraversalPosDg (iInitX, iInitY, iZ, iSeedOFB);
				}
				
				if (!iFirstSweep)
					YXTraversalPosDg (iInitX, iInitY, iInitZ, iSeedOFB);
				else
					Y1TraversalPosDg (iInitX, iCpyInitY, iInitZ, iSeedOFB);
			}
		}
		else
		//
		//	Y ^- - - - - - - - - - -|
		//	  | \				  * 
		//	  |   \ 			*	|
		//	  | 	\		  * 	
		//	  | 	  \ 	*		|
		//	  | 		\ * 		
		//	  | 		* \ 		|
		//	  | 	  * 	\		
		//	  | 	*		  \ 	|
		//	  |   * 			\	
		//	  | *				  \ | 
		//	  *---------------------> Z
		//	any X
		// 
		if (iInitY && iInitZ && (iInitY == iInitZ))
		{
			int iStopZ; // stop points
			int iFirstSweep = 0;
			
			BOOL bTurnZ, bTurnY; // make a 90deg turn during the traversal? 
			bTurnZ = bTurnY = FALSE;
			
			// test bounds for seed cell
			iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
			
			if (iInitY > 0)
			{
				// 1st directions -z -y
				// 2nd directions +x -x
				
				// determine Z extents
				// -z
				if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, -abs(iInitZ)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
				{
					iStopZ = -abs(iInitZ);
					bTurnZ = TRUE;
				}
				else
				{
					// determine stop point
					VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, 0, -1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
					
					if (iInitZ+iOrZ > SIZE_ZM1 && vIntersect.z+iOrZ > SIZE_ZM1) return;
					
					iStopZ = vIntersect.z;
					bTurnZ = FALSE;
				}
				
				if (iZOFB)
					iStopZ = vOFB.z;
				
				// put seed cell within Z bounds
				if (iInitZ+iOrZ > SIZE_ZM1) { iInitZ = SIZE_ZM1-iOrZ; iFirstSweep = 1; }
				if (iInitZ+iOrZ < 0)		{ iInitZ = -iOrZ; iFirstSweep = 1; }
				
				int iCpyInitX = iInitX; // get a copy of original X
				if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
				if (iInitX+iOrX < 0)		iInitX = -iOrX;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
				
				if (!bTurnZ)
				{
					for (iZ = iInitZ-1; iZ >= iStopZ; iZ--)
						XTraversalNegDg (iCpyInitX, iInitY, iZ, iSeedOFB);
				}
				else
				{
					for (iZ = iInitZ-1; iZ > iStopZ; iZ--)
						XTraversalNegDg (iCpyInitX, iInitY, iZ, iSeedOFB);
					
					if (!iZOFB)
						XYTraversalNegDg (iInitX, iInitY, iZ, iSeedOFB);
				}
				
				if (!iFirstSweep)
					XYTraversalNegDg (iInitX, iInitY, iInitZ, iSeedOFB); 
				else
					XTraversalNegDg (iCpyInitX, iInitY, iInitZ, iSeedOFB);
			}
			else
			{
				// 1st directions +z +y
				// 2nd directions +x -x
				
				// determine Z extents
				// +z
				if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, abs(iInitZ)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
				{
					iStopZ = abs(iInitZ);
					bTurnZ = TRUE;
				}
				else
				{
					// determine stop point
					VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, 0, +1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
					
					if (iInitZ+iOrZ < 0 && vIntersect.z+iOrZ < 0) return;
					
					iStopZ = vIntersect.z;
					bTurnZ = FALSE;
				}
				
				if (iZOFB)
					iStopZ = vOFB.z;
				
				// put seed cell within Z bounds
				if (iInitZ+iOrZ > SIZE_ZM1) { iInitZ = SIZE_ZM1-iOrZ; iFirstSweep = 1; }
				if (iInitZ+iOrZ < 0)		{ iInitZ = -iOrZ; iFirstSweep = 1; }
				
				int iCpyInitX = iInitX; // get a copy of original X
				if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
				if (iInitX+iOrX < 0)		iInitX = -iOrX;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
				
				if (!bTurnZ)
				{
					for (iZ = iInitZ+1; iZ <= iStopZ; iZ++)
						XTraversalPosDg (iCpyInitX, iInitY, iZ, iSeedOFB);
				}
				else
				{
					for (iZ = iInitZ+1; iZ < iStopZ; iZ++)
						XTraversalPosDg (iCpyInitX, iInitY, iZ, iSeedOFB);
					
					if (!iZOFB)
						XYTraversalPosDg (iInitX, iInitY, iZ, iSeedOFB);
				}
				
				if (!iFirstSweep)
					XYTraversalPosDg (iInitX, iInitY, iInitZ, iSeedOFB);
				else
					XTraversalPosDg (iCpyInitX, iInitY, iInitZ, iSeedOFB);
			} 
		}
		else
		//
		//	Y ^- - - - - - - - - - -|
		//	  | *				  / 
		//	  |   * 			/	|
		//	  | 	*		  / 	
		//	  | 	  * 	/		|
		//	  | 		* / 		
		//	  | 		/ * 		|
		//	  | 	  / 	*		
		//	  | 	/		  * 	|
		//	  |   / 			*	
		//	  | /				  * | 
		//	  *---------------------> Z
		//	any X
		// 
		if (iInitY && iInitZ && (-iInitY == iInitZ))
		{
			int iStopZ; // stop points
			int iFirstSweep = 0;
			
			BOOL bTurnZ, bTurnY; // make a 90deg turn during the traversal? 
			bTurnZ = bTurnY = FALSE;
			
			// test bounds for seed cell
			iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
			
			if (iInitY > 0)
			{
				// 1st directions +z -y
				// 2nd directions +x -x
				
				// determine Z extents
				// +z
				if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, abs(iInitZ)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
				{
					iStopZ = abs(iInitZ);
					bTurnZ = TRUE;
				}
				else
				{
					// determine stop point
					VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, 0, +1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
					
					if (iInitZ+iOrZ < 0 && vIntersect.z+iOrZ < 0) return;
					
					iStopZ = vIntersect.z;
					bTurnZ = FALSE;
				}
				if (iZOFB)
					iStopZ = vOFB.z;
				
				// put seed cell within Z bounds
				if (iInitZ+iOrZ > SIZE_ZM1) { iInitZ = SIZE_ZM1-iOrZ; iFirstSweep = 1; }
				if (iInitZ+iOrZ < 0)		{ iInitZ = -iOrZ; iFirstSweep = 1; }
				
				int iCpyInitX = iInitX; // get a copy of original X
				if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
				if (iInitX+iOrX < 0)		iInitX = -iOrX;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
				
				if (!bTurnZ)
				{
					for (iZ = iInitZ+1; iZ <= iStopZ; iZ++)
						XTraversalNegDg (iCpyInitX, iInitY, iZ, iSeedOFB);
				}
				else
				{
					for (iZ = iInitZ+1; iZ < iStopZ; iZ++)
						XTraversalNegDg (iCpyInitX, iInitY, iZ, iSeedOFB);
					
					if (!iZOFB)
						XYTraversalNegDg (iInitX, iInitY, iZ, iSeedOFB);
				}
				
				if (!iFirstSweep)
					XYTraversalNegDg (iInitX, iInitY, iInitZ, iSeedOFB); 
				else
					XTraversalNegDg (iCpyInitX, iInitY, iInitZ, iSeedOFB);
			}
			else
			{
				// 1st directions -z +y
				// 2nd directions +x -x
				
				// determine Z extents
				// -z
				if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, -abs(iInitZ)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
				{
					iStopZ = -abs(iInitZ);
					bTurnZ = TRUE;
				}
				else
				{
					// determine stop point
					VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, 0, -1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
					
					if (iInitZ+iOrZ > SIZE_ZM1 && vIntersect.z+iOrZ > SIZE_ZM1) return;
					
					iStopZ = vIntersect.z;
					bTurnZ = FALSE;
				}
				if (iZOFB)
					iStopZ = vOFB.z;
				
				// put seed cell within Z bounds
				if (iInitZ+iOrZ > SIZE_ZM1) { iInitZ = SIZE_ZM1-iOrZ; iFirstSweep = 1; }
				if (iInitZ+iOrZ < 0)		{ iInitZ = -iOrZ; iFirstSweep = 1; }
				
				int iCpyInitX = iInitX; // get a copy of original X
				if (iInitX+iOrZ > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
				if (iInitX+iOrZ < 0)		iInitX = -iOrX;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
				
				if (!bTurnZ)
				{
					for (iZ = iInitZ-1; iZ >= iStopZ; iZ--)
						XTraversalPosDg (iCpyInitX, iInitY, iZ, iSeedOFB);
				}
				else
				{
					for (iZ = iInitZ-1; iZ > iStopZ; iZ--)
						XTraversalPosDg (iCpyInitX, iInitY, iZ, iSeedOFB);
					
					if (!iZOFB)
						XYTraversalPosDg (iInitX, iInitY, iZ, iSeedOFB);
				}
				
				if (!iFirstSweep)
					XYTraversalPosDg (iInitX, iInitY, iInitZ, iSeedOFB);
				else
					XTraversalPosDg (iCpyInitX, iInitY, iInitZ, iSeedOFB);
			} 
		}
		else
		//
		//  Y ^- - - - - - - - - - -|
		//    | \                 * 
		//    |   \             *   |
		//    |     \         *     
		//    |       \     *       |
		//    |         \ *         
		//    |         * \         |
		//    |       *     \       
		//    |     *         \     |
		//    |   *             \   
		//    | *                 \ | 
		//    *---------------------> X
		//  any Z
		// 
		if (iInitY && iInitX && (iInitY == iInitX))
		{
			int iStopY; // stop points
			int iFirstSweep = 0;
			
			BOOL bTurnY, bTurnX; // make a 90deg turn during the traversal? 
			bTurnY = bTurnX = FALSE;
			
			// test bounds for seed cell
			iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
			
			if (iInitX > 0)
			{
				// 1st directions -y -x
				// 2nd directions +z -z
				
				// determine Y extents
				// -y
				if (isInsideVFBnd (iInitX+iOrX, -abs(iInitY)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
				{
					iStopY = -abs(iInitY);
					bTurnY = TRUE;
				}
				else
				{
					// determine stop point
					VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
					
					if (iInitY+iOrY > SIZE_YM1 && vIntersect.y+iOrY > SIZE_YM1) return;
					
					iStopY = vIntersect.y;
					bTurnY = FALSE;
				}
				
				if (iYOFB) // added
					iStopY = vOFB.y; // added
				
				// put seed cell within Y bounds
				if (iInitY+iOrY > SIZE_YM1) { iInitY = SIZE_YM1-iOrY; iFirstSweep = 1; }
				if (iInitY+iOrY < 0)		{ iInitY = -iOrY; iFirstSweep = 1; }
				
				int iCpyInitZ = iInitZ; // get a copy of original Z
				if (iInitZ+iOrZ > SIZE_ZM1) iInitZ = SIZE_ZM1-iOrZ;
				if (iInitZ+iOrZ < 0)		iInitZ = -iOrZ;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
				
				if (!bTurnY)
				{
					for (iY = iInitY-1; iY >= iStopY; iY--)
						Z1TraversalNeg (iInitX, iY, iCpyInitZ, iSeedOFB);
				}
				else
				{
					for (iY = iInitY-1; iY > iStopY; iY--)
						Z1TraversalNeg (iInitX, iY, iCpyInitZ, iSeedOFB);
					
					if (!iYOFB) // added
						ZXTraversalNeg (iInitX, iY, iInitZ, iSeedOFB);
				}
				
				if (!iFirstSweep) // added
					ZXTraversalNeg (iInitX, iInitY, iInitZ, iSeedOFB);
				else
					Z1TraversalNeg (iInitX, iInitY, iCpyInitZ, iSeedOFB);
			}
			else
			{
				// 1st directions +y +x
				// 2nd directions +z -z
				
				// determine Y extents
				// +y
				if (isInsideVFBnd (iInitX+iOrX, abs(iInitY)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
				{
					iStopY = abs(iInitY);
					bTurnY = TRUE;
				}
				else
				{
					// determine stop point
					VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
					
					if (iInitY+iOrY < 0 && vIntersect.y+iOrY < 0) return;
					
					iStopY = vIntersect.y;
					bTurnY = FALSE;
				}
				
				if (iYOFB) // added
					iStopY = vOFB.y; // added
				
				// put seed cell within Y bounds
				if (iInitY+iOrY > SIZE_YM1) { iInitY = SIZE_YM1-iOrY; iFirstSweep = 1; }
				if (iInitY+iOrY < 0)		{ iInitY = -iOrY; iFirstSweep = 1; }
				
				int iCpyInitZ = iInitZ; // get a copy of original Z
				if (iInitZ+iOrZ > SIZE_ZM1) iInitZ = SIZE_ZM1-iOrZ;
				if (iInitZ+iOrZ < 0)		iInitZ = -iOrZ;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
				
				if (!bTurnY)
				{
					for (iY = iInitY+1; iY <= iStopY; iY++)
						Z1TraversalPos (iInitX, iY, iCpyInitZ, iSeedOFB);
				}
				else
				{
					for (iY = iInitY+1; iY < iStopY; iY++)
						Z1TraversalPos (iInitX, iY, iCpyInitZ, iSeedOFB);
					
					if (!iYOFB) // added
						ZXTraversalPos (iInitX, iY, iInitZ, iSeedOFB);
				}
				
				if (!iFirstSweep) // added
					ZXTraversalPos (iInitX, iInitY, iInitZ, iSeedOFB);
				else
					Z1TraversalPos (iInitX, iInitY, iCpyInitZ, iSeedOFB);
			}
		}
		else
		//
		//  Y ^- - - - - - - - - - -|
		//    | *                 / 
		//    |   *             /   |
		//    |     *         /     
		//    |       *     /       |
		//    |         * /         
		//    |         / *         |
		//    |       /     *       
		//    |     /         *     |
		//    |   /             *   
		//    | /                 * | 
		//    *---------------------> X
		//  any Z
		// 
		if (iInitY && iInitX && (-iInitY == iInitX))
		{
			int iStopY; // stop points
			int iFirstSweep = 0;
			
			BOOL bTurnY, bTurnX; // make a 90deg turn during the traversal? 
			bTurnY = bTurnX = FALSE;
			
			// test bounds for seed cell
			iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
			
			if (iInitX > 0)
			{
				// 1st directions +y -x
				// 2nd directions +z -z
				
				// determine Y extents
				// +y
				if (isInsideVFBnd (iInitX+iOrX, abs(iInitY)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
				{
					iStopY = abs(iInitY);
					bTurnY = TRUE;
				}
				else
				{
					// determine stop point
					VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
					
					if (iInitY+iOrY < 0 && vIntersect.y+iOrY < 0) return;
					
					iStopY = vIntersect.y;
					bTurnY = FALSE;
				}
				if (iYOFB) // added
					iStopY = vOFB.y; // added
				
				// put seed cell within Y bounds
				if (iInitY+iOrY > SIZE_YM1) { iInitY = SIZE_YM1-iOrY; iFirstSweep = 1; }
				if (iInitY+iOrY < 0)		{ iInitY = -iOrY; iFirstSweep = 1; }
				
				int iCpyInitZ = iInitZ; // get a copy of original Z
				if (iInitZ+iOrZ > SIZE_ZM1) iInitZ = SIZE_ZM1-iOrZ;
				if (iInitZ+iOrZ < 0)		iInitZ = -iOrZ;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
				
				if (!bTurnY)
				{
					for (iY = iInitY+1; iY <= iStopY; iY++)
						Z1TraversalNeg (iInitX, iY, iCpyInitZ, iSeedOFB);
				}
				else
				{
					for (iY = iInitY+1; iY < iStopY; iY++)
						Z1TraversalNeg (iInitX, iY, iCpyInitZ, iSeedOFB);
					
					if (!iYOFB) // added
						ZXTraversalNeg (iInitX, iY, iInitZ, iSeedOFB);
				}
				
				if (!iFirstSweep) // added
					ZXTraversalNeg (iInitX, iInitY, iInitZ, iSeedOFB);
				else
					Z1TraversalNeg (iInitX, iInitY, iCpyInitZ, iSeedOFB);
			}
			else
			{
				// 1st directions -y +x
				// 2nd directions +z -z
				
				// determine Y extents
				// -y
				if (isInsideVFBnd (iInitX+iOrX, -abs(iInitY)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
				{
					iStopY = -abs(iInitY);
					bTurnY = TRUE;
				}
				else
				{
					// determine stop point
					VECTOR3 vIntersect = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
					
					if (iInitY+iOrY > SIZE_YM1 && vIntersect.y+iOrY > SIZE_YM1) return;
					
					iStopY = vIntersect.y;
					bTurnY = FALSE;
				}
				if (iYOFB) // added
					iStopY = vOFB.y; // added
				
				// put seed cell within Y bounds
				if (iInitY+iOrY > SIZE_YM1) { iInitY = SIZE_YM1-iOrY; iFirstSweep = 1; }
				if (iInitY+iOrY < 0)		{ iInitY = -iOrY; iFirstSweep = 1; }
				
				int iCpyInitZ = iInitZ; // get a copy of original Z
				if (iInitZ+iOrZ > SIZE_ZM1) iInitZ = SIZE_ZM1-iOrZ;
				if (iInitZ+iOrZ < 0)		iInitZ = -iOrZ;
				iSeedOFB = isSeedOutsideBnd (iInitX+iOrX, iInitY+iOrY, iInitZ+iOrZ);
				
				if (!bTurnY)
				{
					for (iY = iInitY-1; iY >= iStopY; iY--)
						Z1TraversalPos (iInitX, iY, iCpyInitZ, iSeedOFB);
				}
				else
				{
					for (iY = iInitY-1; iY > iStopY; iY--)
						Z1TraversalPos (iInitX, iY, iCpyInitZ, iSeedOFB);
					
					if (!iYOFB) // added
						ZXTraversalPos (iInitX, iY, iInitZ, iSeedOFB);
				}
				
				if (!iFirstSweep) // added
					ZXTraversalPos (iInitX, iInitY, iInitZ, iSeedOFB);
				else
					Z1TraversalPos (iInitX, iInitY, iCpyInitZ, iSeedOFB);
			}
		}
	}

	if (iDist == 1) return;

	// convert raster image to vector
	if (bMatrixF)
	{
		if (!bRasterFullF)
		{
			drawConcPol = &Vis3D::drawConcPolF; // set drawConcPol function
			iRasterPlane = raFront;
			fCPMinX = iCPMinXF; fCPMaxX = iCPMaxXF;
			fCPMinY = iCPMinYF; fCPMaxY = iCPMaxYF;
			ras2Vec (iMatrixF, iDistSize); // front
			(this->*drawConcPol)();
			bRasterFullF = bRasterFull;
		}
		bMatrixF = 0;
		memset (iMatrixF, 0, sizeof(int)*iDistSize*iDistSize); // clear raster image
	}

	if (bMatrixB)
	{
		if (!bRasterFullB)
		{
			drawConcPol = &Vis3D::drawConcPolB; // set drawConcPol function
			iRasterPlane = raBack;
			fCPMinX = iCPMinXB; fCPMaxX = iCPMaxXB;
			fCPMinY = iCPMinYB; fCPMaxY = iCPMaxYB;
			ras2Vec (iMatrixB, iDistSize); // back
			(this->*drawConcPol)();
			bRasterFullB = bRasterFull;
		}
		bMatrixB = 0;
		memset (iMatrixB, 0, sizeof(int)*iDistSize*iDistSize); // clear raster image
	}

	if (bMatrixL)
	{
		if (!bRasterFullL)
		{
			drawConcPol = &Vis3D::drawConcPolL; // set drawConcPol function
			iRasterPlane = raLeft;
			fCPMinX = iCPMinXL; fCPMaxX = iCPMaxXL;
			fCPMinY = iCPMinYL; fCPMaxY = iCPMaxYL;
			ras2Vec (iMatrixL, iDistSize); // left
			(this->*drawConcPol)();
			bRasterFullL = bRasterFull;
		}
		bMatrixL = 0;
		memset (iMatrixL, 0, sizeof(int)*iDistSize*iDistSize); // clear raster image
	}

	if (bMatrixR)
	{
		if (!bRasterFullR)
		{
			drawConcPol = &Vis3D::drawConcPolR; // set drawConcPol function
			iRasterPlane = raRight;
			fCPMinX = iCPMinXR; fCPMaxX = iCPMaxXR;
			fCPMinY = iCPMinYR; fCPMaxY = iCPMaxYR;
			ras2Vec (iMatrixR, iDistSize); // right
			(this->*drawConcPol)();
			bRasterFullR = bRasterFull;
		}
		bMatrixR = 0;
		memset (iMatrixR, 0, sizeof(int)*iDistSize*iDistSize); // clear raster image
	}

	if (bMatrixT) 
	{
		if (!bRasterFullT)
		{
			drawConcPol = &Vis3D::drawConcPolT; // set drawConcPol function
			iRasterPlane = raTop;
			fCPMinX = iCPMinXT; fCPMaxX = iCPMaxXT;
			fCPMinY = iCPMinYT; fCPMaxY = iCPMaxYT;
			ras2Vec (iMatrixT, iDistSize); // top
			(this->*drawConcPol)();
			bRasterFullT = bRasterFull;
		}
		bMatrixT = 0;
		memset (iMatrixT, 0, sizeof(int)*iDistSize*iDistSize); // clear raster image
	}

	if (bMatrixD)
	{
		if (!bRasterFullD)
		{
			drawConcPol = &Vis3D::drawConcPolD; // set drawConcPol function
			iRasterPlane = raDown;
			fCPMinX = iCPMinXD; fCPMaxX = iCPMaxXD;
			fCPMinY = iCPMinYD; fCPMaxY = iCPMaxYD;
			ras2Vec (iMatrixD, iDistSize); // down
			(this->*drawConcPol)();
			bRasterFullD = bRasterFull;
		}
		bMatrixD = 0;
		memset (iMatrixD, 0, sizeof(int)*iDistSize*iDistSize); // clear raster image
	}
} // Vis3D::startTraversal

/////////////////////////////////////////////////////////////////////////////
// Traverse the voxels spanned by a line (x1,y1,z1)-(x2,y2,z2) using Bresenham
/////////////////////////////////////////////////////////////////////////////
void Vis3D::bresenhamLine3D(int x1, int y1, int z1, int x2, int y2, int z2)
{
	int i, dx, dy, dz, l, m, n, x_inc, y_inc, z_inc,
		err_1, err_2, dx2, dy2, dz2;

	// relative to the viewer
	x2 -= x1; y2 -= y1; z2 -= z1;
	x1 = y1 = z1 = 0;
	
	int	pixelX = x1;
	int pixelY = y1;
	int pixelZ = z1;
    
	dx = x2 - x1;
	dy = y2 - y1;
	dz = z2 - z1;
    
	x_inc = (dx < 0) ? -1 : 1;
	l = abs(dx);
	y_inc = (dy < 0) ? -1 : 1;
	m = abs(dy);
	z_inc = (dz < 0) ? -1 : 1;
	n = abs(dz);
	dx2 = l << 1;
	dy2 = m << 1;
	dz2 = n << 1;

	if ((l >= m) && (l >= n)) 
	{
		err_1 = dy2 - l;
		err_2 = dz2 - l;
		for (i = 0; i < l; i++) 
		{
			startTraversal (pixelX, pixelY, pixelZ);

			if (bPrematureEnd) // premature end of the traversal
				return;

			if (err_1 > 0) 
			{
				pixelY += y_inc;
				err_1 -= dx2;
			}
			if (err_2 > 0) 
			{
				pixelZ += z_inc;
				err_2 -= dx2;
			}
			err_1 += dy2;
			err_2 += dz2;
			pixelX += x_inc;
		}
	} 
	else if ((m >= l) && (m >= n)) 
	{
		err_1 = dx2 - m;
		err_2 = dz2 - m;
		for (i = 0; i < m; i++) 
		{
			startTraversal (pixelX, pixelY, pixelZ);

			if (bPrematureEnd) // premature end of the traversal
				return;

			if (err_1 > 0) 
			{
				pixelX += x_inc;
				err_1 -= dy2;
			}
			if (err_2 > 0) 
			{
				pixelZ += z_inc;
				err_2 -= dy2;
			}
			err_1 += dx2;
			err_2 += dz2;
			pixelY += y_inc;
		}
	} 
	else 
	{
		err_1 = dy2 - n;
		err_2 = dx2 - n;
		for (i = 0; i < n; i++) 
		{
			startTraversal (pixelX, pixelY, pixelZ);

			if (bPrematureEnd) // premature end of the traversal
				return;

			if (err_1 > 0) 
			{
				pixelY += y_inc;
				err_1 -= dz2;
			}
			if (err_2 > 0) 
			{
				pixelX += x_inc;
				err_2 -= dz2;
			}
			err_1 += dy2;
			err_2 += dx2;
			pixelZ += z_inc;
		}
	}

	startTraversal (pixelX, pixelY, pixelZ);
} // Vis3D::bresenhamLine3D

void Vis3D::show (void)
{
	int iCt;

	glClearColor (1.0, 1.0, 1.0, 0.0);
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// viewpoint position in integer coordinates
	iOrX = ROUND(vPos.x);
	iOrY = ROUND(vPos.y);
	iOrZ = ROUND(vPos.z);

	// transform view frustum
	transformVFC ();

	double FPS = getFPS ();

	bRasterFullF = bRasterFullB = bRasterFullL = bRasterFullR = bRasterFullT = bRasterFullD = false;

	// discretize last visible objects
#ifdef USETBV
	for (iCt = 0; iCt < N_OBJS_DYN; iCt++)
	{
		if (PVS[iCt])
		{
			// object doesn't have a TBV
			putIntoGrid (iCt, false, false);
		}
		else
		{
			// does the object have a TBV?
			if (disks[iCt].HasTBV ())
			{
				// TBV expired
				if (disks[iCt].GetExpDate () <= current_date)
				{
					long last_interval = disks[iCt].GetExpDate () - disks[iCt].GetStartDate ();

					// assign a new TBV
					eraseTBV (iCt);

					disks[iCt].SetTBVX (disks[iCt].GetX ());
					disks[iCt].SetTBVY (disks[iCt].GetY ());
					disks[iCt].SetTBVZ (disks[iCt].GetZ ());
					disks[iCt].SetStartDate (current_date);
					disks[iCt].SetExpDate (current_date + ((last_interval+1) * 1.3));
					insertTBV (iCt);

					disks[iCt].SetTBVState (true);
				}
			}
			else
			{
				long last_interval = disks[iCt].GetExpDate () - disks[iCt].GetStartDate ();

				// assign a TBV
				// object is just getting invisible
				disks[iCt].SetTBVX (disks[iCt].GetX ());
				disks[iCt].SetTBVY (disks[iCt].GetY ());
				disks[iCt].SetTBVZ (disks[iCt].GetZ ());
				disks[iCt].SetStartDate (current_date);
				disks[iCt].SetExpDate (current_date + (last_interval / 2.0));
				insertTBV (iCt);

				putIntoGrid (iCt, false, true);

				disks[iCt].SetTBVState (true);
			}
		}
	}
#endif

	int iVisObjs = 0,	// pot. vis. obj.
		iVisObjsVF = 0, // pot. vis. obj. using vf-culling only
		iVisExact = 0;  // exact visibility

	glViewport (WIN_WIDTH/2, 0, WIN_WIDTH, WIN_HEIGHT+50);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	glOrtho (0, WIN_WIDTH, 0, WIN_HEIGHT+50, 1.0, 10000.0);

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	// clear scene window
	glDisable (GL_CULL_FACE);
	glDisable (GL_DEPTH_TEST);

	glColor3f (0.0, 0.0, 0.0);
	glBegin (GL_QUADS);
		glVertex3f (0,                    0, -10000);
		glVertex3f (0,           WIN_HEIGHT, -10000);
		glVertex3f (WIN_WIDTH/2, WIN_HEIGHT, -10000);
		glVertex3f (WIN_WIDTH/2,          0, -10000);
	glEnd ();

	// draw scene
	glViewport (WIN_WIDTH/2, 0, WIN_WIDTH/2, WIN_HEIGHT);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	gluPerspective (fFOV, 1.0, 0.1, zFar);
	gluLookAt (vPos.x, vPos.y, vPos.z, 
			  (tCorner1.x+tCorner2.x)/2, (tCorner1.y+tCorner2.y)/2, (tCorner1.z+tCorner2.z)/2, 
			   0, 1, 0);
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	glEnable (GL_CULL_FACE);
	glEnable (GL_DEPTH_TEST);
	glShadeModel (GL_SMOOTH);

	GLfloat light0_position[] = { SIZE_X, SIZE_Y, SIZE_Z, 1.0 };
	GLfloat light1_position[] = { 0, SIZE_Y/2, -SIZE_Z/2, 1.0 };
	GLfloat light2_position[] = { SIZE_X, SIZE_Y, 0, 1.0 };
	GLfloat light3_position[] = { SIZE_X/2, SIZE_Y/2, 0, 1.0 };
	GLfloat light1_color[] = { (float)0.5, (float)0.5, (float)0.5 };
	GLfloat light2_color[] = { (float)0.6, (float)0.6, (float)0.6 };
	GLfloat light3_color[] = { (float)0.5, (float)0.5, (float)0.5 };
	glLightfv (GL_LIGHT0, GL_POSITION, light0_position);
	glLightfv (GL_LIGHT1, GL_POSITION, light1_position);
	glLightfv (GL_LIGHT1, GL_DIFFUSE,  light1_color);
	glLightfv (GL_LIGHT2, GL_POSITION, light2_position);
	glLightfv (GL_LIGHT2, GL_DIFFUSE,  light2_color);
	glLightfv (GL_LIGHT3, GL_POSITION, light3_position);
	glLightfv (GL_LIGHT3, GL_DIFFUSE,  light3_color);

	glEnable (GL_LIGHTING);
	glEnable (GL_LIGHT0);
	glEnable (GL_LIGHT1);
	glEnable (GL_LIGHT2);
	glEnable (GL_LIGHT3);

	if (!O.get(MK_OFF_GRID(iOrX, iOrY, iOrZ)))
	{
		bresenhamLine3D (ROUND(vPos.x), ROUND(vPos.y), ROUND(vPos.z), 
						(tCorner1.x+tCorner2.x)/2, 
						(tCorner1.y+tCorner2.y)/2,
						(tCorner1.z+tCorner2.z)/2);
	}
	else
		Add2PVS (0, 0, 0);

	glDisable (GL_LIGHT3);
	glDisable (GL_LIGHT2);
	glDisable (GL_LIGHT1);
	glDisable (GL_LIGHT0);
	glDisable (GL_LIGHTING);

	if (!hiddenGeometry)
	{
		// calculate exact visibility using the item buffer?
		iVisExact = 0;
		if (bCalcExactVis)
		{
			glDisable (GL_LIGHTING);

			// draw potentially visible cylinders
			for (iCt = 0; iCt < N_OBJS_STC + N_OBJS_DYN; iCt++)
			{
				if(PVS[iCt]) 
					disks[iCt].DrawDisk (this);
			}

			glFinish ();
			glReadPixels (WIN_WIDTH/2, 0, WIN_WIDTH/2, WIN_HEIGHT, GL_RED, GL_UNSIGNED_BYTE, ucReadBufferR);
			glReadPixels (WIN_WIDTH/2, 0, WIN_WIDTH/2, WIN_HEIGHT, GL_GREEN, GL_UNSIGNED_BYTE, ucReadBufferG);
			glReadPixels (WIN_WIDTH/2, 0, WIN_WIDTH/2, WIN_HEIGHT, GL_BLUE, GL_UNSIGNED_BYTE, ucReadBufferB);

			memset (PVSEXACT, false, sizeof(bool)*(N_OBJS_DYN+N_OBJS_STC));
			for (iCt = 0; iCt < (WIN_WIDTH/2)*WIN_HEIGHT; iCt++)
			{
				int idx = (ucReadBufferR[iCt] << 16) | (ucReadBufferG[iCt] << 8) | (ucReadBufferB[iCt]);
				if (idx < N_OBJS_DYN+N_OBJS_STC)
					PVSEXACT[idx] = true;
			}

			for (iCt = 0; iCt < N_OBJS_DYN+N_OBJS_STC; iCt++)
				if (PVSEXACT[iCt]) iVisExact++;
		}
	}

	glShadeModel (GL_FLAT);
	glDisable (GL_CULL_FACE);
	glDisable (GL_DEPTH_TEST);

	// visualize data structure
	glViewport (0, 0, WIN_WIDTH/2, WIN_HEIGHT);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	gluPerspective (60, 1.0, 1.0, 300.0);
	gluLookAt (0, 0, 50, 0, 0, 0, 0, 1, 0);

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	glTranslatef (0.0, 0.0, fWorldZTranslate);

    float matRotationInverse[17];
    MatrixInverse ((float*)&matRotationInverse, arcBall.GetRotationMatrix());
    glMultMatrixf (matRotationInverse);
	arcBall.GetTranslationMatrix ((float*)&matRotationInverse);
	glMultMatrixf (matRotationInverse);

	glRotatef (180.0, 0.0, 1.0, 0.0);
	glTranslatef (-(SIZE_X*STEP)/2, -(SIZE_Y*STEP)/2, -(SIZE_Z*STEP)/2);

	glEnable (GL_DEPTH_TEST);
	glEnable (GL_CULL_FACE);

	// draw axis
	glLineWidth (3.0);
	glBegin (GL_LINES);
		glColor3f (1.0, 0.0, 0.0);	// X
		glVertex3f (0, 0, 0);	glVertex3f (SIZE_X*STEP, 0, 0);
		glColor3f (0.0, 1.0, 0.0);	// Y
		glVertex3f (0, 0, 0);	glVertex3f (0, SIZE_Y*STEP, 0);
		glColor3f (0.0, 0.0, 1.0);	// Z
		glVertex3f (0, 0, 0);	glVertex3f (0, 0, SIZE_Z*STEP);
	glEnd ();
	glLineWidth (1.0);

	// draw viewpoint
	glColor3f (0.0, 0.0, 0.0);
	glPointSize (5.0);
	glBegin (GL_POINTS);
		glVertex3f ((vPos.x*STEP)+(STEP/2.0), (vPos.y*STEP)+(STEP/2.0), (vPos.z*STEP)+(STEP/2.0));
	glEnd ();
	glPointSize (1.0);

	// draw view frustum
	glColor3f (0.0, 0.0, 1.0);
	glBegin (GL_LINES);
		glColor3f (0.0, 0.0, 1.0);
		glVertex3f ((vPos.x*STEP)+(STEP/2), (vPos.y*STEP)+(STEP/2), (vPos.z*STEP)+(STEP/2));
		glVertex3f ((tCorner1.x*STEP)+(STEP/2), (tCorner1.y*STEP)+(STEP/2), (tCorner1.z*STEP)+(STEP/2));

		glColor3f (0.0, 0.0, 1.0);
		glVertex3f ((vPos.x*STEP)+(STEP/2), (vPos.y*STEP)+(STEP/2), (vPos.z*STEP)+(STEP/2));
		glVertex3f ((tCorner2.x*STEP)+(STEP/2), (tCorner2.y*STEP)+(STEP/2), (tCorner2.z*STEP)+(STEP/2));

		glColor3f (0.0, 0.0, 1.0);
		glVertex3f ((vPos.x*STEP)+(STEP/2), (vPos.y*STEP)+(STEP/2), (vPos.z*STEP)+(STEP/2));
		glVertex3f ((tCorner3.x*STEP)+(STEP/2), (tCorner3.y*STEP)+(STEP/2), (tCorner3.z*STEP)+(STEP/2));

		glColor3f (0.0, 0.0, 1.0);
		glVertex3f ((vPos.x*STEP)+(STEP/2), (vPos.y*STEP)+(STEP/2), (vPos.z*STEP)+(STEP/2));
		glVertex3f ((tCorner4.x*STEP)+(STEP/2), (tCorner4.y*STEP)+(STEP/2), (tCorner4.z*STEP)+(STEP/2));

		glColor3f (0.0, 0.0, 1.0);

		// 1, 3
		glVertex3f ((tCorner1.x*STEP)+(STEP/2), (tCorner1.y*STEP)+(STEP/2), (tCorner1.z*STEP)+(STEP/2));
		glVertex3f ((tCorner3.x*STEP)+(STEP/2), (tCorner3.y*STEP)+(STEP/2), (tCorner3.z*STEP)+(STEP/2));

		// 3, 2
		glVertex3f ((tCorner3.x*STEP)+(STEP/2), (tCorner3.y*STEP)+(STEP/2), (tCorner3.z*STEP)+(STEP/2));
		glVertex3f ((tCorner2.x*STEP)+(STEP/2), (tCorner2.y*STEP)+(STEP/2), (tCorner2.z*STEP)+(STEP/2));

		// 2, 4
		glVertex3f ((tCorner2.x*STEP)+(STEP/2), (tCorner2.y*STEP)+(STEP/2), (tCorner2.z*STEP)+(STEP/2));
		glVertex3f ((tCorner4.x*STEP)+(STEP/2), (tCorner4.y*STEP)+(STEP/2), (tCorner4.z*STEP)+(STEP/2));

		// 4, 1
		glVertex3f ((tCorner4.x*STEP)+(STEP/2), (tCorner4.y*STEP)+(STEP/2), (tCorner4.z*STEP)+(STEP/2));
		glVertex3f ((tCorner1.x*STEP)+(STEP/2), (tCorner1.y*STEP)+(STEP/2), (tCorner1.z*STEP)+(STEP/2));
	glEnd ();

	glLineStipple (2, 0xAAAA);
	glEnable (GL_LINE_STIPPLE);

	glBegin (GL_LINES);
		glVertex3f ((vPos.x*STEP)+(STEP/2), (vPos.y*STEP)+(STEP/2), (vPos.z*STEP)+(STEP/2));
		glVertex3f ((((tCorner1.x+tCorner2.x)/2)*STEP)+(STEP/2), 
					(((tCorner1.y+tCorner2.y)/2)*STEP)+(STEP/2), 
					(((tCorner1.z+tCorner2.z)/2)*STEP)+(STEP/2));
	glEnd ();

	glDisable (GL_LINE_STIPPLE);

	// draw viewcell
	glColor3f (0.0, 0.0, 0.0);

	// show hidden and opaque cells
	if (bShowData)
	{
		GLfloat light0_position[] = { (float)0.2, (float)0.6, (float)1.0, (float)0.0 };
		GLfloat light0_diffuse[] = { (float)1.0, (float)1.0, (float)1.0, (float)1.0 };
		GLfloat light1_position[] = { (float)-0.2, (float)-0.6, (float)-0.5, (float)0.0 };
		GLfloat light1_diffuse[] = { (float)1.0, (float)1.0, (float)1.0, (float)1.0 };
		glLightfv (GL_LIGHT0, GL_POSITION, light0_position);
		glLightfv (GL_LIGHT0, GL_DIFFUSE,  light0_diffuse);
		glLightfv (GL_LIGHT1, GL_POSITION, light1_position);
		glLightfv (GL_LIGHT1, GL_DIFFUSE,  light1_diffuse);

		glEnable (GL_LIGHTING);
		glEnable (GL_LIGHT0);
		glEnable (GL_LIGHT1);

		GLfloat mat_diffuse_H[] = { (float)0.6, (float)0.55, (float)1.0, (float)1.0 };
		GLfloat mat_diffuse_O[] = { (float)0.6, (float)0.6, (float)0.6, (float)1.0 };
		GLfloat mat_diffuse_I[] = { (float)1.0, (float)1.0, (float)0.4, (float)1.0 };
		GLfloat mat_diffuse_T[] = { (float)0.4, (float)1.0, (float)0.4, (float)1.0 };

		int iZTrans, iYTrans, iXTrans;

		int dsk = 0;

		for (iZTrans = 0; iZTrans < SIZE_Z; iZTrans++)
		{
			for (iYTrans = 0; iYTrans < SIZE_Y; iYTrans++)
			{
				for (iXTrans = 0; iXTrans < SIZE_X; iXTrans++)
				{
					if (H.get(MK_OFF_GRID(iXTrans,iYTrans,iZTrans)))
					{
						glMaterialfv (GL_FRONT, GL_DIFFUSE, mat_diffuse_H);
						drawCell (iXTrans, iYTrans, iZTrans);
					}
					else
					if (O.get(MK_OFF_GRID(iXTrans, iYTrans, iZTrans)))
					{
						glMaterialfv (GL_FRONT, GL_DIFFUSE, mat_diffuse_O);
						drawCell (iXTrans, iYTrans, iZTrans);
					}
				}
			}
		}

		glMaterialfv (GL_FRONT, GL_DIFFUSE, mat_diffuse_I);
		drawCell (ROUND(disks[dsk].GetX()), ROUND(disks[dsk].GetY()), ROUND(disks[dsk].GetZ()));

		glDisable (GL_CULL_FACE);
		glDisable (GL_LIGHTING);

		// Txy
		for (iYTrans = 0; iYTrans < SIZE_Y; iYTrans++)
			for (iXTrans = 0; iXTrans < SIZE_X; iXTrans++)
			{
				if (!TxyN[MK_OFF(iYTrans, iXTrans, SIZE_X)].empty())
				{
					glBegin (GL_QUADS);
						glNormal3f (0.0, 0.0, 1.0);
						glVertex3f ( iXTrans   *STEP,  iYTrans   *STEP, 0);
						glVertex3f ( iXTrans   *STEP, (iYTrans+1)*STEP, 0);
						glVertex3f ((iXTrans+1)*STEP, (iYTrans+1)*STEP, 0);
						glVertex3f ((iXTrans+1)*STEP,  iYTrans   *STEP, 0);
					glEnd ();
				}
			}

		// Tyz
		for (iYTrans = 0; iYTrans < SIZE_Y; iYTrans++)
			for (iZTrans = 0; iZTrans < SIZE_Z; iZTrans++)
			{
				if (!TyzN[MK_OFF(iYTrans, iZTrans, SIZE_Z)].empty())
				{
					glBegin (GL_QUADS);
						glNormal3f (1.0, 0.0, 0.0);
						glVertex3f (0,  iYTrans   *STEP,  iZTrans   *STEP);
						glVertex3f (0,  iYTrans   *STEP, (iZTrans+1)*STEP);
						glVertex3f (0, (iYTrans+1)*STEP, (iZTrans+1)*STEP);
						glVertex3f (0, (iYTrans+1)*STEP,  iZTrans   *STEP);
					glEnd ();
				}
			}

		// Txz
		for (iZTrans = 0; iZTrans < SIZE_Z; iZTrans++)
			for (iXTrans = 0; iXTrans < SIZE_X; iXTrans++)
			{
				if (!TxzN[MK_OFF(iZTrans, iXTrans, SIZE_X)].empty())
				{
					glBegin (GL_QUADS);
						glNormal3f (0.0, 1.0, 0.0);
						glVertex3f ( iXTrans   *STEP, 0,  iZTrans   *STEP);
						glVertex3f ( iXTrans   *STEP, 0, (iZTrans+1)*STEP);
						glVertex3f ((iXTrans+1)*STEP, 0, (iZTrans+1)*STEP);
						glVertex3f ((iXTrans+1)*STEP, 0,  iZTrans   *STEP);
					glEnd ();
				}
			}

		glEnable (GL_LIGHTING);
		glEnable (GL_CULL_FACE);

		GLfloat mat_diffuse_viewpoint[] = { (float)1.0, (float)0.3, (float)0.3, (float)1.0 };
		glMaterialfv (GL_FRONT, GL_DIFFUSE, mat_diffuse_viewpoint);
		drawCell (vPos.x, vPos.y, vPos.z);

		glDisable (GL_LIGHT1);
		glDisable (GL_LIGHT0);
		glDisable (GL_LIGHTING);
	}

	glDisable (GL_CULL_FACE);
	glDisable (GL_DEPTH_TEST);

	reshape (WIN_WIDTH, WIN_HEIGHT + 50);

	// clear grid
	H.reset ();
	O.reset ();
	O.getFromBitset (Os);

	// determine number of visible objects
	for (iCt = 0; iCt < N_OBJS_DYN+N_OBJS_STC; iCt++)
	{
		if(PVS[iCt]) 
			iVisObjs++;

		if (PVSVF[iCt])
			iVisObjsVF++;
	}

	// clear text window
	glColor3f (0.0, 0.0, 0.0);
	glBegin (GL_QUADS);
		glVertex3f (        0, WIN_HEIGHT   , 0);
		glVertex3f (        0, WIN_HEIGHT+50, 0);
		glVertex3f (WIN_WIDTH, WIN_HEIGHT+50, 0);
		glVertex3f (WIN_WIDTH, WIN_HEIGHT   , 0);
	glEnd ();

	// draw text
	glColor3f (1.0, 1.0, 1.0);
	glBegin (GL_LINES);
		glVertex3f (        0, WIN_HEIGHT, 0);
		glVertex3f (WIN_WIDTH, WIN_HEIGHT, 0);
	glEnd ();

	glColor3f (1.0, 1.0, 1.0);
	if (current_date > 100)
	{
		fps_avg += FPS;
		fps_avg_ct++;

		printString (5, WIN_HEIGHT+2+5, "FPS: %.1f  FPS_AVG: %.1f  Frame #%d", FPS, fps_avg / fps_avg_ct, current_date);
	}
	else
		printString (5, WIN_HEIGHT+2+5, "FPS: %.2f  FPS_AVG: 0  Frame #%d", FPS, current_date);

	printString (5, WIN_HEIGHT+2+20, "X: %.1f  Y: %.1f  Z: %.1f  Roll: %.1f  Yaw: %.1f  Pitch: %.1f  FOV: %.1f", vPos.x, vPos.y, vPos.z, rotX, rotY, rotZ, fFOV);
	printString (5, WIN_HEIGHT+2+35, "PVS: %d   PVS w/ VF-culling only: %d   Exact vis.: %d  iDist: %d", iVisObjs, iVisObjsVF, iVisExact, iDist);

	if (hiddenGeometry)
		printString ((WIN_WIDTH/2) + (WIN_WIDTH/4) - 32, WIN_HEIGHT/2, "DISABLED");

	// next frame
	if (bDynamicScene)
	{
#ifdef USETBV
		for (iCt = 0; iCt < N_OBJS_DYN; iCt++)
			disks[iCt].Move ();
#endif
 	}

	memset (PVS, false, sizeof(bool)*(N_OBJS_DYN+N_OBJS_STC)); // clear PVS
	memset (PVSVF, false, sizeof(bool)*(N_OBJS_DYN+N_OBJS_STC)); // clear PVSVF

	current_date++;

	glutSwapBuffers ();
	glFinish ();
} // Vis3D::show

void reshape (int w, int h)
{
	glViewport (0, 0, (GLsizei) w, (GLsizei) h);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	gluOrtho2D (0, w, 0, h);

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	glTranslatef (0.375, 0.375, 0.0);
}

void display (void)
{
	MyVis->show ();
}

void motionFunc (int x, int y)
{
	if (bLeftButtonPressed)
	{
		MyVis->arcBall.LeftButtonDownMove (x, y);
	}
	if (bRightButtonPressed)
	{
		fEndTranslate = (float)y / 5.0;
		fWorldZTranslate = fStartTranslate - fEndTranslate;
	}
}

void mouse (int button, int state, int x, int y)
{
	if (button == GLUT_LEFT_BUTTON)
	{
		if (state == GLUT_DOWN)
		{
			MyVis->arcBall.LeftButtonDown (x, y);
			bLeftButtonPressed = TRUE;
		}
		else if (state == GLUT_UP)
		{
			MyVis->arcBall.LeftButtonUp ();
			bLeftButtonPressed = FALSE;
		}
	}
	else
	if (button == GLUT_RIGHT_BUTTON)
	{
		if (state == GLUT_DOWN)
		{
			fStartTranslate = fWorldZTranslate + ((float)y / 5.0);
			bRightButtonPressed = TRUE;
		}
		else if (state == GLUT_UP)
			bRightButtonPressed = FALSE;
	}

	glutPostRedisplay ();
}

void keyboard (unsigned char key, int x, int y)
{
	switch (key) {
	case 'x':
		if (MyVis->vPos.x > 0) 
			MyVis->vPos.x--;
		break;

	case 'X':
		if (MyVis->vPos.x < SIZE_X-1) 
			MyVis->vPos.x++;
		break;

	case 'y':
		if (MyVis->vPos.y > 0) 
			MyVis->vPos.y--;
		break;

	case 'Y':
		if (MyVis->vPos.y < SIZE_Y-1) 
			MyVis->vPos.y++;
		break;

	case 'z':
		if (MyVis->vPos.z > 0) 
			MyVis->vPos.z--;
		break;

	case 'Z':
		if (MyVis->vPos.z < SIZE_Z-1) 
			MyVis->vPos.z++;
		break;

	case '5':
		(MyVis->rotX < 359) ? MyVis->rotX++ : MyVis->rotX = 0;
		break;

	case '6':
		(MyVis->rotX > 0) ? MyVis->rotX-- : MyVis->rotX = 359;
		break;

	case '7':
		(MyVis->rotY < 359) ? MyVis->rotY++ : MyVis->rotY = 0;
		break;

	case '8':
		(MyVis->rotY > 0) ? MyVis->rotY-- : MyVis->rotY = 359;
		break;

	case '9':
		(MyVis->rotZ < 359) ? MyVis->rotZ++ : MyVis->rotZ = 0;
		break;

	case '0':
		(MyVis->rotZ > 0) ? MyVis->rotZ-- : MyVis->rotZ = 359;
		break;

	case '1':
		if (MyVis->fFOV < 179) 
			MyVis->fFOV++;
		break;
	
	case '2':
		if (MyVis->fFOV > 2) 
			MyVis->fFOV--;
		break;

	case '3':
		MyVis->bShowData = !MyVis->bShowData;
		break;

	case '4':
		MyVis->bDynamicScene = !MyVis->bDynamicScene;
		break;

	case 't':
		MyVis->fusionEnabled = !MyVis->fusionEnabled;
		break;

	case 'h':
		MyVis->hiddenGeometry = !MyVis->hiddenGeometry;
		break;

	case 'e':
		MyVis->bCalcExactVis = !MyVis->bCalcExactVis;
		break;

	case 27: // esc
		delete MyVis;
		exit(0);
		break;
	}

	XMATRIX mX, mY, mZ;

	MatrixIdentity(&MyVis->mView);
	MatrixMultiply(&MyVis->mView, MatrixRotationZ(&mZ, RAD(MyVis->rotZ)), &MyVis->mView);
	MatrixMultiply(&MyVis->mView, MatrixRotationX(&mX, RAD(MyVis->rotX)), &MyVis->mView);
	MatrixMultiply(&MyVis->mView, MatrixRotationY(&mY, RAD(MyVis->rotY)), &MyVis->mView);

	glutPostRedisplay ();
}

void idle (void)
{
	glutPostRedisplay ();
}

int main (int argc, char** argv)
{
	printf ("Dynamic Scene Occlusion Culling using a Regular Grid\n");
	printf ("by Harlen C. Batagelo (2001-2002)\n\n");
	printf ("Grid size: %dx%dx%d  Static objects: %d  Dynamic objects: %d\n\n", SIZE_X, SIZE_Y, 
			SIZE_Z, N_OBJS_STC, N_OBJS_DYN);
	printf ("Keys:\nARROWS   move viewpoint\n");
	printf ("HOME/END rotate\n");
	printf ("1        increase field of view\n");
	printf ("2        decrease field of view\n");
	printf ("3        toggle structure display mode\n");
	printf ("t        toggle occluder fusion\n");
	printf ("h        toggle scene rendering\n");
	printf ("e        toggle exact visibility calculation\n\n");

	printf ("Please wait...");

	MyVis = new Vis3D;

	glutInit (&argc, argv);
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGBA);
	glutInitWindowSize (WIN_WIDTH, WIN_HEIGHT + 50); 
	glutCreateWindow (argv[0]);
	glutSetWindowTitle ("3D dynamic visibility using a regular grid");
	MyVis->init ();
	glutDisplayFunc (display); 
	glutReshapeFunc (reshape);
	glutKeyboardFunc (keyboard);
	glutMouseFunc (mouse);
	glutMotionFunc (motionFunc);
	glutIdleFunc (idle);
	glutMainLoop ();

	delete MyVis;

	return 0;
}