//***************************************************************************
//
// TRAVERSAL.CPP
// by Harlen C. Batagelo
//
//***************************************************************************

#include <GL/glut.h>

#include "disk.h"
#include "common.h"

/////////////////////////////////////////////////////////////////////////////
// Assign a TBV to a given object
/////////////////////////////////////////////////////////////////////////////
void Vis3D::insertTBV (int disk)
{
	int iX0 = ROUND(disks[disk].GetTBVX());
	int iY0 = ROUND(disks[disk].GetTBVY());
	int iZ0 = ROUND(disks[disk].GetTBVZ());
	int iRadius = ROUND(disks[disk].GetExternalRadius() +
		         (float) (disks[disk].GetExpDate() - 
				          disks[disk].GetStartDate()) * 
			     SQUARE3 (disks[disk].GetVelocityX(),
				          disks[disk].GetVelocityY(),
						  disks[disk].GetVelocityZ()));

	// maximum radius
	if (iRadius > 15)
	{
		iRadius = 15;

		// set expiration date based on the new radius
		disks[disk].SetExpDate (ROUND(((iRadius - disks[disk].GetExternalRadius()) /
								SQUARE3(disks[disk].GetVelocityX(),
								        disks[disk].GetVelocityY(),
										disks[disk].GetVelocityZ())) + 
								disks[disk].GetStartDate()));
	}

	// radius must be smaller than the distance from the
	// object center to the viewer
	float fR = ROUND(SQUARE3(vPos.x-iX0, vPos.y-iY0, vPos.z-iZ0));
	if (SQUARE3(iRadius,iRadius,iRadius) >= fR)
	{
		iRadius = (int)(fR/sqrt(3.0))-1;

		if (iRadius < disks[disk].GetExternalRadius())
			iRadius = disks[disk].GetExternalRadius();

		// set expiration date based on the new radius
		disks[disk].SetExpDate (ROUND(((iRadius - disks[disk].GetExternalRadius()) /
								SQUARE3(disks[disk].GetVelocityX(),
								        disks[disk].GetVelocityY(),
									    disks[disk].GetVelocityZ())) + 
								disks[disk].GetStartDate()));
	}

	disks[disk].SetTBVRadius (iRadius);

	// top
	setLineTxy (iX0-iRadius, iX0+iRadius, iY0-iRadius, disk);
	setLineTyz (iZ0-iRadius, iZ0+iRadius, iY0-iRadius, disk);
	setLineTxz (iX0-iRadius, iX0+iRadius, iZ0-iRadius, disk);

	// bottom
	setLineTxy (iX0-iRadius, iX0+iRadius, iY0+iRadius, disk);
	setLineTyz (iZ0-iRadius, iZ0+iRadius, iY0+iRadius, disk);
	setLineTxz (iX0-iRadius, iX0+iRadius, iZ0+iRadius, disk);

	for (int iy = -iRadius; iy <= iRadius; iy++)
	{
		// left
		setTxy (iX0-iRadius, iY0+iy, disk);
		setTyz (iZ0-iRadius, iY0+iy, disk);
		setTxz (iX0-iRadius, iZ0+iy, disk);

		// right
		setTxy (iX0+iRadius, iY0+iy, disk);
		setTyz (iZ0+iRadius, iY0+iy, disk);
		setTxz (iX0+iRadius, iZ0+iy, disk);
	}
} // Vis3D::insertTBV

/////////////////////////////////////////////////////////////////////////////
// Remove the TBV of a given object
/////////////////////////////////////////////////////////////////////////////
void Vis3D::eraseTBV (int disk)
{
	int iX0 = ROUND(disks[disk].GetTBVX ());
	int iY0 = ROUND(disks[disk].GetTBVY ());
	int iZ0 = ROUND(disks[disk].GetTBVZ ());
	int iRadius = disks[disk].GetTBVRadius ();

	// top
	resetLineTxy (iX0-iRadius, iX0+iRadius, iY0-iRadius, disk);
	resetLineTyz (iZ0-iRadius, iZ0+iRadius, iY0-iRadius, disk);
	resetLineTxz (iX0-iRadius, iX0+iRadius, iZ0-iRadius, disk);

	// bottom
	resetLineTxy (iX0-iRadius, iX0+iRadius, iY0+iRadius, disk);
	resetLineTyz (iZ0-iRadius, iZ0+iRadius, iY0+iRadius, disk);
	resetLineTxz (iX0-iRadius, iX0+iRadius, iZ0+iRadius, disk);

	for (int iy = -iRadius; iy <= iRadius; iy++)
	{
		// left
		resetTxy (iX0-iRadius, iY0+iy, disk);
		resetTyz (iZ0-iRadius, iY0+iy, disk);
		resetTxz (iX0-iRadius, iZ0+iy, disk);

		// right
		resetTxy (iX0+iRadius, iY0+iy, disk);
		resetTyz (iZ0+iRadius, iY0+iy, disk);
		resetTxz (iX0+iRadius, iZ0+iy, disk);
	}
} // Vis3D::eraseTBV

int PVSTBV [N_OBJS_DYN]; // used internally by Vis3D::Add2PVS

/////////////////////////////////////////////////////////////////////////////
// Add the objects associated with the given cell to the PVS
/////////////////////////////////////////////////////////////////////////////
void Vis3D::Add2PVS (int iX, int iY, int iZ)
{
	// iterate over the list of objects
	if (!I[MK_OFF_GRID(iX+iOrX, iY+iOrY, iZ+iOrZ)].empty())
	{
		set_of_ids::iterator setIt;
		for (setIt = I[MK_OFF_GRID(iX+iOrX, iY+iOrY, iZ+iOrZ)].begin(); 
			 setIt != I[MK_OFF_GRID(iX+iOrX, iY+iOrY, iZ+iOrZ)].end(); setIt++)
			if (!PVS[*setIt])
			{
				if (!hiddenGeometry && !bCalcExactVis)
					disks[*setIt].DrawDisk (this);

				PVS[*setIt] = true;
			}
	}

#ifdef USETBV

	int iOffTxy = MK_OFF(iY+iOrY, iX+iOrX, SIZE_X),
		iOffTyz = MK_OFF(iY+iOrY, iZ+iOrZ, SIZE_Z),
		iOffTxz = MK_OFF(iZ+iOrZ, iX+iOrX, SIZE_X);

	// iterate over the list of TBVs
	if ((!TxyN[iOffTxy].empty() || !TyzN[iOffTyz].empty()) && 
		(!TxyN[iOffTxy].empty() || !TxzN[iOffTxz].empty()) && 
		(!TyzN[iOffTyz].empty() || !TxzN[iOffTxz].empty()))
	{
		int i = 0;

		set_of_ids::iterator setIt;
		if (TxyN[iOffTxy].size() < TyzN[iOffTyz].size())
		{
			if (TxyN[iOffTxy].size() < TxzN[iOffTxz].size())
			{
				// TxyN
				for (setIt = TxyN[iOffTxy].begin(); setIt != TxyN[iOffTxy].end(); setIt++)
				{
					if ((Txy[*setIt].get(iOffTxy) || Tyz[*setIt].get(iOffTyz)) && 
						(Txy[*setIt].get(iOffTxy) || Txz[*setIt].get(iOffTxz)) && 
						(Tyz[*setIt].get(iOffTyz) || Txz[*setIt].get(iOffTxz)))
						PVSTBV[i++] = *setIt;
				}
			}
			else
			{
				// TxzN
				for (setIt = TxzN[iOffTxz].begin(); setIt != TxzN[iOffTxz].end(); setIt++)
				{
					if ((Txy[*setIt].get(iOffTxy) || Tyz[*setIt].get(iOffTyz)) && 
						(Txy[*setIt].get(iOffTxy) || Txz[*setIt].get(iOffTxz)) && 
						(Tyz[*setIt].get(iOffTyz) || Txz[*setIt].get(iOffTxz)))
						PVSTBV[i++] = *setIt;
				}
			}
		}
		else
		{
			if (TyzN[iOffTyz].size() < TxzN[iOffTxz].size())
			{
				// TyzN
				for (setIt = TyzN[iOffTyz].begin(); setIt != TyzN[iOffTyz].end(); setIt++)
				{
					if ((Txy[*setIt].get(iOffTxy) || Tyz[*setIt].get(iOffTyz)) && 
						(Txy[*setIt].get(iOffTxy) || Txz[*setIt].get(iOffTxz)) && 
						(Tyz[*setIt].get(iOffTyz) || Txz[*setIt].get(iOffTxz)))
						PVSTBV[i++] = *setIt;
				}
			}
			else
			{
				// TxzN
				for (setIt = TxzN[iOffTxz].begin(); setIt != TxzN[iOffTxz].end(); setIt++)
				{
					if ((Txy[*setIt].get(iOffTxy) || Tyz[*setIt].get(iOffTyz)) && 
						(Txy[*setIt].get(iOffTxy) || Txz[*setIt].get(iOffTxz)) && 
						(Tyz[*setIt].get(iOffTyz) || Txz[*setIt].get(iOffTxz)))
						PVSTBV[i++] = *setIt;
				}
			}
		}

		for(int iCt = 0; iCt < i; iCt++)
		{
			putIntoGrid (PVSTBV[iCt], false, false);
			eraseTBV (PVSTBV[iCt]);
			disks[PVSTBV[iCt]].SetTBVState (false);
		}
	}
#endif
} // Vis3D::Add2PVS

/////////////////////////////////////////////////////////////////////////////
// Look for occluders and occludees. Return if tested cell is hidden
/////////////////////////////////////////////////////////////////////////////
void Vis3D::TestCell (int iX, int iY, int iZ)
{
	if ((abs(iX) > iDist) || (abs(iY) > iDist) || (abs(iZ) > iDist)) // test bounds
		return;

	int iXTrans, iYTrans;

	// Look for opaque or hidden cells
	if (!H.get(MK_OFF_GRID(iX+iOrX,iY+iOrY,iZ+iOrZ)))
	{
		if (O.get(MK_OFF_GRID(iX+iOrX,iY+iOrY,iZ+iOrZ)))
		{
			if (iZ == iDist)
			{
				// front plane (x' = -x) (y' = -y) (z' = z++)
				iXTrans = -iX+iDist;
				iYTrans = -iY+iDist;

				if (iYTrans < iCPMinYF) iCPMinYF = iYTrans; // determine min/max
				if (iYTrans > iCPMaxYF) iCPMaxYF = iYTrans;
				if (iXTrans < iCPMinXF) iCPMinXF = iXTrans;
				if (iXTrans > iCPMaxXF) iCPMaxXF = iXTrans;

				iMatrixF[MK_OFF(iYTrans,iXTrans,iDistSize)] = 1;
				bMatrixF = 1;
			}
			else if (iZ == -iDist)
			{
				// back plane (x' = x) (y' = -y) (z' = z--)
				iXTrans =  iX+iDist;
				iYTrans = -iY+iDist;

				if (iYTrans < iCPMinYB) iCPMinYB = iYTrans; // determine min/max
				if (iYTrans > iCPMaxYB) iCPMaxYB = iYTrans;
				if (iXTrans < iCPMinXB) iCPMinXB = iXTrans;
				if (iXTrans > iCPMaxXB) iCPMaxXB = iXTrans;

				iMatrixB[MK_OFF(iYTrans,iXTrans,iDistSize)] = 1;
				bMatrixB = 1;
			}

			if (iX == iDist)
			{
				// left plane (x' = z) (y' = -y) (z' = x++)
				iXTrans =  iZ+iDist;
				iYTrans = -iY+iDist;

				if (iYTrans < iCPMinYL) iCPMinYL = iYTrans; // determine min/max
				if (iYTrans > iCPMaxYL) iCPMaxYL = iYTrans;
				if (iXTrans < iCPMinXL) iCPMinXL = iXTrans;
				if (iXTrans > iCPMaxXL) iCPMaxXL = iXTrans;

				iMatrixL[MK_OFF(iYTrans,iXTrans,iDistSize)] = 1;
				bMatrixL = 1;
			}
			else if (iX == -iDist)
			{
				// right plane (x' = -z) (y' = -y) (z' = x--)
				iXTrans = -iZ+iDist;
				iYTrans = -iY+iDist;

				if (iYTrans < iCPMinYR) iCPMinYR = iYTrans; // determine min/max
				if (iYTrans > iCPMaxYR) iCPMaxYR = iYTrans;
				if (iXTrans < iCPMinXR) iCPMinXR = iXTrans;
				if (iXTrans > iCPMaxXR) iCPMaxXR = iXTrans;

				iMatrixR[MK_OFF(iYTrans,iXTrans,iDistSize)] = 1;
				bMatrixR = 1;
			}

			if (iY == iDist)
			{
				// top plane (x' = -x) (y' = z) (z' = y++)
				iXTrans = -iX+iDist;
				iYTrans =  iZ+iDist;

				if (iYTrans < iCPMinYT) iCPMinYT = iYTrans; // determine min/max
				if (iYTrans > iCPMaxYT) iCPMaxYT = iYTrans;
				if (iXTrans < iCPMinXT) iCPMinXT = iXTrans;
				if (iXTrans > iCPMaxXT) iCPMaxXT = iXTrans;

				iMatrixT[MK_OFF(iYTrans,iXTrans,iDistSize)] = 1;
				bMatrixT = 1;
			}
			else if (iY == -iDist)
			{
				// down plane (x' = -x) (y' = -z) (z' = y--)
				iXTrans = -iX+iDist;
				iYTrans = -iZ+iDist;

				if (iYTrans < iCPMinYD) iCPMinYD = iYTrans; // determine min/max
				if (iYTrans > iCPMaxYD) iCPMaxYD = iYTrans;
				if (iXTrans < iCPMinXD) iCPMinXD = iXTrans;
				if (iXTrans > iCPMaxXD) iCPMaxXD = iXTrans;

				iMatrixD[MK_OFF(iYTrans,iXTrans,iDistSize)] = 1;
				bMatrixD = 1;
			}
		}
		else
			// flag for premature end of traversal
			bPrematureEnd = false;

		// Add the objects associated with the given cell to the PVS
		Add2PVS (iX, iY, iZ);
	}

	// Add the objects associated with the given cell to the PVSVF

	// iterate over the list of objects
	if (!I[MK_OFF_GRID(iX+iOrX, iY+iOrY, iZ+iOrZ)].empty())
	{
		set_of_ids::iterator setIt;
		for (setIt = I[MK_OFF_GRID(iX+iOrX, iY+iOrY, iZ+iOrZ)].begin(); 
			 setIt != I[MK_OFF_GRID(iX+iOrX, iY+iOrY, iZ+iOrZ)].end(); setIt++)
			if (!PVSVF[*setIt]) PVSVF[*setIt] = true;
	} 
} // Vis3D::TestCell

//
// The following procedures are applied only to the region
// pointed in the diagram below:
//
//  X ^- - - - - - - - - - -|
//    | \                 / 
//    |   \     here    /   |
//    |     \         /     
//    |       \     /       |
//    |         \ /         
//    |         / \         |
//    |       /     \       
//    |     /         \     |
//    |   /             \   
//    | /                 \ | 
//    *---------------------> Z
//  any Y
// 

void Vis3D::YXTraversalNeg (int iX, int iY, int iZ, int iSeedXOFB, int iSeedYOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	int iStopYNeg, iStopYPos; // stop points
	int iInitX, iInitY, iInitZ; // seed cell

	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// put seed cell within Y bounds
	if (iSeedYOFB)
	{
		if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
		if (iInitY+iOrY < 0)		iInitY = -iOrY;
	}

	// determine Y extents
	// -y
	if (isInsideVFBnd (iInitX+iOrX, -abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iYOFB)
			iStopYNeg = -abs(iInitX);
		else
			iStopYNeg = vOFB.y;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iYOFB)
			iStopYNeg = vBound.y;
		else
			iStopYNeg = vOFB.y;
	}

	// +y
	if (isInsideVFBnd (iInitX+iOrX, abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iYOFB)
			iStopYPos = abs(iInitX);
		else
			iStopYPos = vOFB.y;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iYOFB)
			iStopYPos = vBound.y;
		else
			iStopYPos = vOFB.y;
	}

	for (iY = iInitY-1; iY >= iStopYNeg; iY--)
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iInitZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopX;
		(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

		if (!iSeedXOFB)
		{
			for (iX = iInitX; iX >= iStopX; iX--)
				TestCell (iX, iY, iInitZ);
		}
		else if (vBound.x+iOrX < SIZE_X) // added
		{
			for (iX = SIZE_XM1-iOrX; iX >= iStopX; iX--)
				TestCell (iX, iY, iInitZ);
		}
	}


	for (iY = iInitY; iY <= iStopYPos; iY++)
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iInitZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopX;
		(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

		if (!iSeedXOFB)
		{
			for (iX = iInitX; iX >= iStopX; iX--)
				TestCell (iX, iY, iInitZ);
		}
		else if (vBound.x+iOrX < SIZE_X) // added
		{
			for (iX = SIZE_XM1-iOrX; iX >= iStopX; iX--)
				TestCell (iX, iY, iInitZ);
		}
	}
}

void Vis3D::Y1TraversalNeg (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	BOOL bTurnNeg, bTurnPos; // make a 90deg turn during the traversal? 
	int iStopYNeg,  // stop points
		iStopYPos;
	int iInitX, iInitY, iInitZ; // seed cell
	int iCpyInitY = iY;

	int iOFBNeg, iOFBStopYNeg,
		iOFBPos, iOFBStopYPos;

	bTurnNeg = bTurnPos = FALSE;
	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// determine Y extents
	// -y
	if (isInsideVFBnd (iInitX+iOrX, -abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopYNeg = -abs(iInitX);
		bTurnNeg = TRUE;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopYNeg = vBound.y;
		bTurnNeg = FALSE;
	}
	iOFBNeg = iYOFB;
	iOFBStopYNeg = vOFB.y;

	// +y
	if (isInsideVFBnd (iInitX+iOrX, abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopYPos = abs(iInitX);
		bTurnPos = TRUE;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopYPos = vBound.y;
		bTurnPos = FALSE;
	}
	iOFBPos = iYOFB;
	iOFBStopYPos = vOFB.y;

	// Y negative traversal

	if (!iOFBNeg)
	{
		if (!iSeedOFB)
		{
			if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
			if (iInitY+iOrY < 0)		iInitY = -iOrY;

			for (iY = iInitY; iY >= iStopYNeg; iY--)
				TestCell (iInitX, iY, iInitZ);
		}

		if (bTurnNeg)
		{
			// determine stop point (x extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iStopYNeg, iZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopX;
			(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

			if  (!iSeedOFB)
			{
				for (iX = iInitX; iX >= iStopX; iX--)
					TestCell (iX, iStopYNeg, iZ);
			}
			else if (vBound.x+iOrX < SIZE_X) // added
			{
				for (iX = SIZE_XM1-iOrX; iX >= iStopX; iX--)
					TestCell (iX, iStopYNeg, iZ);
			}
		}
	}
	else
	{
		if (!iSeedOFB)
		{
			if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
			if (iInitY+iOrY < 0)		iInitY = -iOrY;

			for (iY = iInitY-1; iY >= iOFBStopYNeg; iY--)
				TestCell (iInitX, iY, iInitZ);
		}
	}

	// Y positive traversal
	iInitY = iCpyInitY;

	if (!iOFBPos)
	{
		if (!iSeedOFB)
		{
			if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
			if (iInitY+iOrY < 0)		iInitY = -iOrY;

			for (iY = iInitY; iY <= iStopYPos; iY++)
				TestCell (iInitX, iY, iInitZ);
		}

		if (bTurnPos)
		{
			// determine stop point (x extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iStopYPos, iZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopX;
			(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

			if (!iSeedOFB)
			{
				for (iX = iInitX; iX >= iStopX; iX--)
					TestCell (iX, iStopYPos, iZ);
			}	
			else if (vBound.x+iOrX < SIZE_X) // added
			{
				for (iX = SIZE_XM1-iOrX; iX >= iStopX; iX--)
					TestCell (iX, iStopYPos, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
		if (iInitY+iOrY < 0)		iInitY = -iOrY;

		for (iY = iInitY; iY <= iOFBStopYPos; iY++) 
			TestCell (iInitX, iY, iInitZ);
	}
}

//
// The following procedures are applied only to the region
// pointed in the diagram below:
//
//  X ^- - - - - - - - - - -|
//    | \                 / 
//    |   \             /   |
//    |     \         /     
//    |       \     /       |
//    |         \ /         
//    |         / \         |
//    |       /     \       
//    |     /         \     |
//    |   /     here    \   
//    | /                 \ | 
//    *---------------------> Z
//  any Y
// 

void Vis3D::YXTraversalPos (int iX, int iY, int iZ, int iSeedXOFB, int iSeedYOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	int iStopYNeg, iStopYPos; // stop points
	int iInitX, iInitY, iInitZ; // seed cell

	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// put seed cell within Y bounds
	if (iSeedYOFB)
	{
		if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
		if (iInitY+iOrY < 0)		iInitY = -iOrY;
	}

	// determine Y extents
	// -y
	if (isInsideVFBnd (iInitX+iOrX, -abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iYOFB)
			iStopYNeg = -abs(iInitX);
		else
			iStopYNeg = vOFB.y;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iYOFB)
			iStopYNeg = vBound.y;
		else
			iStopYNeg = vOFB.y;
	}

	// +y
	if (isInsideVFBnd (iInitX+iOrX, abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iYOFB)
			iStopYPos = abs(iInitX);
		else
			iStopYPos = vOFB.y;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iYOFB)
			iStopYPos = vBound.y;
		else
			iStopYPos = vOFB.y;
	}

	for (iY = iInitY-1; iY >= iStopYNeg; iY--)
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iInitZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopX;
		(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

		if (!iSeedXOFB)
		{
			for (iX = iInitX; iX <= iStopX; iX++)
				TestCell (iX, iY, iInitZ);
		}
		else if (vBound.x+iOrX >= 0) // added
		{
			for (iX = -iOrX; iX <= iStopX; iX++)
				TestCell (iX, iY, iInitZ);
		}
	}

	for (iY = iInitY; iY <= iStopYPos; iY++)
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iInitZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopX;
		(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

		if (!iSeedXOFB)
		{
			for (iX = iInitX; iX <= iStopX; iX++)
				TestCell (iX, iY, iInitZ);
		}
		else if (vBound.x+iOrX >= 0) // added
		{
			for (iX = -iOrX; iX <= iStopX; iX++)
				TestCell (iX, iY, iInitZ);
		}
	}
}

void Vis3D::Y1TraversalPos (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	BOOL bTurnNeg, bTurnPos; // make a 90deg turn during the traversal? 
	int iStopYNeg, // stop points
		iStopYPos;
	int iInitX, iInitY, iInitZ; // seed cell
	int iCpyInitY = iY;

	int iOFBNeg, iOFBStopYNeg,
		iOFBPos, iOFBStopYPos;

	bTurnNeg = bTurnPos = FALSE;
	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// determine Y extents
	// -y
	if (isInsideVFBnd (iInitX+iOrX, -abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopYNeg = -abs(iInitX);
		bTurnNeg = TRUE;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopYNeg = vBound.y;
		bTurnNeg = FALSE;
	}
	iOFBNeg = iYOFB;
	iOFBStopYNeg = vOFB.y;

	// +y
	if (isInsideVFBnd (iInitX+iOrX, abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopYPos = abs(iInitX);
		bTurnPos = TRUE;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopYPos = vBound.y;
		bTurnPos = FALSE;
	}
	iOFBPos = iYOFB;
	iOFBStopYPos = vOFB.y;

	// Y negative traversal

	if (!iOFBNeg)
	{
		if (!iSeedOFB)
		{
			if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
			if (iInitY+iOrY < 0)		iInitY = -iOrY;

			for (iY = iInitY; iY >= iStopYNeg; iY--)
				TestCell (iInitX, iY, iInitZ);
		}

		if (bTurnNeg)
		{
			// determine stop point (x extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iStopYNeg, iZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopX;
			(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

			if (!iSeedOFB)
			{
				for (iX = iInitX; iX <= iStopX; iX++)
					TestCell (iX, iStopYNeg, iZ);
			}
			else if (vBound.x+iOrX >= 0) // added
			{
				for (iX = -iOrX; iX <= iStopX; iX++)
					TestCell (iX, iStopYNeg, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
		if (iInitY+iOrY < 0)		iInitY = -iOrY;

		for (iY = iInitY-1; iY >= iOFBStopYNeg; iY--)
			TestCell (iInitX, iY, iInitZ);
	}

	// Y positive traversal
	iInitY = iCpyInitY;

	if (!iOFBPos)
	{
		if (!iSeedOFB)
		{
			if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
			if (iInitY+iOrY < 0)		iInitY = -iOrY;

			for (iY = iInitY; iY <= iStopYPos; iY++)
				TestCell (iInitX, iY, iInitZ);
		}

		if (bTurnPos)
		{
			// determine stop point (x extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iStopYPos, iZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopX;
			(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

			if (!iSeedOFB)
			{
				for (iX = iInitX; iX <= iStopX; iX++)
					TestCell (iX, iStopYPos, iZ);
			}
			else if (vBound.x+iOrX >= 0) // added
			{
				for (iX = -iOrX; iX <= iStopX; iX++)
					TestCell (iX, iStopYPos, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
		if (iInitY+iOrY < 0)		iInitY = -iOrY;

		for (iY = iInitY; iY <= iOFBStopYPos; iY++) 
			TestCell (iInitX, iY, iInitZ);
	} 
}

//
// The following procedures are applied only to the region
// pointed in the diagram below:
//
//  X ^- - - - - - - - - - -|
//    | \                 / 
//    |   \             /   |
//    |     \         /     
//    |       \     /       |
//    |         \ /   here 
//    |         / \         |
//    |       /     \       
//    |     /         \     |
//    |   /             \   
//    | /                 \ | 
//    *---------------------> Z
//  any Y
// 

void Vis3D::YZTraversalNeg (int iX, int iY, int iZ, int iSeedYOFB, int iSeedZOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	int iStopYNeg, iStopYPos; // stop points
	int iInitX, iInitY, iInitZ; // seed cell

	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// put seed cell within Y bounds
	if (iSeedYOFB)
	{
		if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
		if (iInitY+iOrY < 0)		iInitY = -iOrY;
	}

	// determine Y extents
	// -y
	if (isInsideVFBnd (iInitX+iOrX, -abs(iInitZ)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iYOFB)
			iStopYNeg = -abs(iInitZ);
		else
			iStopYNeg = vOFB.y;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iYOFB)
			iStopYNeg = vBound.y;
		else
			iStopYNeg = vOFB.y;
	}

	// +y
	if (isInsideVFBnd (iInitX+iOrX, abs(iInitZ)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iYOFB)
			iStopYPos = abs(iInitZ);
		else
			iStopYPos = vOFB.y;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iYOFB)
			iStopYPos = vBound.y;
		else
			iStopYPos = vOFB.y;
	}

	for (iY = iInitY-1; iY >= iStopYNeg; iY--)
	{
		// determine stop point (z extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iInitZ), VECTOR3(0, 0, -1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopZ;
		(iZOFB) ? iStopZ = vOFB.z : iStopZ = vBound.z;

		if (!iSeedZOFB)
		{
			for (iZ = iInitZ; iZ >= iStopZ; iZ--)
				TestCell (iInitX, iY, iZ);
		}
		else if (vBound.z+iOrZ < SIZE_Z)
		{
			for (iZ = SIZE_ZM1-iOrZ; iZ >= iStopZ; iZ--)
				TestCell (iInitX, iY, iZ);
		}
	}


	for (iY = iInitY; iY <= iStopYPos; iY++)
	{
		// determine stop point (z extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iInitZ), VECTOR3(0, 0, -1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopZ;
		(iZOFB) ? iStopZ = vOFB.z : iStopZ = vBound.z;

		if (!iSeedZOFB)
		{
			for (iZ = iInitZ; iZ >= iStopZ; iZ--)
				TestCell (iInitX, iY, iZ);
		}
		else if (vBound.z+iOrZ < SIZE_Z) // added
		{
			for (iZ = SIZE_ZM1-iOrZ; iZ >= iStopZ; iZ--)
				TestCell (iInitX, iY, iZ);
		}
	}
}

void Vis3D::Y2TraversalNeg (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	BOOL bTurnNeg, bTurnPos; // make a 90deg turn during the traversal? 
	int iStopYNeg,  // stop points
		iStopYPos;
	int iInitX, iInitY, iInitZ; // seed cell
	int iCpyInitY = iY;

	int iOFBNeg, iOFBStopYNeg,
		iOFBPos, iOFBStopYPos;

	bTurnNeg = bTurnPos = FALSE;
	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// determine Y extents
	// -y
	if (isInsideVFBnd (iInitX+iOrX, -abs(iInitZ)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopYNeg = -abs(iInitZ);
		bTurnNeg = TRUE;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopYNeg = vBound.y;
		bTurnNeg = FALSE;
	}
	iOFBNeg = iYOFB;
	iOFBStopYNeg = vOFB.y;

	// +y
	if (isInsideVFBnd (iInitX+iOrX, abs(iInitZ)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopYPos = abs(iInitZ);
		bTurnPos = TRUE;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopYPos = vBound.y;
		bTurnPos = FALSE;
	}
	iOFBPos = iYOFB;
	iOFBStopYPos = vOFB.y;

	// Y negative traversal

	if (!iOFBNeg)
	{
		if (!iSeedOFB)
		{
			if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
			if (iInitY+iOrY < 0)		iInitY = -iOrY;

			for (iY = iInitY; iY >= iStopYNeg; iY--)
				TestCell (iInitX, iY, iInitZ);
		}

		if (bTurnNeg)
		{
			// determine stop point (z extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iX, iStopYNeg, iInitZ), VECTOR3(0, 0, -1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopZ;
			(iZOFB) ? iStopZ = vOFB.z : iStopZ = vBound.z;

			if  (!iSeedOFB)
			{
				for (iZ = iInitZ; iZ >= iStopZ; iZ--)
					TestCell (iX, iStopYNeg, iZ);
			}
			else if (vBound.z+iOrZ < SIZE_Z) // added
			{
				for (iZ = SIZE_ZM1-iOrZ; iZ >= iStopZ; iZ--)
					TestCell (iX, iStopYNeg, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
		if (iInitY+iOrY < 0)		iInitY = -iOrY;

		for (iY = iInitY-1; iY >= iOFBStopYNeg; iY--)
			TestCell (iInitX, iY, iInitZ);
	}

	// Y positive traversal
	iInitY = iCpyInitY;

	if (!iOFBPos)
	{
		if (!iSeedOFB)
		{
			if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
			if (iInitY+iOrY < 0)		iInitY = -iOrY;

			for (iY = iInitY; iY <= iStopYPos; iY++)
				TestCell (iInitX, iY, iInitZ);
		}

		if (bTurnPos)
		{
			// determine stop point (z extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iX, iStopYPos, iInitZ), VECTOR3(0, 0, -1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopZ;
			(iZOFB) ? iStopZ = vOFB.z : iStopZ = vBound.z;

			if (!iSeedOFB)
			{
				for (iZ = iInitZ; iZ >= iStopZ; iZ--)
					TestCell (iX, iStopYPos, iZ);
			}	
			else if (vBound.z+iOrZ < SIZE_Z) // added
			{
				for (iZ = SIZE_ZM1-iOrZ; iZ >= iStopZ; iZ--)
					TestCell (iX, iStopYPos, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
		if (iInitY+iOrY < 0)		iInitY = -iOrY;

		for (iY = iInitY; iY <= iOFBStopYPos; iY++) 
			TestCell (iInitX, iY, iInitZ);
	}
}

//
// The following procedures are applied only to the region
// pointed in the diagram below:
//
//  X ^- - - - - - - - - - -|
//    | \                 / 
//    |   \             /   |
//    |     \         /     
//    |       \     /       |
//    |  here   \ /         
//    |         / \         |
//    |       /     \       
//    |     /         \     |
//    |   /             \   
//    | /                 \ | 
//    *---------------------> Z
//  any Y
// 

void Vis3D::YZTraversalPos (int iX, int iY, int iZ, int iSeedYOFB, int iSeedZOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	int iStopYNeg, iStopYPos; // stop points
	int iInitX, iInitY, iInitZ; // seed cell

	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// put seed cell within Y bounds
	if (iSeedYOFB)
	{
		if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
		if (iInitY+iOrY < 0)		iInitY = -iOrY;
	}

	// determine Y extents
	// -y
	if (isInsideVFBnd (iInitX+iOrX, -abs(iInitZ)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iYOFB)
			iStopYNeg = -abs(iInitZ);
		else
			iStopYNeg = vOFB.y;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iYOFB)
			iStopYNeg = vBound.y;
		else
			iStopYNeg = vOFB.y;
	}

	// +y
	if (isInsideVFBnd (iInitX+iOrX, abs(iInitZ)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iYOFB)
			iStopYPos = abs(iInitZ);
		else
			iStopYPos = vOFB.y;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iX, iInitY, iInitZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iYOFB)
			iStopYPos = vBound.y;
		else
			iStopYPos = vOFB.y;
	}

	for (iY = iInitY-1; iY >= iStopYNeg; iY--)
	{
		// determine stop point (z extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iInitZ), VECTOR3(0, 0, +1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopZ;
		(iZOFB) ? iStopZ = vOFB.z : iStopZ = vBound.z;

		if (!iSeedZOFB)
		{
			for (iZ = iInitZ; iZ <= iStopZ; iZ++)
				TestCell (iInitX, iY, iZ);
		}
		else if (vBound.z+iOrZ >= 0) // added
		{
			for (iZ = -iOrZ; iZ <= iStopZ; iZ++)
				TestCell (iInitX, iY, iZ);
		}
	}

	for (iY = iInitY; iY <= iStopYPos; iY++)
	{
		// determine stop point (z extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iInitZ), VECTOR3(0, 0, +1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopZ;
		(iZOFB) ? iStopZ = vOFB.z : iStopZ = vBound.z;

		if (!iSeedZOFB)
		{
			for (iZ = iInitZ; iZ <= iStopZ; iZ++)
				TestCell (iInitX, iY, iZ);
		}
		else if (vBound.z+iOrZ >= 0) // added
		{
			for (iZ = -iOrZ; iZ <= iStopZ; iZ++)
				TestCell (iInitX, iY, iZ);
		}
	}
}

void Vis3D::Y2TraversalPos (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	BOOL bTurnNeg, bTurnPos; // make a 90deg turn during the traversal? 
	int iStopYNeg, // stop points
		iStopYPos;
	int iInitX, iInitY, iInitZ; // seed cell
	int iCpyInitY = iY;

	int iOFBNeg, iOFBStopYNeg,
		iOFBPos, iOFBStopYPos;

	bTurnNeg = bTurnPos = FALSE;
	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// determine Y extents
	// -y
	if (isInsideVFBnd (iInitX+iOrX, -abs(iInitZ)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopYNeg = -abs(iInitZ);
		bTurnNeg = TRUE;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopYNeg = vBound.y;
		bTurnNeg = FALSE;
	}
	iOFBNeg = iYOFB;
	iOFBStopYNeg = vOFB.y;

	// +y
	if (isInsideVFBnd (iInitX+iOrX, abs(iInitZ)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopYPos = abs(iInitZ);
		bTurnPos = TRUE;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopYPos = vBound.y;
		bTurnPos = FALSE;
	}
	iOFBPos = iYOFB;
	iOFBStopYPos = vOFB.y;

	// Y negative traversal

	if (!iOFBNeg)
	{
		if (!iSeedOFB)
		{
			if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
			if (iInitY+iOrY < 0)		iInitY = -iOrY;

			for (iY = iInitY; iY >= iStopYNeg; iY--)
				TestCell (iInitX, iY, iInitZ);
		}

		if (bTurnNeg)
		{
			// determine stop point (z extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iX, iStopYNeg, iInitZ), VECTOR3(0, 0, +1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopZ;
			(iZOFB) ? iStopZ = vOFB.z : iStopZ = vBound.z;

			if (!iSeedOFB)
			{
				for (iZ = iInitZ; iZ <= iStopZ; iZ++)
					TestCell (iX, iStopYNeg, iZ);
			}
			else if (vBound.z+iOrZ >= 0) // added
			{
				for (iZ = -iOrZ; iZ <= iStopZ; iZ++)
					TestCell (iX, iStopYNeg, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
		if (iInitY+iOrY < 0)		iInitY = -iOrY;

		for (iY = iInitY-1; iY >= iOFBStopYNeg; iY--)
			TestCell (iInitX, iY, iInitZ);
	}

	// Y positive traversal
	iInitY = iCpyInitY;

	if (!iOFBPos)
	{
		if (!iSeedOFB)
		{
			if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
			if (iInitY+iOrY < 0)		iInitY = -iOrY;

			for (iY = iInitY; iY <= iStopYPos; iY++)
				TestCell (iInitX, iY, iInitZ);
		}

		if (bTurnPos)
		{
			// determine stop point (z extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iX, iStopYPos, iInitZ), VECTOR3(0, 0, +1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopZ;
			(iZOFB) ? iStopZ = vOFB.z : iStopZ = vBound.z;

			if (!iSeedOFB)
			{
				for (iZ = iInitZ; iZ <= iStopZ; iZ++)
					TestCell (iX, iStopYPos, iZ);
			}
			else if (vBound.z+iOrZ >= 0) // added
			{
				for (iZ = -iOrZ; iZ <= iStopZ; iZ++)
					TestCell (iX, iStopYPos, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
		if (iInitY+iOrY < 0)		iInitY = -iOrY;

		for (iY = iInitY; iY <= iOFBStopYPos; iY++) 
			TestCell (iInitX, iY, iInitZ);
	}
}

//
// The following procedures are applied only to the region
// pointed in the diagram below:
//
//  Y ^- - - - - - - - - - -|
//    | \                 / 
//    |   \     here    /   |
//    |     \         /     
//    |       \     /       |
//    |         \ /         
//    |         / \         |
//    |       /     \       
//    |     /         \     |
//    |   /             \   
//    | /                 \ | 
//    *---------------------> Z
//  any X
// 

void Vis3D::XYTraversalNeg (int iX, int iY, int iZ, int iSeedXOFB, int iSeedYOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	int iStopXNeg, iStopXPos; // stop points
	int iInitX, iInitY, iInitZ; // seed cell

	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// put seed cell within X bounds
	if (iSeedXOFB)
	{
		if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
		if (iInitX+iOrX < 0)		iInitX = -iOrX;
	}

	// determine X extents
	// -x
	if (isInsideVFBnd (-abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iXOFB)
			iStopXNeg = -abs(iInitY);
		else
			iStopXNeg = vOFB.x;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iXOFB)
			iStopXNeg = vBound.x;
		else
			iStopXNeg = vOFB.x;
	}

	// +x
	if (isInsideVFBnd (abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iXOFB)
			iStopXPos = abs(iInitY);
		else
			iStopXPos = vOFB.x;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iXOFB)
			iStopXPos = vBound.x;
		else
			iStopXPos = vOFB.x;
	}

	for (iX = iInitX-1; iX >= iStopXNeg; iX--)
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopY;
		(iYOFB) ? iStopY = vOFB.y :	iStopY = vBound.y;

		if (!iSeedYOFB)
		{
			for (iY = iInitY; iY >= iStopY; iY--)
				TestCell (iX, iY, iInitZ);
		}
		else if (vBound.y+iOrY < SIZE_Y) // added
		{
			for (iY = SIZE_YM1-iOrY; iY >= iStopY; iY--)
				TestCell (iX, iY, iInitZ);
		}
	}


	for (iX = iInitX; iX <= iStopXPos; iX++)
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopY;
		(iYOFB) ? iStopY = vOFB.y :	iStopY = vBound.y;

		if (!iSeedYOFB)
		{
			for (iY = iInitY; iY >= iStopY; iY--)
				TestCell (iX, iY, iInitZ);
		}
		else if (vBound.y+iOrY < SIZE_Y) // added
		{
			for (iY = SIZE_YM1-iOrY; iY >= iStopY; iY--)
				TestCell (iX, iY, iInitZ);
		}
	}
}

void Vis3D::XTraversalNeg (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	BOOL bTurnNeg, bTurnPos; // make a 90deg turn during the traversal? 
	int iStopXNeg,  // stop points
		iStopXPos;
	int iInitX, iInitY, iInitZ; // seed cell
	int iCpyInitX = iX;

	int iOFBNeg, iOFBStopXNeg,
		iOFBPos, iOFBStopXPos;

	bTurnNeg = bTurnPos = FALSE;
	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// determine X extents
	// -x
	if (isInsideVFBnd (-abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopXNeg = -abs(iInitY);
		bTurnNeg = TRUE;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopXNeg = vBound.x;
		bTurnNeg = FALSE;
	}
	iOFBNeg = iXOFB;
	iOFBStopXNeg = vOFB.x;

	// +x
	if (isInsideVFBnd (abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopXPos = abs(iInitY);
		bTurnPos = TRUE;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopXPos = vBound.x;
		bTurnPos = FALSE;
	}
	iOFBPos = iXOFB;
	iOFBStopXPos = vOFB.x;

	// X negative traversal

	if (!iOFBNeg)
	{
		if (!iSeedOFB)
		{
			if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
			if (iInitX+iOrX < 0)		iInitX = -iOrX;

			for (iX = iInitX; iX >= iStopXNeg; iX--)
				TestCell (iX, iInitY, iInitZ);
		}

		if (bTurnNeg)
		{
			// determine stop point (y extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iStopXNeg, iInitY, iZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopY;
			(iYOFB) ? iStopY = vOFB.y : iStopY = vBound.y;

			if  (!iSeedOFB)
			{
				for (iY = iInitY; iY >= iStopY; iY--)
					TestCell (iStopXNeg, iY, iZ);
			}
			else if (vBound.y+iOrY < SIZE_Y) // added
			{
				for (iY = SIZE_YM1-iOrY; iY >= iStopY; iY--)
					TestCell (iStopXNeg, iY, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
		if (iInitX+iOrX < 0)		iInitX = -iOrX;

		for (iX = iInitX-1; iX >= iOFBStopXNeg; iX--)
			TestCell (iX, iInitY, iInitZ);
	}

	// X positive traversal
	iInitX = iCpyInitX;

	if (!iOFBPos)
	{
		if (!iSeedOFB)
		{
			if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
			if (iInitX+iOrX < 0)		iInitX = -iOrX;

			for (iX = iInitX; iX <= iStopXPos; iX++)
				TestCell (iX, iInitY, iInitZ);
		}

		if (bTurnPos)
		{
			// determine stop point (y extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iStopXPos, iInitY, iZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopY;
			(iYOFB) ? iStopY = vOFB.y : iStopY = vBound.y;

			if (!iSeedOFB)
			{
				for (iY = iInitY; iY >= iStopY; iY--)
					TestCell (iStopXPos, iY, iZ);
			}	
			else if (vBound.y+iOrY < SIZE_Y) // added
			{
				for (iY = SIZE_YM1-iOrY; iY >= iStopY; iY--)
					TestCell (iStopXPos, iY, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
		if (iInitX+iOrX < 0)		iInitX = -iOrX;

		for (iX = iInitX; iX <= iOFBStopXPos; iX++) 
			TestCell (iX, iInitY, iInitZ);
	}
}

//
// The following procedures are applied only to the region
// pointed in the diagram below:
//
//  Y ^- - - - - - - - - - -|
//    | \                 / 
//    |   \             /   |
//    |     \         /     
//    |       \     /       |
//    |         \ /         
//    |         / \         |
//    |       /     \       
//    |     /         \     |
//    |   /     here    \   
//    | /                 \ | 
//    *---------------------> Z
//  any X
// 

void Vis3D::XYTraversalPos (int iX, int iY, int iZ, int iSeedXOFB, int iSeedYOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	int iStopXNeg, iStopXPos; // stop points
	int iInitX, iInitY, iInitZ; // seed cell

	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// put seed cell within X bounds
	if (iSeedXOFB)
	{
		if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
		if (iInitX+iOrX < 0)		iInitX = -iOrX;
	}

	// determine X extents
	// -x
	if (isInsideVFBnd (-abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iXOFB)
			iStopXNeg = -abs(iInitY);
		else
			iStopXNeg = vOFB.x;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iXOFB)
			iStopXNeg = vBound.x;
		else
			iStopXNeg = vOFB.x;
	}

	// +x
	if (isInsideVFBnd (abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iXOFB)
			iStopXPos = abs(iInitY);
		else
			iStopXPos = vOFB.x;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iXOFB)
			iStopXPos = vBound.x;
		else
			iStopXPos = vOFB.x;
	}

	for (iX = iInitX-1; iX >= iStopXNeg; iX--)
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iX, iInitY, iInitZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopY;
		(iYOFB) ? iStopY = vOFB.y :	iStopY = vBound.y;

		if (!iSeedYOFB)
		{
			for (iY = iInitY; iY <= iStopY; iY++)
				TestCell (iX, iY, iInitZ);
		}
		else if (vBound.y+iOrY >= 0)
		{
			for (iY = -iOrY; iY <= iStopY; iY++)
				TestCell (iX, iY, iInitZ);
		}
	}

	for (iX = iInitX; iX <= iStopXPos; iX++)
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iX, iInitY, iInitZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopY;
		(iYOFB) ? iStopY = vOFB.y :	iStopY = vBound.y;

		if (!iSeedYOFB)
		{
			for (iY = iInitY; iY <= iStopY; iY++)
				TestCell (iX, iY, iInitZ);
		}
		else if (vBound.y+iOrY >= 0)
		{
			for (iY = -iOrY; iY <= iStopY; iY++)
				TestCell (iX, iY, iInitZ);
		}
	}
}

void Vis3D::XTraversalPos (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	BOOL bTurnNeg, bTurnPos; // make a 90deg turn during the traversal? 
	int iStopXNeg, // stop points
		iStopXPos;
	int iInitX, iInitY, iInitZ; // seed cell
	int iCpyInitX = iX;

	int iOFBNeg, iOFBStopXNeg,
		iOFBPos, iOFBStopXPos;

	bTurnNeg = bTurnPos = FALSE;
	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// determine X extents
	// -x
	if (isInsideVFBnd (-abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopXNeg = -abs(iInitY);
		bTurnNeg = TRUE;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopXNeg = vBound.x;
		bTurnNeg = FALSE;
	}
	iOFBNeg = iXOFB;
	iOFBStopXNeg = vOFB.x;

	// +x
	if (isInsideVFBnd (abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopXPos = abs(iInitY);
		bTurnPos = TRUE;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopXPos = vBound.x;
		bTurnPos = FALSE;
	}
	iOFBPos = iXOFB;
	iOFBStopXPos = vOFB.x;

	// X negative traversal

	if (!iOFBNeg)
	{
		if (!iSeedOFB)
		{
			if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
			if (iInitX+iOrX < 0)		iInitX = -iOrX;

			for (iX = iInitX; iX >= iStopXNeg; iX--)
				TestCell (iX, iInitY, iInitZ);
		}

		if (bTurnNeg)
		{
			// determine stop point (y extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iStopXNeg, iInitY, iZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopY;
			(iYOFB) ? iStopY = vOFB.y : iStopY = vBound.y;

			if (!iSeedOFB)
			{
				for (iY = iInitY; iY <= iStopY; iY++)
					TestCell (iStopXNeg, iY, iZ);
			}
			else if (vBound.y+iOrY >= 0) // added
			{
				for (iY = -iOrY; iY <= iStopY; iY++)
					TestCell (iStopXNeg, iY, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
		if (iInitX+iOrX < 0)		iInitX = -iOrX;

		for (iX = iInitX-1; iX >= iOFBStopXNeg; iX--)
			TestCell (iX, iInitY, iInitZ);
	}

	// X positive traversal
	iInitX = iCpyInitX;

	if (!iOFBPos)
	{
		if (!iSeedOFB)
		{
			if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
			if (iInitX+iOrX < 0)		iInitX = -iOrX;

			for (iX = iInitX; iX <= iStopXPos; iX++)
				TestCell (iX, iInitY, iInitZ);
		}

		if (bTurnPos)
		{
			// determine stop point (y extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iStopXPos, iInitY, iZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopY;
			(iYOFB) ? iStopY = vOFB.y : iStopY = vBound.y;

			if (!iSeedOFB)
			{
				for (iY = iInitY; iY <= iStopY; iY++)
					TestCell (iStopXPos, iY, iZ);
			}
			else if (vBound.y+iOrY >= 0) // added
			{
				for (iY = -iOrY; iY <= iStopY; iY++)
					TestCell (iStopXPos, iY, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
		if (iInitX+iOrX < 0)		iInitX = -iOrX;

		for (iX = iInitX; iX <= iOFBStopXPos; iX++) 
			TestCell (iX, iInitY, iInitZ);
	}
}

//
// The following procedures are applied only to the region
// pointed in the diagram below:
//
//  X ^- - - - - - - - - - -|
//    | \                 / 
//    |   \     here    /   |
//    |     \         /     
//    |       \     /       |
//    |         \ /         
//    |         / \         |
//    |       /     \       
//    |     /         \     |
//    |   /             \   
//    | /                 \ | 
//    *---------------------> Y
//  any Z
// 

void Vis3D::ZXTraversalNeg (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	int iStopZNeg, iStopZPos; // stop points
	int iInitX, iInitY, iInitZ; // seed cell

	iInitX = iX; iInitY = iY; iInitZ = iZ;
	
	// chunk removed
	// put seed cell within Z bounds
//	if (iSeedOFB)
//	{
//		if (iInitZ+iOrZ > SIZE_ZM1) iInitZ = SIZE_ZM1-iOrZ;
//		if (iInitZ+iOrZ < 0)		iInitZ = -iOrZ;
//	}

	// determine Z extents
	// -z
	if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, -abs(iInitX)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iZOFB)
			iStopZNeg = -abs(iInitX);
		else
			iStopZNeg = vOFB.z;
	}
	else
	{
		// determine stop point (z extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, 0, -1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iZOFB)
			iStopZNeg = vBound.z;
		else
			iStopZNeg = vOFB.z;
	}

	// +z
	if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, abs(iInitX)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iZOFB)
			iStopZPos = abs(iInitX);
		else
			iStopZPos = vOFB.z;
	}
	else
	{
		// determine stop point (z extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iZ), VECTOR3(0, 0, +1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iZOFB)
			iStopZPos = vBound.z;
		else
			iStopZPos = vOFB.z;
	}

	for (iZ = iInitZ-1; iZ >= iStopZNeg; iZ--)
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iInitZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopX;
		(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

		if (!iSeedOFB)
		{
			for (iX = iInitX; iX >= iStopX; iX--)
				TestCell (iX, iInitY, iZ);
		}
		else
		for (iX = SIZE_XM1-iOrX; iX >= iStopX; iX--)
			TestCell (iX, iInitY, iZ);
	}


	for (iZ = iInitZ; iZ <= iStopZPos; iZ++)
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopX;
		(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

		if (!iSeedOFB)
		{
			for (iX = iInitX; iX >= iStopX; iX--)
				TestCell (iX, iInitY, iZ);
		}
		else
		for (iX = SIZE_XM1-iOrX; iX >= iStopX; iX--)
			TestCell (iX, iInitY, iZ);
	}
}

void Vis3D::Z1TraversalNeg (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	BOOL bTurnNeg, bTurnPos; // make a 90deg turn during the traversal? 
	int iStopZNeg,  // stop points
		iStopZPos;
	int iInitX, iInitY, iInitZ; // seed cell

	int iOFBNeg, iOFBStopZNeg,
		iOFBPos, iOFBStopZPos;

	bTurnNeg = bTurnPos = FALSE;
	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// put cell within Z bounds
	int iCpyInitZ = iInitZ;
	if (iInitZ+iOrZ > SIZE_ZM1) iInitZ = SIZE_ZM1-iOrZ;
	if (iInitZ+iOrZ < 0)		iInitZ = -iOrZ;

	// determine Z extents
	// -z
	if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, -abs(iInitX)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopZNeg = -abs(iInitX);
		bTurnNeg = TRUE;
	}
	else
	{
		// determine stop point (z extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iCpyInitZ), VECTOR3(0, 0, -1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopZNeg = vBound.z;
		bTurnNeg = FALSE;
	}
	iOFBNeg = iZOFB;
	iOFBStopZNeg = vOFB.z;

	// +z
	if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, abs(iInitX)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopZPos = abs(iInitX);
		bTurnPos = TRUE;
	}
	else
	{
		// determine stop point (z extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iCpyInitZ), VECTOR3(0, 0, +1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopZPos = vBound.z;
		bTurnPos = FALSE;
	}
	iOFBPos = iZOFB;
	iOFBStopZPos = vOFB.z;

	// Z negative traversal

	if (!iOFBNeg)
	{
		if (!iSeedOFB)
		{
			for (iZ = iInitZ; iZ >= iStopZNeg; iZ--)
				TestCell (iInitX, iInitY, iZ);
		}

		if (bTurnNeg)
		{
			// determine stop point (x extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iStopZNeg), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopX;
			(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

			if  (!iSeedOFB)
			{
				for (iX = iInitX; iX >= iStopX; iX--)
					TestCell (iX, iY, iStopZNeg);
			}
			else
			for (iX = SIZE_XM1-iOrX; iX >= iStopX; iX--)
				TestCell (iX, iY, iStopZNeg);
		}
	}
	else if (!iSeedOFB)
	{
		for (iZ = iInitZ-1; iZ >= iOFBStopZNeg; iZ--)
			TestCell (iInitX, iInitY, iZ);
	}

	// Z positive traversal

	if (!iOFBPos)
	{
		if (!iSeedOFB)
		{
			for (iZ = iInitZ; iZ <= iStopZPos; iZ++)
				TestCell (iInitX, iInitY, iZ);
		}

		if (bTurnPos)
		{
			// determine stop point (x extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iStopZPos), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopX;
			(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

			if (!iSeedOFB)
			{
				for (iX = iInitX; iX >= iStopX; iX--)
					TestCell (iX, iY, iStopZPos);
			}	
			else
			for (iX = SIZE_XM1-iOrX; iX >= iStopX; iX--)
				TestCell (iX, iY, iStopZPos);
		}
	}
	else if (!iSeedOFB)
	{
		for (iZ = iInitZ; iZ <= iOFBStopZPos; iZ++)
			TestCell (iInitX, iInitY, iZ);
	}
}

//
// The following procedures are applied only to the region
// pointed in the diagram below:
//
//  X ^- - - - - - - - - - -|
//    | \                 / 
//    |   \             /   |
//    |     \         /     
//    |       \     /       |
//    |         \ /         
//    |         / \         |
//    |       /     \       
//    |     /         \     |
//    |   /     here    \   
//    | /                 \ | 
//    *---------------------> Z
//  any Y
// 

void Vis3D::ZXTraversalPos (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	int iStopZNeg, iStopZPos; // stop points
	int iInitX, iInitY, iInitZ; // seed cell

	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// determine Z extents
	// -z
	if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, -abs(iInitX)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iZOFB)
			iStopZNeg = -abs(iInitX);
		else
			iStopZNeg = vOFB.z;
	}
	else
	{
		// determine stop point (z extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, 0, -1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iZOFB)
			iStopZNeg = vBound.z;
		else
			iStopZNeg = vOFB.z;
	}

	// +z
	if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, abs(iInitX)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iZOFB)
			iStopZPos = abs(iInitX);
		else
			iStopZPos = vOFB.z;
	}
	else
	{
		// determine stop point (z extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iZ), VECTOR3(0, 0, +1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iZOFB)
			iStopZPos = vBound.z;
		else
			iStopZPos = vOFB.z;
	}

	for (iZ = iInitZ-1; iZ >= iStopZNeg; iZ--)
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopX;
		(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

		if (!iSeedOFB)
		{
			for (iX = iInitX; iX <= iStopX; iX++)
				TestCell (iX, iInitY, iZ);
		}
		else
		for (iX = -iOrX; iX <= iStopX; iX++)
			TestCell (iX, iInitY, iZ);
	}

	for (iZ = iInitZ; iZ <= iStopZPos; iZ++)
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopX;
		(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

		if (!iSeedOFB)
		{
			for (iX = iInitX; iX <= iStopX; iX++)
				TestCell (iX, iInitY, iZ);
		}
		else
		for (iX = -iOrX; iX <= iStopX; iX++)
			TestCell (iX, iInitY, iZ);
	}
}

void Vis3D::Z1TraversalPos (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	BOOL bTurnNeg, bTurnPos; // make a 90deg turn during the traversal? 
	int iStopZNeg, // stop points
		iStopZPos;
	int iInitX, iInitY, iInitZ; // seed cell

	int iOFBNeg, iOFBStopZNeg,
		iOFBPos, iOFBStopZPos;

	bTurnNeg = bTurnPos = FALSE;
	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// put cell within Z bounds
	int iCpyInitZ = iInitZ;
	if (iInitZ+iOrZ > SIZE_ZM1) iInitZ = SIZE_ZM1-iOrZ;
	if (iInitZ+iOrZ < 0)		iInitZ = -iOrZ;

	// determine Z extents
	// -z
	if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, -abs(iInitX)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopZNeg = -abs(iInitX);
		bTurnNeg = TRUE;
	}
	else
	{
		// determine stop point (z extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iCpyInitZ), VECTOR3(0, 0, -1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopZNeg = vBound.z;
		bTurnNeg = FALSE;
	}
	iOFBNeg = iZOFB;
	iOFBStopZNeg = vOFB.z;

	// +z
	if (isInsideVFBnd (iInitX+iOrX, iInitY+iOrY, abs(iInitX)+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopZPos = abs(iInitX);
		bTurnPos = TRUE;
	}
	else
	{
		// determine stop point (z extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iCpyInitZ), VECTOR3(0, 0, +1), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopZPos = vBound.z;
		bTurnPos = FALSE;
	}
	iOFBPos = iZOFB;
	iOFBStopZPos = vOFB.z;

	// Z negative traversal

	if (!iOFBNeg)
	{
		if (!iSeedOFB)
		{
			for (iZ = iInitZ; iZ >= iStopZNeg; iZ--)
				TestCell (iInitX, iInitY, iZ);
		}

		if (bTurnNeg)
		{
			// determine stop point (x extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iStopZNeg), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopX;
			(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

			if (!iSeedOFB)
			{
				for (iX = iInitX; iX <= iStopX; iX++)
					TestCell (iX, iY, iStopZNeg);
			}
			else
			for (iX = -iOrX; iX <= iStopX; iX++)
				TestCell (iX, iY, iStopZNeg);
		}
	}
	else if (!iSeedOFB)
	{
		for (iZ = iInitZ-1; iZ >= iOFBStopZNeg; iZ--)
			TestCell (iInitX, iInitY, iZ);
	}

	// Z positive traversal

	if (!iOFBPos)
	{
		if (!iSeedOFB)
		{
			for (iZ = iInitZ; iZ <= iStopZPos; iZ++)
				TestCell (iInitX, iInitY, iZ);
		}

		if (bTurnPos)
		{
			// determine stop point (x extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iStopZPos), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopX;
			(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

			if (!iSeedOFB)
			{
				for (iX = iInitX; iX <= iStopX; iX++)
					TestCell (iX, iY, iStopZPos);
			}
			else
			for (iX = -iOrX; iX <= iStopX; iX++)
				TestCell (iX, iY, iStopZPos);
		}
	}
	else if (!iSeedOFB)
	{
		for (iZ = iInitZ; iZ <= iOFBStopZPos; iZ++)
			TestCell (iInitX, iInitY, iZ);
	}
}

//***************************************************************************
//
//
//
// SPECIAL 'DIAGONAL' CASES
//
//
//
//***************************************************************************

void Vis3D::YXTraversalNegDg (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	int iStopYNeg, iStopYPos; // stop points
	int iInitX, iInitY, iInitZ; // seed cell

	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// determine Y extents
	// -y
	if (isInsideVFBnd (iInitX+iOrX, -abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iYOFB)
			iStopYNeg = -abs(iInitX);
		else
			iStopYNeg = vOFB.y;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iYOFB)
			iStopYNeg = vBound.y;
		else
			iStopYNeg = vOFB.y;
	}

	// +y
	if (isInsideVFBnd (iInitX+iOrX, abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iYOFB)
			iStopYPos = abs(iInitX);
		else
			iStopYPos = vOFB.y;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iYOFB)
			iStopYPos = vBound.y;
		else
			iStopYPos = vOFB.y;
	}

	for (iY = iInitY-1; iY >= iStopYNeg; iY--)
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iInitZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopX;
		(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

		if (!iSeedOFB)
		{
			for (iX = iInitX; iX >= iStopX; iX--)
				TestCell (iX, iY, iInitZ);
		}
		else
		{
			for (iX = SIZE_XM1-iOrX; iX >= iStopX; iX--)
				TestCell (iX, iY, iInitZ);
		}
	}


	for (iY = iInitY; iY <= iStopYPos; iY++)
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iInitZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopX;
		(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

		if (!iSeedOFB)
		{
			for (iX = iInitX; iX >= iStopX; iX--)
				TestCell (iX, iY, iInitZ);
		}
		else
		{
			for (iX = SIZE_XM1-iOrX; iX >= iStopX; iX--)
				TestCell (iX, iY, iInitZ);
		}
	}
}

void Vis3D::Y1TraversalNegDg (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	BOOL bTurnNeg, bTurnPos; // make a 90deg turn during the traversal? 
	int iStopYNeg,  // stop points
		iStopYPos;
	int iInitX, iInitY, iInitZ; // seed cell

	int iOFBNeg, iOFBStopYNeg,
		iOFBPos, iOFBStopYPos;

	bTurnNeg = bTurnPos = FALSE;
	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// put cell within Y bounds
	int iCpyInitY = iInitY;
	if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
	if (iInitY+iOrY < 0)		iInitY = -iOrY;

	// determine Y extents
	// -y
	if (isInsideVFBnd (iInitX+iOrX, -abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopYNeg = -abs(iInitX);
		bTurnNeg = TRUE;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iCpyInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopYNeg = vBound.y;
		bTurnNeg = FALSE;
	}
	iOFBNeg = iYOFB;
	iOFBStopYNeg = vOFB.y;

	// +y
	if (isInsideVFBnd (iInitX+iOrX, abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopYPos = abs(iInitX);
		bTurnPos = TRUE;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iCpyInitY, iInitZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopYPos = vBound.y;
		bTurnPos = FALSE;
	}
	iOFBPos = iYOFB;
	iOFBStopYPos = vOFB.y;

	// Y negative traversal

	if (!iOFBNeg)
	{
		if (!iSeedOFB)
		{
			for (iY = iInitY; iY >= iStopYNeg; iY--)
				TestCell (iInitX, iY, iInitZ);
		}

		if (bTurnNeg)
		{
			// determine stop point (x extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iStopYNeg, iZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopX;
			(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

			if  (!iSeedOFB)
			{
				for (iX = iInitX; iX >= iStopX; iX--)
					TestCell (iX, iStopYNeg, iZ);
			}
			else
			{
				for (iX = SIZE_XM1-iOrX; iX >= iStopX; iX--)
					TestCell (iX, iStopYNeg, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		for (iY = iInitY-1; iY >= iOFBStopYNeg; iY--)
			TestCell (iInitX, iY, iInitZ);
	}

	// Y positive traversal

	if (!iOFBPos)
	{
		if (!iSeedOFB)
		{
			for (iY = iInitY; iY <= iStopYPos; iY++)
				TestCell (iInitX, iY, iInitZ);
		}

		if (bTurnPos)
		{
			// determine stop point (x extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iStopYPos, iZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopX;
			(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

			if (!iSeedOFB)
			{
				for (iX = iInitX; iX >= iStopX; iX--)
					TestCell (iX, iStopYPos, iZ);
			}	
			else
			{
				for (iX = SIZE_XM1-iOrX; iX >= iStopX; iX--)
					TestCell (iX, iStopYPos, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		for (iY = iInitY; iY <= iOFBStopYPos; iY++)
			TestCell (iInitX, iY, iInitZ);
	}
}

void Vis3D::YXTraversalPosDg (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	int iStopYNeg, iStopYPos; // stop points
	int iInitX, iInitY, iInitZ; // seed cell

	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// determine Y extents
	// -y
	if (isInsideVFBnd (iInitX+iOrX, -abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iYOFB)
			iStopYNeg = -abs(iInitX);
		else
			iStopYNeg = vOFB.y;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iYOFB)
			iStopYNeg = vBound.y;
		else
			iStopYNeg = vOFB.y;
	}

	// +y
	if (isInsideVFBnd (iInitX+iOrX, abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iYOFB)
			iStopYPos = abs(iInitX);
		else
			iStopYPos = vOFB.y;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iYOFB)
			iStopYPos = vBound.y;
		else
			iStopYPos = vOFB.y;
	}

	for (iY = iInitY-1; iY >= iStopYNeg; iY--)
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iInitZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopX;
		(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

		if (!iSeedOFB)
		{
			for (iX = iInitX; iX <= iStopX; iX++)
				TestCell (iX, iY, iInitZ);
		}
		else
		{
			for (iX = -iOrX; iX <= iStopX; iX++)
				TestCell (iX, iY, iInitZ);
		}
	}

	for (iY = iInitY; iY <= iStopYPos; iY++)
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iY, iInitZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopX;
		(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

		if (!iSeedOFB)
		{
			for (iX = iInitX; iX <= iStopX; iX++)
				TestCell (iX, iY, iInitZ);
		}
		else
		{
			for (iX = -iOrX; iX <= iStopX; iX++)
				TestCell (iX, iY, iInitZ);
		}
	}
}

void Vis3D::Y1TraversalPosDg (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	BOOL bTurnNeg, bTurnPos; // make a 90deg turn during the traversal? 
	int iStopYNeg, // stop points
		iStopYPos;
	int iInitX, iInitY, iInitZ; // seed cell

	int iOFBNeg, iOFBStopYNeg,
		iOFBPos, iOFBStopYPos;

	bTurnNeg = bTurnPos = FALSE;
	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// put cell within Y bounds
	int iCpyInitY = iInitY;
	if (iInitY+iOrY > SIZE_YM1) iInitY = SIZE_YM1-iOrY;
	if (iInitY+iOrY < 0)		iInitY = -iOrY;

	// determine Y extents
	// -y
	if (isInsideVFBnd (iInitX+iOrX, -abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopYNeg = -abs(iInitX);
		bTurnNeg = TRUE;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iCpyInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopYNeg = vBound.y;
		bTurnNeg = FALSE;
	}
	iOFBNeg = iYOFB;
	iOFBStopYNeg = vOFB.y;

	// +y
	if (isInsideVFBnd (iInitX+iOrX, abs(iInitX)+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopYPos = abs(iInitX);
		bTurnPos = TRUE;
	}
	else
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iCpyInitY, iInitZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopYPos = vBound.y;
		bTurnPos = FALSE;
	}
	iOFBPos = iYOFB;
	iOFBStopYPos = vOFB.y;

	// Y negative traversal

	if (!iOFBNeg)
	{
		if (!iSeedOFB)
		{
			for (iY = iInitY; iY >= iStopYNeg; iY--)
				TestCell (iInitX, iY, iInitZ);
		}

		if (bTurnNeg)
		{
			// determine stop point (x extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iStopYNeg, iZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopX;
			(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

			if (!iSeedOFB)
			{
				for (iX = iInitX; iX <= iStopX; iX++)
					TestCell (iX, iStopYNeg, iZ);
			}
			else
			{
				for (iX = -iOrX; iX <= iStopX; iX++)
					TestCell (iX, iStopYNeg, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		for (iY = iInitY-1; iY >= iOFBStopYNeg; iY--)
			TestCell (iInitX, iY, iInitZ);
	}

	// Y positive traversal

	if (!iOFBPos)
	{
		if (!iSeedOFB)
		{
			for (iY = iInitY; iY <= iStopYPos; iY++)
				TestCell (iInitX, iY, iInitZ);
		}

		if (bTurnPos)
		{
			// determine stop point (x extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iStopYPos, iZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopX;
			(iXOFB) ? iStopX = vOFB.x : iStopX = vBound.x;

			if (!iSeedOFB)
			{
				for (iX = iInitX; iX <= iStopX; iX++)
					TestCell (iX, iStopYPos, iZ);
			}
			else
			{
				for (iX = -iOrX; iX <= iStopX; iX++)
					TestCell (iX, iStopYPos, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		for (iY = iInitY; iY <= iOFBStopYPos; iY++)
			TestCell (iInitX, iY, iInitZ);
	} 
}

//
// The following procedures are applied only to the region
// pointed in the diagram below:
//
//  Y ^- - - - - - - - - - -|
//    | \                 / 
//    |   \     here    /   |
//    |     \         /     
//    |       \     /       |
//    |         \ /         
//    |         / \         |
//    |       /     \       
//    |     /         \     |
//    |   /             \   
//    | /                 \ | 
//    *---------------------> Z
//  any X
// 

void Vis3D::XYTraversalNegDg (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	int iStopXNeg, iStopXPos; // stop points
	int iInitX, iInitY, iInitZ; // seed cell

	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// determine X extents
	// -x
	if (isInsideVFBnd (-abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iXOFB)
			iStopXNeg = -abs(iInitY);
		else
			iStopXNeg = vOFB.x;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iXOFB)
			iStopXNeg = vBound.x;
		else
			iStopXNeg = vOFB.x;
	}

	// +x
	if (isInsideVFBnd (abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iXOFB)
			iStopXPos = abs(iInitY);
		else
			iStopXPos = vOFB.x;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iXOFB)
			iStopXPos = vBound.x;
		else
			iStopXPos = vOFB.x;
	}

	for (iX = iInitX-1; iX >= iStopXNeg; iX--)
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopY;
		(iYOFB) ? iStopY = vOFB.y :	iStopY = vBound.y;

		if (!iSeedOFB)
		{
			for (iY = iInitY; iY >= iStopY; iY--)
				TestCell (iX, iY, iInitZ);
		}
		else if (vBound.y+iOrY < SIZE_Y) // added
		{
			for (iY = SIZE_YM1-iOrY; iY >= iStopY; iY--)
				TestCell (iX, iY, iInitZ);
		}
	}


	for (iX = iInitX; iX <= iStopXPos; iX++)
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iX, iInitY, iInitZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopY;
		(iYOFB) ? iStopY = vOFB.y :	iStopY = vBound.y;

		if (!iSeedOFB)
		{
			for (iY = iInitY; iY >= iStopY; iY--)
				TestCell (iX, iY, iInitZ);
		}
		else if (vBound.y+iOrY < SIZE_Y) // added
		{
			for (iY = SIZE_YM1-iOrY; iY >= iStopY; iY--)
				TestCell (iX, iY, iInitZ);
		}
	}
}

void Vis3D::XTraversalNegDg (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	BOOL bTurnNeg, bTurnPos; // make a 90deg turn during the traversal? 
	int iStopXNeg,  // stop points
		iStopXPos;
	int iInitX, iInitY, iInitZ; // seed cell

	int iOFBNeg, iOFBStopXNeg,
		iOFBPos, iOFBStopXPos;

	bTurnNeg = bTurnPos = FALSE;
	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// put cell within X bounds
	int iCpyInitX = iInitX;
	if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
	if (iInitX+iOrX < 0)		iInitX = -iOrX;

	// determine X extents
	// -x
	if (isInsideVFBnd (-abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopXNeg = -abs(iInitY);
		bTurnNeg = TRUE;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iCpyInitX, iInitY, iInitZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopXNeg = vBound.x;
		bTurnNeg = FALSE;
	}
	iOFBNeg = iXOFB;
	iOFBStopXNeg = vOFB.x;

	// +x
	if (isInsideVFBnd (abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopXPos = abs(iInitY);
		bTurnPos = TRUE;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iCpyInitX, iInitY, iInitZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopXPos = vBound.x;
		bTurnPos = FALSE;
	}
	iOFBPos = iXOFB;
	iOFBStopXPos = vOFB.x;

	// X negative traversal

	if (!iOFBNeg)
	{
		if (!iSeedOFB)
		{
			for (iX = iInitX; iX >= iStopXNeg; iX--)
				TestCell (iX, iInitY, iInitZ);
		}

		if (bTurnNeg)
		{
			// determine stop point (y extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iStopXNeg, iInitY, iZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopY;
			(iYOFB) ? iStopY = vOFB.y : iStopY = vBound.y;

			if  (!iSeedOFB)
			{
				for (iY = iInitY; iY >= iStopY; iY--)
					TestCell (iStopXNeg, iY, iZ);
			}
			else if (vBound.y+iOrY < SIZE_Y) // added
			{
				for (iY = SIZE_YM1-iOrY; iY >= iStopY; iY--)
					TestCell (iStopXNeg, iY, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		for (iX = iInitX-1; iX >= iOFBStopXNeg; iX--)
			TestCell (iX, iInitY, iInitZ);
	}

	// X positive traversal
	iInitX = iCpyInitX;

	if (!iOFBPos)
	{
		if (!iSeedOFB)
		{
			for (iX = iInitX; iX <= iStopXPos; iX++)
				TestCell (iX, iInitY, iInitZ);
		}

		if (bTurnPos)
		{
			// determine stop point (y extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iStopXPos, iInitY, iZ), VECTOR3(0, -1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopY;
			(iYOFB) ? iStopY = vOFB.y : iStopY = vBound.y;

			if (!iSeedOFB)
			{
				for (iY = iInitY; iY >= iStopY; iY--)
					TestCell (iStopXPos, iY, iZ);
			}	
			else if (vBound.y+iOrY < SIZE_Y) // added
			{
				for (iY = SIZE_YM1-iOrY; iY >= iStopY; iY--)
					TestCell (iStopXPos, iY, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		for (iX = iInitX; iX <= iOFBStopXPos; iX++) 
			TestCell (iX, iInitY, iInitZ);
	}
}

//
// The following procedures are applied only to the region
// pointed in the diagram below:
//
//  Y ^- - - - - - - - - - -|
//    | \                 / 
//    |   \             /   |
//    |     \         /     
//    |       \     /       |
//    |         \ /         
//    |         / \         |
//    |       /     \       
//    |     /         \     |
//    |   /     here    \   
//    | /                 \ | 
//    *---------------------> Z
//  any X
// 

void Vis3D::XYTraversalPosDg (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	int iStopXNeg, iStopXPos; // stop points
	int iInitX, iInitY, iInitZ; // seed cell

	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// determine X extents
	// -x
	if (isInsideVFBnd (-abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iXOFB)
			iStopXNeg = -abs(iInitY);
		else
			iStopXNeg = vOFB.x;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iInitZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iXOFB)
			iStopXNeg = vBound.x;
		else
			iStopXNeg = vOFB.x;
	}

	// +x
	if (isInsideVFBnd (abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		if (!iXOFB)
			iStopXPos = abs(iInitY);
		else
			iStopXPos = vOFB.x;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iInitX, iInitY, iZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		if (!iXOFB)
			iStopXPos = vBound.x;
		else
			iStopXPos = vOFB.x;
	}

	for (iX = iInitX-1; iX >= iStopXNeg; iX--)
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iX, iInitY, iInitZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopY;
		(iYOFB) ? iStopY = vOFB.y :	iStopY = vBound.y;

		if (!iSeedOFB)
		{
			for (iY = iInitY; iY <= iStopY; iY++)
				TestCell (iX, iY, iInitZ);
		}
		else if (vBound.y+iOrY >= 0)
		{
			for (iY = -iOrY; iY <= iStopY; iY++)
				TestCell (iX, iY, iInitZ);
		}
	}

	for (iX = iInitX; iX <= iStopXPos; iX++)
	{
		// determine stop point (y extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iX, iInitY, iInitZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

		int iStopY;
		(iYOFB) ? iStopY = vOFB.y :	iStopY = vBound.y;

		if (!iSeedOFB)
		{
			for (iY = iInitY; iY <= iStopY; iY++)
				TestCell (iX, iY, iInitZ);
		}
		else if (vBound.y+iOrY >= 0)
		{
			for (iY = -iOrY; iY <= iStopY; iY++)
				TestCell (iX, iY, iInitZ);
		}
	}
}

void Vis3D::XTraversalPosDg (int iX, int iY, int iZ, int iSeedOFB)
{
	int iOFB, iXOFB, iYOFB, iZOFB; // out-of-bounds flag
	VECTOR3 vOFB; // out-of-bounds vector

	BOOL bTurnNeg, bTurnPos; // make a 90deg turn during the traversal? 
	int iStopXNeg, // stop points
		iStopXPos;
	int iInitX, iInitY, iInitZ; // seed cell

	int iOFBNeg, iOFBStopXNeg,
		iOFBPos, iOFBStopXPos;

	bTurnNeg = bTurnPos = FALSE;
	iInitX = iX; iInitY = iY; iInitZ = iZ;

	// put cell within X bounds
	int iCpyInitX = iInitX;
	if (iInitX+iOrX > SIZE_XM1) iInitX = SIZE_XM1-iOrX;
	if (iInitX+iOrX < 0)		iInitX = -iOrX;

	// determine X extents
	// -x
	if (isInsideVFBnd (-abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopXNeg = -abs(iInitY);
		bTurnNeg = TRUE;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iCpyInitX, iInitY, iInitZ), VECTOR3(-1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopXNeg = vBound.x;
		bTurnNeg = FALSE;
	}
	iOFBNeg = iXOFB;
	iOFBStopXNeg = vOFB.x;

	// +x
	if (isInsideVFBnd (abs(iInitY)+iOrX, iInitY+iOrY, iInitZ+iOrZ, vOFB, iOFB, iXOFB, iYOFB, iZOFB))
	{
		iStopXPos = abs(iInitY);
		bTurnPos = TRUE;
	}
	else
	{
		// determine stop point (x extent)
		VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iCpyInitX, iInitY, iInitZ), VECTOR3(+1, 0, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);
		iStopXPos = vBound.x;
		bTurnPos = FALSE;
	}
	iOFBPos = iXOFB;
	iOFBStopXPos = vOFB.x;

	// X negative traversal

	if (!iOFBNeg)
	{
		if (!iSeedOFB)
		{
			for (iX = iInitX; iX >= iStopXNeg; iX--)
				TestCell (iX, iInitY, iInitZ);
		}

		if (bTurnNeg)
		{
			// determine stop point (y extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iStopXNeg, iInitY, iZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopY;
			(iYOFB) ? iStopY = vOFB.y : iStopY = vBound.y;

			if (!iSeedOFB)
			{
				for (iY = iInitY; iY <= iStopY; iY++)
					TestCell (iStopXNeg, iY, iZ);
			}
			else if (vBound.y+iOrY >= 0) // added 
			{
				for (iY = -iOrY; iY <= iStopY; iY++)
					TestCell (iStopXNeg, iY, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		for (iX = iInitX-1; iX >= iOFBStopXNeg; iX--)
			TestCell (iX, iInitY, iInitZ);
	}

	// X positive traversal
	iInitX = iCpyInitX;

	if (!iOFBPos)
	{
		if (!iSeedOFB)
		{
			for (iX = iInitX; iX <= iStopXPos; iX++)
				TestCell (iX, iInitY, iInitZ);
		}

		if (bTurnPos)
		{
			// determine stop point (y extent)
			VECTOR3 vBound = rayFrustumIntersectBnd (VECTOR3(iStopXPos, iInitY, iZ), VECTOR3(0, +1, 0), vOFB, iOFB, iXOFB, iYOFB, iZOFB);

			int iStopY;
			(iYOFB) ? iStopY = vOFB.y : iStopY = vBound.y;

			if (!iSeedOFB)
			{
				for (iY = iInitY; iY <= iStopY; iY++)
					TestCell (iStopXPos, iY, iZ);
			}
			else if (vBound.y+iOrY >= 0) // added
			{
				for (iY = -iOrY; iY <= iStopY; iY++)
					TestCell (iStopXPos, iY, iZ);
			}
		}
	}
	else if (!iSeedOFB)
	{
		for (iX = iInitX; iX <= iOFBStopXPos; iX++) 
			TestCell (iX, iInitY, iInitZ);
	}
}
